/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5694 $
 * $Date: 2012-06-27 14:31:05 +0200 (Mi, 27. Jun 2012) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.Alida.dataio.provider.cmdline;

import de.unihalle.informatik.Alida.dataio.ALDDataIOManagerCmdline;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOManagerException;
import de.unihalle.informatik.Alida.annotations.ALDDataIOProvider;
import de.unihalle.informatik.Alida.helpers.ALDClassInfo;
import de.unihalle.informatik.Alida.helpers.ALDParser;


import java.lang.reflect.Field;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.LinkedList;

/**
 * Abstract class providing basic methods for cmdline DataIO providers
 * according to Alida conventions.
 *<p>
 * These conventions are detailed in the documentation if the methods <code>readData</code> and <code>writeData</code> in this class below.
 * They handle IO from/to file and reading derived classes of the class handled by an dataIO provider.
 *<p>
 * Classes extending this class are expected to override the methods <code>parse</code> and
 * <code>formatAsString</code> which do the actual reading/parsing or writing/formating
 * subsequent to generic handling of Alida convention with respect to derived classes
 * and IO form/to file.
 *
 * @author posch
 *
 */

public abstract class ALDStandardizedDataIOCmdline implements ALDDataIOCmdline {

    /** As a convention a parameter value starting with this character indicates
     * that the actual parameter should be read from or writen to a file.
     * The filename is the remaining string after removing this character.
     */
    public static final char FILEIO_CHAR = '@';

    /** As a convention this character startes a derived class name,
     *  and is terminated by ':'
     */
	public static final char DERIVEDCLASS_CHAR = '$';


	/** Returns an object instantiated from valueString.
     * For the class of the object to be read see {@link ALDDataIOManagerCmdline#readData(Field,Class,String)}.
     * This method is assumed to directly parse the <code>valueString</code> and make no
     * prior interpretation regarding a file to use or derived class to return.
     *
	 * @param	field Field of object to be returned
	 * @param cl Class of object to be returned.
	 * @param valueString   Source from where to read data (e.g. a filename).
	 * @return Object read from valueString
	 *
	 *  @see de.unihalle.informatik.Alida.dataio.ALDDataIOManagerCmdline
	 */
	abstract public Object parse( Field field, Class<?> cl, String valueString);

	/** Returns the string representations of this object
     * This method is assumed to directly format the <code>obj</code> into the string return and make no
     * prior interpretation regarding a file to be used.
     *
     * @param obj	object to format
	 * @return string representations of the object
	 */
	abstract public String formatAsString( Object obj);

	/** Returns the external representations of this object using a formatString.
     * The <code>formatString</code> may be used define parts of the object to be formated
     * or specify format charateristics.
     * However, <code>formatString</code> should not interpreted in order to decide to
     * output the object directly to a file (as this is handled by <code>writeData</code>).
	 *<p>
	 *  The default implementation is equivalent to <code>formatAsString(obj)</code>.
     * @param obj	object to format
	 */
	public String formatAsString( Object obj, String formatString) {
		return formatAsString( obj);
	}

    /** Read an object using <code>valueString</code>.
     * For the class of the object to be read see {@link ALDDataIOCmdline}.
     * <p>
	 * If <code>valueString</code> starts with <code>FILEIO_CHAR</code> the value is read from a file
     * where the name of this file is the remaining string of <code>valueString</code>.
     * subsequent to <code>FILEIO_CHAR</code>.
     * <p>
     * The next step in interpreting the value string is scrutinze
     * whether an instance of a class deriving the class defined in <code>field</code>  or 
     * the class <code>cl</code> is to be returned.
     * This is indicated by starting the value string (either passed directly via <code>valueString</code>
	 * or the string read from file) with <code>DERIVEDCLASS_CHAR</code>.
	 * by <code>field</code> 
     * If this is the case all charaters up, but excluding, the next colon are
     * interpreted as this fully qualified class name.
     * Next it is checked if this class indeed is a proper extension
     * which in turn required this class to be annotated with {@link de.unihalle.informatik.Alida.annotations.ALDDerivedClass}.
     * If this check passes
     * a dataIO provider of this class is looked up and its <code>readData</code>
     * invoked with the value string after the deliminating colon.
     * <p>
     * If no deriving class is requested, the parse method of the 
	 * class defined in <code>field</code>  or 
     * the class <code>cl</code> is invoked on the value string.
     *
     * @param field Field of object to be returned.
     * @param cl Class of object to be returned.
     * @param valueString   Source from where to read data (e.g. a filename).
     * @return      Object read from source.
     *
	 */
	/* (non-Javadoc)
	 * @see de.unihalle.informatik.Alida.helpers.ALDDataIO#writeData(java.lang.Object, java.lang.String)
	 */
	@Override
	public Object readData(Field field, Class<?> cl, String valueString) {

		if ( field != null )
			cl = field.getType();

		// potentially read the value string from file
		valueString = getValueStringFromFile( valueString);

		// handle derived classes
		valueString = valueString.trim();
		if ( valueString.charAt(0) != DERIVEDCLASS_CHAR ) {
			// and parse the string
			return parse( field, cl, valueString);
		}
		
		int indexOfSep = valueString.indexOf( ':', 1);
    if ( indexOfSep == -1 ) {
    	System.err.println( "ALDStandardizedDataIOCmdline::readData derived class without <:> in " + valueString);
    	return null;
    }

    String className = valueString.substring( 1, indexOfSep);
    valueString = valueString.substring( indexOfSep+1);
    LinkedList<Class> derivedClassesFound = new LinkedList<Class>();

    Collection<Class> derivedClasses = ALDClassInfo.lookupExtendingClasses( cl);
    for ( Class derivedClass : derivedClasses ) {
    	if ( className.equals( derivedClass.getName()) ) {
    		derivedClassesFound.clear();
    		derivedClassesFound.add( derivedClass);
    		break;
    	} else if ( derivedClass.getName().endsWith( className) ) {
    		derivedClassesFound.add( derivedClass);
    	}
    }

    if ( derivedClassesFound.size() != 1 ) {
    	System.err.println( "ALDStandardizedDataIOCmdline::readData found " +
    			derivedClassesFound.size() + " derived classes matching <" + className + ">");
    	System.err.println( "      derived classes available:");
    	for ( Class derivedClass : derivedClasses ) {
    		System.err.println( "\t" + derivedClass.getName());
    	}
    	return null;
    }

    ALDDataIOCmdline provider = null;
    try {
	    provider = (ALDDataIOCmdline)(ALDDataIOManagerCmdline.getInstance().getProvider( derivedClassesFound.peek(), ALDDataIOCmdline.class));
    } catch (ALDDataIOManagerException e) {
	    // TODO Auto-generated catch block
	    e.printStackTrace();
    }
    if ( provider != null ) {
    	return provider.readData( null, derivedClassesFound.peek(), valueString);
    } else {
    	System.err.println( "ALDStandardizedDataIOCmdline::readData can't find provider for " + className);
    	return null;
    }
	}
	
	/**
	 * Return a string representation of given object value or print it to a file.
	 * <p>
	 * If the given <code>formatString</code> starts with <code>FILEIO_CHAR</code> standard out is used as
	 * target location, otherwise the string remaining subsequent to <code>FILEIO_CHAR</code> is interpreted as file name.
	 * In turn, if the remaining string starts with <code>+</code> the output is apended to the file (with <code>+</code> 
	 * removed as filename).
	 */
	/* (non-Javadoc)
	 * @see de.unihalle.informatik.Alida.helpers.ALDDataIO#writeData(java.lang.Object, java.lang.String)
	 */
	@Override
  	public String writeData(Object obj, String formatString) {
		// writer object
		BufferedWriter bufWriter = null;
			

		if ( formatString.length() > 0 && formatString.charAt( 0) == FILEIO_CHAR ) {
	    	try {
				boolean append = false;
				String filename;
				if ( formatString.length() > 1 && formatString.charAt( 1) == '+' ) {
					append = true;
					filename = formatString.substring(2);
				} else {
					filename = formatString.substring(1);
				}

	      		bufWriter = new BufferedWriter(new FileWriter( filename, append));
				bufWriter.write(this.formatAsString( obj) + "\n");
				bufWriter.flush();
				bufWriter.close();
      		} catch (IOException e) {
      			System.err.println("ALDEnumDataIOCmdline: cannot open writer on file " + 
      				"\"" + formatString + "\", printing to standard out!");
  				bufWriter = new BufferedWriter(new PrintWriter(System.out));
				try {
					bufWriter.write(this.formatAsString( obj) + "\n");
					bufWriter.flush();
					bufWriter.close();
      			} catch (IOException e2) {
      				System.err.println("ALDEnumDataIOCmdline: printing to standard out failed");
					return null;
      			}
      		}

			return null;
    	} else {
			if ( formatString.length() > 0 && 
				ALDParser.brackets.get( formatString.charAt(0)) != null) {
				
				return this.formatAsString( obj, formatString);
			} else {
				return this.formatAsString( obj);
			}
		}
	}

	/** Checks if the <code>formatString</code> passed starts with <code>FILEIO_CHAR</code>.
	 * This indicated that the value should be read from a file
     * where the name of this file is the remaining string of <code>formatString</code>
	 * subsequent to <code>FILEIO_CHAR</code>.
     * If this is case the complete content of the file is return as one string.
	 * Otherwise return the <code>formatString</code> unchanged.
	 *
	 * @param formatString
	 * @return argument or string read from file
	 */
	public static String getValueStringFromFile( String formatString) {
		formatString = formatString.trim();
		if ( formatString.length() == 0 )
			return formatString;

		formatString = formatString.trim();
		if ( formatString.charAt(0) == FILEIO_CHAR ) {
			String filename = formatString.substring( 1);
			try {
				BufferedReader reader = new BufferedReader( new FileReader(filename));
				StringBuffer buf = new StringBuffer();
				String line;
				while ( (line = reader.readLine()) != null ) {
					buf.append( line);
				}
				return( new String( buf));
			} catch (FileNotFoundException e) {
				System.err.println( "ALDNativeDataIOCmdline::readData cannot open reader for file " +
						filename);
			} catch (IOException e) {
				System.err.println( "ALDNativeDataIOCmdline::readData cannot read line from " +
						filename);
			}
		} 
		return formatString;
	}
}
