/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 6037 $
 * $Date: 2012-09-24 19:07:26 +0200 (Mo, 24. Sep 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.dataio.provider.helpers;

import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import de.unihalle.informatik.Alida.operator.*;


import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.dataio.ALDDataIOManagerSwing;
import de.unihalle.informatik.Alida.dataio.provider.swing.components.ALDSwingComponent;
import de.unihalle.informatik.Alida.dataio.provider.swing.components.ALDSwingComponentLabel;
import de.unihalle.informatik.Alida.dataio.provider.swing.events.ALDSwingValueChangeEvent;
import de.unihalle.informatik.Alida.dataio.provider.swing.events.ALDSwingValueChangeListener;
import de.unihalle.informatik.Alida.dataio.provider.swing.events.ALDSwingValueChangeReporter;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOManagerException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException.ALDDataIOProviderExceptionType;
import de.unihalle.informatik.Alida.gui.editor.ALDEditorFrame;

/**
 * Class for configuring sets of parameters via GUI in Alida.
 * <p>
 * This class provides a panel containing GUI elements to configure all
 * parameters of a set provided at initialization. Note that parameter labels
 * can be marked and unmarked externally. This feature is, e.g., be used during
 * parameter validation.
 * <p>
 * This panel allows to validate its parameters, i.e. if a certain parameter is
 * declared as required, but reading the parameter results in 'null', the
 * parameter is marked in the GUI.
 * 
 * @author moeller
 */

public class ALDParameterPanel extends ALDSwingValueChangeReporter implements
		ALDSwingValueChangeListener {

	/**
	 * The main panel containing everything.
	 */
	private JPanel mainPanel;

	/**
	 * Local debug flag, not accessible from outside of the class.
	 */
	private boolean debug = false;

	/**
	 * Flag indicating if null return values are ok.
	 * <p>
	 * Mostly null returns are not ok on top level operator configuration,
	 * however, operators being passed as parameters to other operators might
	 * accept null parameters.
	 */
	private boolean nullReturnsForbidden = false;

	/**
	 * Flag to disable editability of GUI elements.
	 * <p>
	 * If the flag is set <code>true</code>, parameter editing is not possible,
	 * i.e. the graphical elements in the window are not editable.
	 */
	private boolean onlyDisplay = false;

	/**
	 * List of parameter descriptors managed by this panel.
	 */
	private LinkedList<? extends ALDParameterDescriptor> paramDescriptors;

	/**
	 * Optional title string for panel.
	 */
	private String titleString;

	/**
	 * Maps a descriptor to its <code>JLabel</code> component in the panel.
	 */
	private HashMap<ALDParameterDescriptor, JLabel> labelMap = null;

	/**
	 * Maps a parameter descriptor to its component in the GUI.
	 */
	private HashMap<ALDParameterDescriptor, ALDSwingComponent> componentMap = null;

	/**
	 * Maps a parameter descriptor to its panel in the GUI.
	 */
	private HashMap<ALDParameterDescriptor, JPanel> panelMap = null;

	// TODO modify setParameter and validateParameters with the editorFrame
	private ALDEditorFrame editorFrame = null;
	private String link = null;

	public void setFrame(ALDEditorFrame editorFrame, String link) {
		this.editorFrame = editorFrame;
		this.link = link;
	}

	/**
	 * Default constructor.
	 * 
	 * @param paramDescrips
	 *            List of parameter descriptors to manage.
	 * @param title
	 *            Title of panel.
	 * @param topLevelCall
	 *            If true, null returns are forbidden.
	 */
	public ALDParameterPanel(
			LinkedList<? extends ALDParameterDescriptor> paramDescrips,
			String title, boolean topLevelCall) {
		this.mainPanel = new JPanel();
		this.paramDescriptors = paramDescrips;
		this.titleString = title;
		this.nullReturnsForbidden = topLevelCall;
		this.buildInputPanel();
	}

	/**
	 * Constructor.
	 * 
	 * @param paramDescrips
	 *            List of parameter descriptors to manage.
	 * @param title
	 *            Title of panel.
	 * @param topLevelCall
	 *            If true, null returns are forbidden.
	 * @param noEdit
	 *            If true, elements cannot be edited.
	 */
	public ALDParameterPanel(
			LinkedList<? extends ALDParameterDescriptor> paramDescrips,
			String title, boolean topLevelCall, boolean noEdit) {
		this.mainPanel = new JPanel();
		this.paramDescriptors = paramDescrips;
		this.titleString = title;
		this.nullReturnsForbidden = topLevelCall;
		this.onlyDisplay = noEdit;
		this.buildInputPanel();
	}

	/**
	 * Method to access associated panel.
	 * 
	 * @return Reference to the main panel.
	 */
	public JPanel getJPanel() {
		return this.mainPanel;
	}

	/**
	 * Changes the view mode.
	 * <p>
	 * There are the view modes 'standard' and 'advanced' available in the GUI.
	 * In standard mode only the parameters annotated as standard are visible,
	 * while in advanced mode all parameters are visible.
	 * 
	 * @param mode
	 *            Desired display mode.
	 */
	public void changeViewMode(Parameter.ExpertMode mode) {
		Set<ALDParameterDescriptor> keys = this.labelMap.keySet();
		for (ALDParameterDescriptor descr : keys) {
			if (mode.equals(Parameter.ExpertMode.STANDARD)
					&& descr.getHandlingMode().equals(
							Parameter.ExpertMode.ADVANCED)) {
				if (this.labelMap.get(descr) != null)
					this.labelMap.get(descr).setVisible(false);
				if (this.componentMap.get(descr) != null)
					this.componentMap.get(descr).getJComponent()
							.setVisible(false);
			} else if (descr.getHandlingMode().equals(
					Parameter.ExpertMode.STANDARD)) {
				if (this.labelMap.get(descr) != null)
					this.labelMap.get(descr).setVisible(true);
				if (this.componentMap.get(descr) != null)
					this.componentMap.get(descr).getJComponent()
							.setVisible(true);
			} else {
				if (this.labelMap.get(descr) != null)
					this.labelMap.get(descr).setVisible(true);
				if (this.componentMap.get(descr) != null)
					this.componentMap.get(descr).getJComponent()
							.setVisible(true);
			}
		}
	}

	/**
	 * Get value of given parameter from the GUI element.
	 * 
	 * @param descr
	 *            Descriptor of desired parameter
	 * @return Value of parameter, might be null.
	 * @throws ALDDataIOException
	 */
	public Object readParameter(ALDParameterDescriptor descr)
			throws ALDDataIOProviderException {

		if (this.debug)
			System.out.println("Reading parameter " + descr.getName());

		// get the value out of the parameter field
		if (!this.componentMap.containsKey(descr)) {
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.UNSPECIFIED_ERROR,
					"[ALDParameterPanel::readParameter()] Descriptor not found!");
		}
		try {
			Object obj = null;
			// TODO return an ALDLinkException if value is null or runOpNode
			// failed
			if (descr.isLink()) {
				String link = (String) ALDDataIOManagerSwing.getInstance()
						.readData(descr.getField(), String.class,
								this.componentMap.get(descr));
				if (link == null || link.isEmpty() || !link.contains("@"))
					return null;
//				if (this.editorFrame.getValue(link) == null
//						&& descr.isReverse())
//					if (!this.editorFrame.runOpNode(link.split("@")[0], true))
//						return null;
				obj = this.editorFrame.getValue(link);
			} else {
				obj = ALDDataIOManagerSwing.getInstance().readData(
						descr.getField(), descr.getMyclass(),
						this.componentMap.get(descr));
			}
			return obj;
		} catch (ALDDataIOException ex) {
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.UNSPECIFIED_ERROR,
					"[ALDParameterPanel::readParameter()] "
							+ "Reading parameter value failed!\n" + "--> "
							+ ex.getCommentString());
			// if (descr.isRequired()) {
			// Object[] options = { "OK" };
			// JOptionPane.showOptionDialog(
			// null,
			// "Reading parameter \"" + descr.getLabel()
			// + "\" failed!" + "Reason:\n"
			// + ex.getCommentString(), "Warning",
			// JOptionPane.DEFAULT_OPTION,
			// JOptionPane.WARNING_MESSAGE, null, options, options[0]);
			// return null;
			// }
		}
	}

	/**
	 * Function to enable or disable the panel.
	 * 
	 * @param b
	 *            If true, panel is enabled, otherwise components are disabled.
	 */
	public void setEnabled(boolean b) {
		Set<ALDParameterDescriptor> keys = this.labelMap.keySet();
		for (ALDParameterDescriptor descr : keys) {
			this.componentMap.get(descr).getJComponent().setEnabled(b);
		}
	}

	/**
	 * Set value of given parameter, i.e. pass the value to GUI component.
	 * 
	 * @param descr
	 *            Descriptor of parameter to be updated.
	 * @param value
	 *            New value of parameter to be displayed.
	 * @throws ALDDataIOException
	 */
	public void setParameter(ALDParameterDescriptor descr, Object value)
			throws ALDDataIOProviderException {

		// check if parameter is registered
		if (!this.componentMap.containsKey(descr)) {
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.UNSPECIFIED_ERROR,
					"[ALDParameterPanel::setParameter()] Descriptor not found!");
		}
		try {
			// ask the GUI component for updating its value
			ALDDataIOManagerSwing.getInstance().setValue(descr.getField(),
					descr.getMyclass(), this.componentMap.get(descr), value);
			this.mainPanel.updateUI();
		} catch (ALDDataIOException ex) {
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.UNSPECIFIED_ERROR,
					"[ALDParameterPanel::setParameter()] Setting parameter value failed!\n"
							+ "--> " + ex.getCommentString());
		}
	}

	/**
	 * Sets label color of parameter to red (e.g. if validation failed).
	 * 
	 * @param descr
	 *            Descriptor of parameter to mark.
	 */
	public void markParameter(ALDParameterDescriptor descr) {
		// get the value out of the parameter field
		if (!this.componentMap.containsKey(descr))
			return; // do nothing, if parameter unknown
		this.labelMap.get(descr).setForeground(java.awt.Color.RED);
	}

	/**
	 * Sets label color of parameter to black or of link to blue (e.g. if
	 * validation successful).
	 * 
	 * @param descr
	 *            Descriptor of parameter to unmark.
	 */
	public void unmarkParameter(ALDParameterDescriptor descr) {
		// get the value out of the parameter field
		if (!this.componentMap.containsKey(descr))
			return; // do nothing, if parameter unknown
		if (!descr.isLink())
			this.labelMap.get(descr).setForeground(java.awt.Color.BLACK);
		else
			this.labelMap.get(descr).setForeground(java.awt.Color.BLUE);
	}

	/**
	 * Sets label colors of all parameter to black and of all links to blue.
	 */
	public void unmarkAllParameters() {
		Set<ALDParameterDescriptor> keys = this.labelMap.keySet();
		for (ALDParameterDescriptor descr : keys) {
			if (!descr.isLink())
				this.labelMap.get(descr).setForeground(java.awt.Color.BLACK);
			else
				this.labelMap.get(descr).setForeground(java.awt.Color.BLUE);
		}
	}

	/**
	 * Builds up the input panel.
	 */
	private void buildInputPanel() {
		this.componentMap = new HashMap<ALDParameterDescriptor, ALDSwingComponent>();
		this.labelMap = new HashMap<ALDParameterDescriptor, JLabel>();
		this.panelMap = new HashMap<ALDParameterDescriptor, JPanel>();
		BoxLayout ylayout = new BoxLayout(this.mainPanel, BoxLayout.Y_AXIS);
		this.mainPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		this.mainPanel.setLayout(ylayout);

		// and setup GUI elements
		if (this.paramDescriptors.size() > 0) {
			if (this.titleString != null && !this.titleString.isEmpty()) {
				// add new panel for label
				JPanel titlePanel = new JPanel(new FlowLayout(
						FlowLayout.LEADING));
				// new title label
				JLabel titleLabel = new JLabel(" " + this.titleString + " ");
				// element separator before label
				JSeparator labelSep1 = new JSeparator(SwingConstants.HORIZONTAL);
				labelSep1.setPreferredSize(new Dimension(200, 1));
				// add element separator before label
				titlePanel.add(labelSep1);
				titlePanel.add(titleLabel);
				// element separator after label
				JSeparator labelSep2 = new JSeparator(SwingConstants.HORIZONTAL);
				labelSep2.setPreferredSize(new Dimension(200, 1));
				// add element separator after label
				titlePanel.add(labelSep2);
				this.mainPanel.add(titlePanel);
			}
			// sort descriptors according to GUI order into hash table
			HashMap<Integer, Vector<ALDParameterDescriptor>> guiOrderHash = new HashMap<Integer, Vector<ALDParameterDescriptor>>();
			for (ALDParameterDescriptor descr : this.paramDescriptors) {
				Integer order = new Integer(descr.getDataIOOrder());
				if (guiOrderHash.containsKey(order)) {
					guiOrderHash.get(order).add(descr);
				} else {
					Vector<ALDParameterDescriptor> paramVec = new Vector<ALDParameterDescriptor>();
					paramVec.add(descr);
					guiOrderHash.put(order, paramVec);
				}
			}
			// TODO Birgit - Maybe sorting the keys can be done easier...
			Set<Integer> keys = guiOrderHash.keySet();
			LinkedList<Integer> keyList = new LinkedList<Integer>();
			for (Integer key : keys) {
				keyList.add(key);
			}
			Collections.sort(keyList);
			for (Integer key : keyList) {
				Vector<ALDParameterDescriptor> descrips = guiOrderHash.get(key);
				for (ALDParameterDescriptor descr : descrips) {
					try {
						if (this.debug)
							System.out.println("ALDParameterPanel: "
									+ "requesting GUI element for parameter "
									+ descr.getName() + " of class "
									+ descr.getMyclass().getName());
						JPanel paramPanel = new JPanel(new FlowLayout(
								FlowLayout.LEADING));
						JLabel nameLabel = new JLabel(descr.getLabel());
						this.labelMap.put(descr, nameLabel);
						nameLabel.setToolTipText("<html>" + "Class: "
								+ descr.getMyclass().getSimpleName() + "<br>"
								+ descr.getExplanation() + "</html>");
						paramPanel.add(nameLabel);

						ALDSwingComponent aldElement = null;
						JComponent guiElement = null;
						if (!this.onlyDisplay) {
							try {
								aldElement = ALDDataIOManagerSwing
										.getInstance().createGUIElement(
												descr.getField(),
												descr.getMyclass(),
												descr.getDefaultValue());
								guiElement = aldElement.getJComponent();
							} catch (ALDDataIOManagerException exp) {
								// check if a provider was found
								ALDSwingComponentLabel warning = new ALDSwingComponentLabel(
										"ERROR: provider returned no element!");
								warning.getJComponent().setForeground(
										java.awt.Color.red);
								this.componentMap.put(descr, warning);
								paramPanel.add(warning.getJComponent());
								// add new element to panel
								this.mainPanel.add(paramPanel);
								// this.panelMap.put(descr, paramPanel);
							}
						} else {
							try {
								guiElement = ALDDataIOManagerSwing
										.getInstance().writeData(
												descr.getDefaultValue());
							} catch (ALDDataIOManagerException exp) {
								// check if a provider was found
								ALDSwingComponentLabel warning = new ALDSwingComponentLabel(
										"ERROR: provider returned no element!");
								warning.getJComponent().setForeground(
										java.awt.Color.red);
								this.componentMap.put(descr, warning);
								paramPanel.add(warning.getJComponent());
								// add new element to panel
								this.mainPanel.add(paramPanel);
								// this.panelMap.put(descr, paramPanel);
							}
						}
						if (aldElement != null && guiElement != null) {
							aldElement.addValueChangeEventListener(this);
							this.componentMap.put(descr, aldElement);
							paramPanel.add(guiElement);
							this.panelMap.put(descr, paramPanel);
							// add new element to panel
							this.mainPanel.add(paramPanel);
						}
					} catch (Exception e) {
						System.out.println(e.getMessage());
						e.printStackTrace();
					}
				}
			}
		}
	}

	/**
	 * Check if all GUI elements can deliver a proper (non-null) value.
	 * <p>
	 * Parameters delivering null as value are marked red or orange. Red color
	 * is used for panels used in top-level context, orange otherwise.
	 * 
	 * @return Null, if successful, otherwise list of failing parameters.
	 */
	public HashMap<ALDParameterDescriptor, String> validateParameters() {

		// first reset all labels to black or blue (from probable previous
		// validations)
		for (Map.Entry<ALDParameterDescriptor, JLabel> entry : this.labelMap
				.entrySet()) {
			JLabel label = entry.getValue();
			if (!entry.getKey().isLink())
				label.setForeground(java.awt.Color.BLACK);
			else
				label.setForeground(java.awt.Color.BLUE);
		}

		HashMap<ALDParameterDescriptor, String> failedList = new HashMap<ALDParameterDescriptor, String>();
		Object readValue = null;
		for (Map.Entry<ALDParameterDescriptor, ALDSwingComponent> entry : this.componentMap
				.entrySet()) {
			readValue = null;
			ALDParameterDescriptor descr = entry.getKey();
			ALDSwingComponent guiElement = entry.getValue();
			if (descr.isRequired()) {
				// skip output parameters of operators
				if (descr instanceof ALDOpParameterDescriptor) {
					if (((ALDOpParameterDescriptor) descr).getDirection() == Parameter.Direction.OUT)
						continue;
				}
				try {
					// TODO
					if (descr.isLink()) {
						String linkSource = (String) ALDDataIOManagerSwing
								.getInstance().readData(descr.getField(),
										String.class, guiElement);
						if (linkSource != null && !linkSource.isEmpty()
								&& linkSource.contains("@")) {
							if (!this.editorFrame.validateLink(linkSource, link
									+ "@" + descr.getName())) {
								failedList.put(descr, "Wrong link received!");
								this.labelMap.get(descr).setForeground(
										java.awt.Color.RED);
							}
						}
					}
					else {
						// readValue = ALDDataIOManagerSwing.getInstance()
						// .readData(descr.getField(), descr.getMyclass(),
						// guiElement);
						// if (readValue == null && !this.nullReturnsForbidden)
						// this.labelMap.get(descr).setForeground(
						// java.awt.Color.ORANGE);
						// else if (readValue == null)
						// this.labelMap.get(descr).setForeground(
						// java.awt.Color.RED);
						readValue = ALDDataIOManagerSwing.getInstance()
								.readData(descr.getField(), descr.getMyclass(),
										guiElement);
						if (readValue == null && !this.nullReturnsForbidden)
							this.labelMap.get(descr).setForeground(
									java.awt.Color.ORANGE);
						else if (readValue == null) {
							failedList.put(descr, "Null value received!");
							this.labelMap.get(descr).setForeground(
									java.awt.Color.RED);
						}
					}
				} catch (ALDDataIOException e) {
					failedList.put(descr, e.getCommentString());
					this.labelMap.get(descr).setForeground(java.awt.Color.RED);
				}
			}
		}
		if (failedList.size() > 0)
			return failedList;
		return null;
	}

	@Override
	public void handleValueChangeEvent(ALDSwingValueChangeEvent event) {
		this.fireALDSwingValueChangeEvent(event);
	}

	/**
	 * Replace the panel of given parameter to insert a link.
	 * 
	 * @param descr
	 *            Descriptor of parameter.
	 * @param value
	 *            Default value of the new panel.
	 * @param editorFrame
	 */
	public void changePanel(ALDParameterDescriptor descr, Object value) {
		if (this.panelMap.containsKey(descr)) {

			this.panelMap.get(descr).remove(
					this.componentMap.get(descr).getJComponent());
			this.componentMap.remove(descr);

			ALDSwingComponent elem = null;
			JComponent guiElement = null;
			try {
				if (descr.isLink()) {
					elem = ALDDataIOManagerSwing.getInstance()
							.createGUIElement(descr.getField(), String.class,
									null);
					guiElement = elem.getJComponent();
					// elem.addValueChangeEventListener(this);
					guiElement.addKeyListener(new LinkTextBox(descr));
					/*
					 * ALDDataIOManagerSwing.getInstance().setValue(
					 * descr.getField(), String.class, elem, value);
					 */
					this.labelMap.get(descr).setForeground(java.awt.Color.BLUE);
				} else {
					elem = ALDDataIOManagerSwing.getInstance()
							.createGUIElement(descr.getField(),
									descr.getMyclass(), value);
					guiElement = elem.getJComponent();
					// elem.addValueChangeEventListener(this);
					/*
					 * ALDDataIOManagerSwing.getInstance().setValue(
					 * descr.getField(), descr.getMyclass(), elem,
					 * descr.getDefaultValue());
					 */
					this.labelMap.get(descr)
							.setForeground(java.awt.Color.BLACK);
				}
				if (elem != null && guiElement != null) {
					elem.addValueChangeEventListener(this);
					this.componentMap.put(descr, elem);
					this.panelMap.get(descr).add(guiElement);
					this.panelMap.get(descr).updateUI();
					if (descr.isLink())
						ALDDataIOManagerSwing.getInstance().setValue(
								descr.getField(), String.class, elem, value);
					else
						ALDDataIOManagerSwing.getInstance().setValue(
								descr.getField(), descr.getMyclass(), elem,
								descr.getDefaultValue());
				}
			} catch (ALDDataIOException e) {
				e.printStackTrace();
			}
		}
	}

	// TODO comments
	public void copyConfiguration(ALDParameterPanel parameterPanel)
			throws ALDDataIOException {
		for (ALDParameterDescriptor descr : this.componentMap.keySet()) {
			if (!descr.isLink()) {
				Object value = null;
				value = ALDDataIOManagerSwing.getInstance().readData(
						descr.getField(), descr.getMyclass(),
						this.componentMap.get(descr));

				for (ALDParameterDescriptor descr2 : parameterPanel.componentMap
						.keySet())
					if (descr2.getName().equals(descr.getName()))
						ALDDataIOManagerSwing.getInstance().setValue(
								descr2.getField(), descr2.getMyclass(),
								parameterPanel.componentMap.get(descr2), value);
			}
		}
	}

	/**
	 * KeyListener for the text box of links.
	 */
	class LinkTextBox implements KeyListener {
		private ALDParameterDescriptor descr;

		public LinkTextBox(ALDParameterDescriptor descr) {
			this.descr = descr;
		}

		@Override
		public void keyPressed(KeyEvent e) {
			if (e.getKeyChar() == KeyEvent.VK_ENTER) {
				try {
					String scr = (String) ALDDataIOManagerSwing.getInstance()
							.readData(descr.getField(), String.class,
									componentMap.get(descr));
					if (scr.contains("@")) {
						if (editorFrame.validateLink(scr,
								link + "@" + descr.getName())) {
							editorFrame.drawEdge(scr,
									link + "@" + descr.getName());
						}
					} else
						JOptionPane.showMessageDialog(null, "Error!",
								"Warning", JOptionPane.OK_CANCEL_OPTION);
				} catch (ALDDataIOException e1) {
					e1.printStackTrace();
				}
			}
		}

		@Override
		public void keyReleased(KeyEvent e) {

		}

		@Override
		public void keyTyped(KeyEvent e) {

		}
	}
}
