/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5601 $
 * $Date: 2012-05-22 14:12:02 +0200 (Di, 22. Mai 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.dataio.provider.helpers;

import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import javax.swing.table.DefaultTableModel;




/**
 * GUI window for displaying tables in Alida. 
 * 
 * @author moeller
 */

public class ALDTableWindow extends JFrame implements ActionListener {

	/**
	 * Data table (swing tables are nicer than imageJ tables).
	 */
	private JTable dataTab;
	
	/**
	 * Reference object to the data.
	 */
	private DefaultTableModel dataTabModel;
	
	/**
	 * For convenience: always open last directory for saving.
	 */
	private File lastDir;

	/**
	 * Flag to indicate if headers should be saved to file.
	 */
	private boolean saveHeaders = true;
	
	/**
	 * Default constructor.
	 * @param mtm	Associated table model.
	 */
	public ALDTableWindow(DefaultTableModel mtm) {
		this.dataTabModel= mtm;

		// init current directory with user directory
		this.lastDir= new File(System.getProperty("user.dir"));

		this.setupResultTable();
	}
	
	/**
	 * Shows the table window in graphical environment.
	 */
	public void openWindow() {
		this.setVisible(true);
		this.validate();
	}

	/**
	 * Closes the GUI table window.
	 * <p>
	 * Attention, if you do not store a reference to the window,
	 * it cannot be opened again!
	 */
	public void closeWindow() {
		this.setVisible(false);
	}
	
	/**
	 * Function to enable/disable saving of headers to file.
	 * By default headers are saved also.
	 * @param b	If false, headers are ignored on saving data to file.
	 */
	public void setSaveHeaders(boolean b) {
		this.saveHeaders = b;
	}
	
	/**
	 * Saves the contents of the table to a user-specified file.
	 * <p>
	 * The file format is TSV, i.e. tabulator-separated values.
	 * The default ending is '.txt'. The user can select the file
	 * name through a file open dialog which pops-up on call of the function.
	 */
	public void saveTable() {

		File file= null;
		
		JFrame dummy= new JFrame();
		JFileChooser getSaveFileDialog= new JFileChooser(); 
		getSaveFileDialog.setFileFilter(new DataTabFileFilter());
		getSaveFileDialog.setFileSelectionMode(JFileChooser.FILES_ONLY);
		getSaveFileDialog.setCurrentDirectory(this.lastDir); 
		getSaveFileDialog.setSelectedFile(new File("ResultTab.txt"));
		getSaveFileDialog.setApproveButtonText("Speichern");
		getSaveFileDialog.setDialogTitle("Select file name");
		int returnVal = getSaveFileDialog.showOpenDialog(dummy);
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            file = getSaveFileDialog.getSelectedFile();
            this.lastDir= getSaveFileDialog.getCurrentDirectory();
        }
        
        // obviously we did not get any name, can't do anything...
		if (file== null)
			return;
		
		// ... otherwise write table to selected file
		try{
			if (file.exists()) {
				int ret= 
					JOptionPane.showConfirmDialog(null,"The file exists already, would "
					+ "you like to overwrite it?", "Warning",JOptionPane.YES_NO_OPTION);
				switch(ret)
				{
				case JOptionPane.NO_OPTION:
				case JOptionPane.CANCEL_OPTION:
					return;
				case JOptionPane.YES_OPTION:
					// nothing special to do, just proceed
				}
			}
			FileWriter ow= new FileWriter(file.getPath());
			StringBuffer [] tab= tableToString(this.dataTabModel);
			// note: first row contains the header which might be meaningless, 
			//       i.e. sometimes they should be ignored
			int firstIndex = (this.saveHeaders ? 0 : 1);
			for (int i=firstIndex;i<tab.length;++i)
				ow.write(tab[i].toString());
			ow.close();
		} catch (IOException e) {
			JOptionPane.showMessageDialog(null,"Error!!! " +
					"Could not open output file " + file.getPath() + "!");
		}
	}


	/* (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	@Override
  public void actionPerformed(ActionEvent e) {
		
		// deletes all data from the table and sets row and column counts to zero
		if (e.getActionCommand()=="clear") {
			for (int i=0;i<this.dataTabModel.getRowCount();++i)
				for (int j=0;j<this.dataTabModel.getColumnCount();++j)
					this.dataTabModel.setValueAt(null, i, j);
		}
		// closes the GUI window
		if (e.getActionCommand()=="close" ) {
			this.setVisible(false);
		}
		// saves the table data in TSV format
		if (e.getActionCommand()=="save") {
			this.saveTable();
		}		
	}
	
	/**
	 * Initializes the data table window.
	 */
	private void setupResultTable() {
		
		// panel for the table
		JPanel tabPanel;

		// instantiate result table and put into scroll pane
		this.dataTab= new JTable(this.dataTabModel);
		JScrollPane scrollpane= new JScrollPane(this.dataTab);
		scrollpane.setPreferredSize(new Dimension (400,300));
		this.dataTab.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

		// add a toolbar
		JButton tabSave= new JButton("Save");
		tabSave.setActionCommand("save");
		tabSave.addActionListener(this);
		tabSave.setSize(new Dimension(80,35));
		JButton tabClear= new JButton("Clear");
		tabClear.setActionCommand("clear");
		tabClear.addActionListener(this);
		tabClear.setSize(new Dimension(80,35));
		JButton tabClose= new JButton("Close");
		tabClose.setActionCommand("close");
		tabClose.addActionListener(this);
		tabClose.setSize(new Dimension(80,35));
		JPanel tabToolPanel= new JPanel();
		tabToolPanel.setLayout(new FlowLayout());
		tabToolPanel.add(tabSave);
		tabToolPanel.add(tabClear);
		tabToolPanel.add(tabClose);
		JToolBar tabTools= new JToolBar("");
		tabTools.add(tabToolPanel);
		
		tabPanel= new JPanel();
		tabPanel.setLayout(new BorderLayout());
		tabPanel.add(scrollpane, BorderLayout.CENTER);
		tabPanel.add(tabTools, BorderLayout.SOUTH);

		this.setSize(900, 250);
		this.setTitle("");
		this.add(tabPanel);
		
		// ... then show it!
		this.setVisible(false);
	}
	
	/**
	 * Converts the contents of the table to a string array
	 * in CSV format (suitable for import in Excel).
	 * 
	 * @return array with contents of table (arranged line-wise)
	 */
	public static StringBuffer [] tableToString(DefaultTableModel model) {
		
		// number of lines in table (without headlines)
		int lines= model.getRowCount();
		int cols = model.getColumnCount();

		StringBuffer [] tabS= new StringBuffer[lines+1];
		StringBuffer headings= new StringBuffer();
		StringBuffer newline;
		
		// transform headlines...
		for (int j=0; j<cols; ++j) {
			headings.append(model.getColumnName(j) + "\t");
		}
		headings.append("\n");
		tabS[0]= headings;
		
		// ... and then the rows
		String buf= null;
		for (int i= 0; i<lines; ++i) {

			newline= new StringBuffer();
			newline.append(model.getValueAt(i, 0)+"\t");
			for (int j=1; j<model.getColumnCount()-1;++j) {

				if (model.getValueAt(i,j) == null) {
					newline.append(0 + "\t");
					continue;
				}
				buf= (model.getValueAt(i,j)).toString();
				// append to line
				newline.append(buf + "\t");
			}
			if (   model.getColumnCount() > 1 
					&& model.getValueAt(i, model.getColumnCount()-1) != null) {
				
				buf= (model.getValueAt(i, model.getColumnCount()-1)).toString();

				newline.append(buf + "\n");
			}
			else {
				newline.append("\n");
			}
			tabS[i+1]= newline;
		}
		return tabS;
	}

	/**
	 * Internal class that realizes a FileFilter for text
	 * files where MiToBo table data is stored.
	 *	
	 * @author moeller
	 */
	protected class DataTabFileFilter extends FileFilter {
		
		/* (non-Javadoc)
		 * @see javax.swing.filechooser.FileFilter#accept(java.io.File)
		 */
		@Override
		public boolean accept(File f) {
			return (f.getName().endsWith(".txt") || f.isDirectory());
		}
		
		/* (non-Javadoc)
		 * @see javax.swing.filechooser.FileFilter#getDescription()
		 */
		@Override
		public String getDescription() {
			return "MiToBo Data Table Files (*.txt)";
		}
	}
}
