/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5825 $
 * $Date: 2012-07-24 09:35:34 +0200 (Di, 24. Jul 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.dataio.provider.swing;



import de.unihalle.informatik.Alida.annotations.ALDDataIOProvider;
import de.unihalle.informatik.Alida.dataio.provider.swing.components.ALDSwingComponent;
import de.unihalle.informatik.Alida.dataio.provider.swing.components.ALDSwingComponentTextField;
import de.unihalle.informatik.Alida.dataio.provider.swing.events.ALDSwingValueChangeEvent;
import de.unihalle.informatik.Alida.dataio.provider.swing.events.ALDSwingValueChangeListener;
import de.unihalle.informatik.Alida.datatypes.ALDDirectoryString;
import de.unihalle.informatik.Alida.datatypes.ALDFileString;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException.ALDDataIOProviderExceptionType;
import de.unihalle.informatik.Alida.helpers.ALDFilePathManipulator;

import javax.swing.*;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.lang.reflect.Field;
import java.util.*;

/**
 * Class for loading/saving file and directory paths via GUI.
 * <p>
 * This class provides a panel for a GUI which contains a text entry field
 * for the path, and in addition a button that opens a dialog for browsing the
 * file system. By this it allows to select files and directories in a 
 * comfortable manner. 
 * 
 * @author moeller
 *
 */
@ALDDataIOProvider

public class ALDFileDirectoryDataIOSwing implements ALDDataIOSwing {

	/**
	 * Interface method to announce class for which IO is provided for.
	 * 
	 * @return  Collection of classes provided.
	 */
	@Override
	public Collection<Class<?>> providedClasses() {
		LinkedList<Class<?>> classes = new LinkedList<Class<?>>();
		classes.add(ALDFileString.class);
		classes.add(ALDDirectoryString.class);
		return classes;
	}
	
	/* (non-Javadoc)
	 * @see de.unihalle.informatik.Alida.helpers.ALDDataIO#readData(java.lang.String)
	 */
	@Override
	public ALDSwingComponent createGUIElement(
																			Field field, Class<?> cl, Object obj) {
		return new FileDirectoryDataIOPanel(field, cl, obj);
	}

	@Override
  public void setValue(
  		Field field, Class<?> cl, ALDSwingComponent guiElement, Object value) 
  	throws ALDDataIOProviderException {
		if (   !(value instanceof ALDDirectoryString)
				&& !(value instanceof ALDFileString))
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.INVALID_GUI_ELEMENT, 
					"ALDFileDirDataIOSwing: setValue() received wrong object type!");
		if (!(guiElement instanceof FileDirectoryDataIOPanel))
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.INVALID_GUI_ELEMENT, 
					"ALDFileDirDataIOSwing: setValue() received invalid GUI element!");
		((FileDirectoryDataIOPanel)guiElement).setValue(field,cl,value);
  }

	/* (non-Javadoc)
	 * @see de.unihalle.informatik.Alida.dataio.provider.swing.ALDDataIOSwing#readData(java.lang.reflect.Field, java.lang.Class, javax.swing.JComponent)
	 */
	@Override
	public Object readData(
			Field field, Class<?> cl, ALDSwingComponent guiElement)
		throws ALDDataIOProviderException {
		if (!(guiElement instanceof FileDirectoryDataIOPanel))
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.INVALID_GUI_ELEMENT, 
					"ALDFileDirDataIOSwing: readData received invalid GUI element!");
		return ((FileDirectoryDataIOPanel)guiElement).readData(field,cl);
	}

	/* (non-Javadoc)
	 * @see de.unihalle.informatik.Alida.dataio.provider.swing.ALDDataIOSwing#writeData(java.lang.Object)
	 */
	@Override
	public JComponent writeData(Object obj) throws ALDDataIOProviderException {
		if (   !(obj instanceof ALDFileString) 
				&& !(obj instanceof ALDDirectoryString)) {
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.OBJECT_TYPE_ERROR,
					"ALDFileDirDataIOSwing: writeData received object of wrong type!");
		}
		if (obj instanceof ALDFileString)
			return new JTextField(((ALDFileString)obj).getFileName());
		return new JTextField(((ALDDirectoryString)obj).getDirectoryName());
	}

	/**
	 * GUI panel for IO of file and directory names.
	 * 
	 * @author moeller
	 */
	protected class FileDirectoryDataIOPanel extends ALDSwingComponent 
		implements ActionListener, ALDSwingValueChangeListener {

		/**
		 * Swing component to be integrated in GUI.
		 */
		private JPanel ioPanel = null;
		
		/**
		 * Last directory selected by user.
		 */
		private File lastDirectory;
		
		/**
		 * Last file selected by user.
		 */
		private File lastFile;
		
		/**
		 * Flag indicating if panel refers to a directory parameter or not.
		 */
		private boolean isDirectory = false;
		
		/**
		 * Text field in the panel.
		 */
		private ALDSwingComponentTextField textField = null;
		
		/**
		 * Default constructor.
		 * 
		 * @param field	Field to consider.
		 * @param cl		Class to consider.
		 * @param obj		Default object.
		 */
		protected FileDirectoryDataIOPanel(Field field, Class<?> cl, Object obj) {
			this.ioPanel = new JPanel();
			
			if (cl.equals(ALDDirectoryString.class)) {
				this.isDirectory = true;
			}
			
			this.ioPanel.setLayout(new GridLayout(1,2));
			
			this.textField = 
					new ALDSwingComponentTextField(String.class, 20);
			this.textField.setText("/tmp");
			this.textField.getJComponent().setSize(70,5);
			this.textField.addValueChangeEventListener(this);
			String defValue = System.getProperty("user.dir");
			if (obj != null) {
				if (obj instanceof ALDFileString) {
					defValue = ((ALDFileString)obj).getFileName();
				}
				else {
					defValue = ((ALDDirectoryString)obj).getDirectoryName();
				}
			}
			this.lastDirectory = new File(defValue);
			this.textField.setText(defValue);
			this.ioPanel.add(this.textField.getJComponent());
			JButton selectFileDir= new JButton("Choose...");
			selectFileDir.setActionCommand("choose");
			selectFileDir.addActionListener(this);
			this.ioPanel.add(selectFileDir);
		}
		
		@Override
		public JPanel getJComponent() {
			return this.ioPanel;
		}
		
		/**
		 * Returns the contents of the text field.
		 * 
		 * @param field	Field to consider.
		 * @param cl		Class to consider.
		 * @param value	Object value to be set.
		 * @throws ALDDataIOProviderException 
		 */
		public void setValue(Field field, Class<?> cl, Object value) 
			throws ALDDataIOProviderException {
			if (cl.equals(ALDFileString.class)) {
				this.textField.setText(((ALDFileString)value).getFileName());
				this.lastDirectory = 
					new File(((ALDFileString)value).getFileName());
			}
			else {
				this.textField.setText(((ALDDirectoryString)value).getDirectoryName());
				this.lastDirectory = 
					new File(((ALDDirectoryString)value).getDirectoryName());
			}
		}

		/**
		 * Returns the contents of the text field.
		 * 
		 * @param field	Field to consider.
		 * @param cl		Class to consider.
		 * @return	Read object.
		 * @throws ALDDataIOProviderException 
		 */
		public Object readData(Field field, Class<?> cl) 
			throws ALDDataIOProviderException {
			if (cl.equals(ALDFileString.class))
				return new ALDFileString(this.textField.getText());
			if (cl.equals(ALDDirectoryString.class))
				return new ALDDirectoryString(this.textField.getText());
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.OBJECT_TYPE_ERROR,
					"ALDFileDirectoryDataIOSwing: unknown class requested!");
		}

		@Override
	  public void actionPerformed(ActionEvent e) {
			String command = e.getActionCommand();
			if (command.equals("choose")) {
				// open file chooser
				JFileChooser getDirDialog= new JFileChooser();
				if (    this.textField.getText() != null 
						&& !this.textField.getText().isEmpty()) {
					String selection = this.textField.getText();
					String file = 
						ALDFilePathManipulator.removeLeadingDirectories(selection);
					String dir = ALDFilePathManipulator.getPath(selection);
					if (!file.isEmpty())
						getDirDialog.setSelectedFile(new File(file));
					if (!dir.isEmpty())
						getDirDialog.setCurrentDirectory(new File(dir));
					else {
						if (selection.substring(0,1).equals(
								System.getProperty("file.separator"))) {
							getDirDialog.setCurrentDirectory(new File(selection));
						}
					}
				}
				else {
					getDirDialog.setCurrentDirectory(this.lastDirectory);
					getDirDialog.setSelectedFile(this.lastFile);
				}
				if (this.isDirectory) {
					getDirDialog.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
					getDirDialog.setDialogTitle("Select a directory...");
				}
				else
					getDirDialog.setDialogTitle("Select a file...");		
				
				int returnVal = getDirDialog.showOpenDialog(null);
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					File file = getDirDialog.getSelectedFile();
					this.textField.setText(file.getPath());
					this.lastDirectory = getDirDialog.getCurrentDirectory();
					this.lastFile = getDirDialog.getSelectedFile();
				}
			}
	  }

		@Override
		public void handleValueChangeEvent(ALDSwingValueChangeEvent event) {
			this.fireALDSwingValueChangeEvent(new ALDSwingValueChangeEvent(this));
		}
	}
}
