/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 6088 $
 * $Date: 2012-10-10 15:31:20 +0200 (Mi, 10. Okt 2012) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.Alida.gui;

import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.event.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.util.*;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;



import de.unihalle.informatik.Alida.annotations.*;
import de.unihalle.informatik.Alida.annotations.ALDAOperator.Level;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.gui.ALDOperatorChooserTreeNode;
import de.unihalle.informatik.Alida.helpers.ALDClassInfo;

/**
 * Main window for selecting Alida annotated operators for running.
 * For a selected operator a configuration window may be created by double clicking or
 * hitting the configure button.
 * <p>
 * This class holds two different trees of operators, one for the standard level, one for
 * the application level.
 * These trees contain all <code>ALDAOperator</code> annotated operators allowed to be executed in a GUI,
 * and annotated as <code>Level.STANDARD</code> and <code>Level.APPLICATION</code>, respectively.
 * Deriving classes may override the method <code>additionalOperators</code> to extend the trees.
 * <p>
 * In deriving classes the method <code>getOpControlFrame</code> may be overridden to define a different
 * frame to handle the configuration of the selected operator. The frame created in by 
 * <code>ALDChooseOpNameFrame</code> is
 * {@link ALDOperatorControlFrame}.
 *   
 * @author Stefan Posch
 * @author Birgit Moeller
 */

public class ALDChooseOpNameFrame extends JFrame 
	implements  ActionListener, TreeSelectionListener {

	/**
	 * Debug flag (not accessible from outside).
	 */
	private boolean debug = false;
	
	/**
	 * Path of selected operator.
	 */
	protected String opFullname = null;

	/**
	 * Name of selected operator.
	 */
	protected String opName = null;
	
	/**
	 * JComponent containing both trees of available operators.
	 */
	protected ALDOperatorChooserTree opTree;

	/**
	 * Main panel of chooser window.
	 */
	protected JRootPane mainPanel;
	
	/**
	 * Label to display selected operator.
	 */
	protected JLabel opNameJText;
	
	/**
	 * Scroll pane displaying the operator tree.
	 */
	protected JScrollPane opTreePane;

	/**
	 * Icon to be shown in about box.
	 */
	protected ImageIcon aboutIcon;
	
	/**
	 * Title of frame.
	 */
	protected String titleString = 
		"Alida - OpRunner: simply choose an operator...";
	
	/**extendOpNames
	 * Constructor.
	 * 
	 * @param selectAction action to be performed upon selection of an operator
	 */
	public ALDChooseOpNameFrame() {
		if ( debug )
			System.out.println( " ALDChooseOpNameFrame instantiate");
		
		// instantiate operator trees
		Collection<String> standardOps = ALDClassInfo.lookupOperators(ALDAOperator.Level.STANDARD, ALDAOperator.ExecutionMode.SWING);
		standardOps.addAll( additionalOperators( ALDAOperator.Level.STANDARD));
		Collection<String> allicationOps = ALDClassInfo.lookupOperators(ALDAOperator.Level.APPLICATION, ALDAOperator.ExecutionMode.SWING);
		allicationOps.addAll( additionalOperators( ALDAOperator.Level.APPLICATION));

		this.opTree = new ALDOperatorChooserTree( standardOps, allicationOps);
        MouseListener ml = new MyMouseAdapter(this);
        this.opTree.addMouseListener(ml);
        this.opTree.addTreeSelectionListener(this);
        
		// global language settings
		Locale.setDefault(Locale.ENGLISH);
		JComponent.setDefaultLocale(Locale.ENGLISH);

		// ------------------------------------------------------------
		// setup icon
		String iconDataName = "/share/logo/Alida_logo.png";
		Image img = null;
		BufferedImage bi = null;
		Graphics g = null;
		InputStream is = null;
		try {
			ImageIcon icon;
			File iconDataFile = new File("./" + iconDataName);
			if(iconDataFile.exists()) {
				icon = new ImageIcon("./" + iconDataName);
				img = icon.getImage();
			}
			// try to find it inside a jar archive....
			else {
				is = ALDChooseOpNameFrame.class.getResourceAsStream(iconDataName);
				if (is == null) {
					System.err.println("Warning - cannot find icons...");
					img = new BufferedImage(20,20,BufferedImage.TYPE_INT_ARGB);
				}
				else {
					img = ImageIO.read(is);
				}
				bi= new BufferedImage(20,20,BufferedImage.TYPE_INT_ARGB);
				g = bi.createGraphics();
				g.drawImage(img, 0, 0, 20, 20, null);
			}
		} catch (IOException ex) {
			System.err.println("ALDChooseOpNameFrame - problems loading icons...!");
			img = new BufferedImage(20,20,BufferedImage.TYPE_INT_ARGB);
			bi= new BufferedImage(20,20,BufferedImage.TYPE_INT_ARGB);
			g = bi.createGraphics();
			g.drawImage(img, 0, 0, 20, 20, null);
		}
		this.aboutIcon = new ImageIcon(img);
		
		// ------------------------------------------------------------
		// set up main window
		this.mainPanel = new JRootPane();
		BoxLayout ylayout = new BoxLayout( this.mainPanel, BoxLayout.Y_AXIS);
		this.mainPanel.setLayout( ylayout);

		this.opTreePane = new JScrollPane( this.opTree);
		this.mainPanel.add(this.opTreePane,0);
		
		// do some global settings
		this.setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);

		// TODO Does this anything? -> if not, remove...
		this.mainPanel.setBorder(BorderFactory.createEmptyBorder(10,10,10,10));

		// add panel for displaying selected operator
		JLabel lab =  new JLabel("Selected Operator: ");
		this.opNameJText = new JLabel( "- none -");
		JPanel selelectePanel = new JPanel(); 
		selelectePanel.add( lab);
		selelectePanel.add(Box.createRigidArea(new Dimension(10, 0)));
		selelectePanel.add( this.opNameJText);
		this.mainPanel.add( selelectePanel);

		// add buttons
		JButton configButton = new JButton("Configure Operator...");
		configButton.setActionCommand( "config");
		configButton.addActionListener(this);
		configButton.setBounds(50, 60, 80, 30);
		configButton.setAlignmentX(Component.LEFT_ALIGNMENT);

		JButton quitButton = new JButton("Quit");
		quitButton.setActionCommand( "quit");
		quitButton.addActionListener(this);
		quitButton.setBounds(50, 60, 80, 30);
		quitButton.setAlignmentX(Component.LEFT_ALIGNMENT);

		JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout(new BoxLayout(buttonPanel, BoxLayout.LINE_AXIS));
		buttonPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
		buttonPanel.add(Box.createHorizontalGlue());
		buttonPanel.add(configButton);
		buttonPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		buttonPanel.add(quitButton);

		this.mainPanel.add(buttonPanel);

		// add a nice menu bar
		JMenuBar mainWindowMenu = new JMenuBar();
		JMenu fileM = new JMenu("File");
		JMenuItem quitItem = new JMenuItem("Quit");
		quitItem.setActionCommand("quit");
		quitItem.addActionListener(this);
		fileM.add(quitItem);
		JMenu opSelectM = new JMenu("Operator Level");
		ButtonGroup operatorLevelGroup = new ButtonGroup();
		// default
		JRadioButtonMenuItem radioItemApplication =
			new JRadioButtonMenuItem("Application");
		radioItemApplication.setActionCommand("viewApps");
		radioItemApplication.addActionListener(this);
		JRadioButtonMenuItem radioItemStandard = 
			new JRadioButtonMenuItem("Standard");
		radioItemStandard.setActionCommand("viewStd");
		radioItemStandard.addActionListener(this);
		operatorLevelGroup.add(radioItemApplication);
		operatorLevelGroup.add(radioItemStandard);
		opSelectM.add(radioItemApplication);
		opSelectM.add(radioItemStandard);

		if ( this.opTree.getLevel() == Level.APPLICATION)
			radioItemApplication.setSelected(true);
		else
			radioItemStandard.setSelected(true);

		JMenu helpM = this.generateHelpMenu();
		mainWindowMenu.add(fileM);
		mainWindowMenu.add(opSelectM);
		mainWindowMenu.add(Box.createHorizontalGlue());
		mainWindowMenu.add(helpM);

		// ... and go setup the frame
		this.setTitle(this.titleString);
		this.getContentPane().setPreferredSize(new Dimension(500, 700));
		this.pack();
		this.setJMenuBar(mainWindowMenu);
		this.add(this.mainPanel);
		this.setDefaultCloseOperation( DISPOSE_ON_CLOSE);
		this.setVisible(true);
	}
	
    /** This method may be overridden in extending classes to add operators to the
     * list of available operators.
     * 
     * @param level Level to which operators are to be added
     */
	protected Collection<String> additionalOperators( ALDAOperator.Level level) {
		return new LinkedList<String>();
	}
	
    /**
     * Method to get a control frame for the chosen operator.
     * <p>
     * Method is supposed to be overwritten for instantiating specific control
     * frames, i.e. for handling IJ 2.0 plugins by wrappers.
     *  
     * @param name  Name of the operator.
     * @return  Corresponding control frame.
     */
    protected ALDOperatorControlFrame getOpControlFrame(String name) {
    	ALDOperatorControlFrame frame = null;
			try {
				frame = new ALDOperatorControlFrame(name);
			} catch (ALDOperatorException e) {
				Object[] okOption = { "OK" };
				JOptionPane.showOptionDialog(
						null,
						"Could not instante chosen operator \"" + name + "\"\n" +
					  "Something seems to be wrong with its class implementation...",
					  "Warning", JOptionPane.DEFAULT_OPTION,
					  JOptionPane.WARNING_MESSAGE, null, okOption, okOption[0]);
			}
    	return frame;
    }

	
	/**
	 * Sets up the help menu.
	 * 
	 * @return Generated help menu.
	 */
	protected JMenu generateHelpMenu() {
		JMenu helpM = new JMenu("Help");
		JMenuItem itemHelp = new JMenuItem("Online Help");
		JMenuItem itemAbout = new JMenuItem("About Alida");
		itemHelp.addActionListener(
				OnlineHelpDisplayer.getHelpActionListener(itemHelp,"welcome"));
		itemAbout.setActionCommand("showAbout");
		itemAbout.addActionListener(this);
		helpM.add(itemHelp);
		helpM.add(itemAbout);
		return helpM;
	}	
	
	/**
	 * Show an about box window.
	 * <p>
	 * Method is supposed to be overwritten by subclasses.
	 */
	protected void showAboutBox() {
		Object[] options = { "OK" };
		String msg = "<html>Alida - Advanced Library for Integrated Development<p>"
		   + "\t of Data Analysis Applications, Release 1.2<p><p>"
		   + "\u00a9 2010 - 2012   "
		   + "Martin Luther University Halle-Wittenberg<p><p>"
		   + "Institute of Computer Science, Faculty of Natural Sciences III<p><p>"
		   + "Email: alida@informatik.uni-halle.de <p>"
		   + "Internet: <i>www.informatik.uni-halle.de/alida</i><p>"
		   + "License: GPL 3.0, <i>http://www.gnu.org/licenses/gpl.html</i></html>";

		JOptionPane.showOptionDialog(null, new JLabel(msg),
		    "Information about Alida", JOptionPane.DEFAULT_OPTION,
		    JOptionPane.INFORMATION_MESSAGE, this.aboutIcon, options, options[0]);
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	@Override
	public final void actionPerformed(ActionEvent e) {

		// local variables
		String command = e.getActionCommand();

		// configure button has been pressed
		if ( command.equals("config") ) {
			// do we have an operator name?
			if (this.opFullname != null ) {
				ALDOperatorControlFrame controlFrame =
						this.getOpControlFrame( ALDChooseOpNameFrame.this.opFullname);
				// if frame is null, a problem occurred...
				if (controlFrame != null)
					controlFrame.setVisible(true);
			}
		}

		// quit button has been pressed
		else if ( command.equals("quit") ) {
			this.setVisible( false);
			// stop the current thread, note that control windows of operators and
			// other graphical elements remain unaffected
			Thread.currentThread().interrupt();
		}

		/* handle tree view options */
		else if (command.equals("viewApps")) {
			opTree.setLevel( ALDAOperator.Level.APPLICATION);
		}
		else if (command.equals("viewStd")) {
			opTree.setLevel( ALDAOperator.Level.STANDARD);
		}
		
		/* display about box */
		else if (command.equals("showAbout")) {
			this.showAboutBox();
		}
	}

	// maybe this functionality can be moved to the mouse adapter <code>MyMouseAdapter</code>

	@Override
	public void valueChanged(TreeSelectionEvent event) {
		ALDOperatorChooserTreeNode node = (ALDOperatorChooserTreeNode) this.opTree.getLastSelectedPathComponent();
		// if sub-tree has been folded in, last selected node is null -			
		// do nothing in that case
		if (node == null)
			return;
		if (node.isOperator()) {
			this.opFullname = new String(node.getFullName());
			this.opName = this.opTree.getLastSelectedPathComponent().toString();

            this.opNameJText.setText(
                            this.opTree.getLastSelectedPathComponent().toString());
            this.opNameJText.setToolTipText( this.opFullname);

		} else {
			this.opFullname = null;
			this.opName = null;
		}
	}

    /**
     * Local MouseAdapter class. Handles double click on operator to invoke the action.
     * 
     * @author posch
     */
    class MyMouseAdapter extends MouseAdapter {

        /**
         * chooser frame this mouse adapter is associated with.
         */
    	ALDChooseOpNameFrame chooseOpNameFrame;

        /**
         * Constructor.
         * @param chooseOpNameFrame chooser frame this mouse adapter is associated with
         */
        MyMouseAdapter(ALDChooseOpNameFrame chooseOpNameFrame) {
            this.chooseOpNameFrame = chooseOpNameFrame;
        }

        @Override
        public void mousePressed(MouseEvent e) {
        	int selRow =
        			ALDChooseOpNameFrame.this.opTree.getRowForLocation(e.getX(), e.getY());
        	if(selRow != -1) {
        		if(e.getClickCount() == 2) {
        			//myDoubleClick(selRow, selPath);
        			ALDOperatorChooserTreeNode node = (ALDOperatorChooserTreeNode)ALDChooseOpNameFrame.this.opTree.getLastSelectedPathComponent();
        			if ( node.isOperator() ) {
        				ALDChooseOpNameFrame.this.opName = new String( node.getFullName());
        				ALDChooseOpNameFrame.this.opNameJText.setText( node.toString());
        				ALDChooseOpNameFrame.this.opNameJText.setToolTipText(ALDChooseOpNameFrame.this.opName);
        				ALDOperatorControlFrame controlFrame =
        						this.chooseOpNameFrame.getOpControlFrame( ALDChooseOpNameFrame.this.opName);
        				// if frame is null, an error occurred
        				if (controlFrame != null)
        					controlFrame.setVisible(true);
        			}
        		}
        	}
        }
    }

}
