/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

package de.unihalle.informatik.Alida.gui;

import javax.swing.*;
import javax.swing.tree.*;

import java.util.*;
import java.io.BufferedReader;
import java.io.FileReader;



import de.unihalle.informatik.Alida.annotations.*;
import de.unihalle.informatik.Alida.helpers.ALDEnvironmentConfig;

/**
 * This class extends <code>JTree</code> to hold <code>ALDOperator</code>s to choose from.
 * There are two levels which determine the operators available, namely
 * the standard and application level.
 *  <p>
 *  The initial level of operators displayed is by default the application level.
 *  This may be overridden by a JVM property alida_oprunner_level or
 *  the environment variable ALIDA_OPRUNNER_LEVEL.
 *  <p>
 *  A list of operators to be unfolded at start up may be supplied in a file.
 *  This file is assumed to contain one operator name per line.
 *  The name of this file may be specified by a JVM property alida_oprunner_favoriteops or
 *  the environment variable ALIDA_OPRUNNER_FAVORITEOPS
 *  
 * @author Stefan Posch
 */

//TODO: currently we are not able to retain folding/unfolding between switching of levels
//      as belongs to the JTree and not the TreeModel


public class ALDOperatorChooserTree extends JTree  {

	/**
	 * Debug flag (not accessible from outside).
	 */
	private boolean debug = false;

	/**
	 * Mode of usage.
	 * <p>
	 * In application mode only operators annotated as applications are
	 * available, otherwise all annotated operators are displayed.
	 */
	protected ALDAOperator.Level level = ALDAOperator.Level.APPLICATION;

	/**
	 * List of favorite operators to be unfolded at startup.
	 */
	Vector<String> favoriteOperators;

	/**
	 * Tree of available operators for standard level.
	 */
	protected DefaultTreeModel standardTreeModel;

	/**
	 * Tree of available operators for standard level.
	 */
	protected DefaultTreeModel applicationTreeModel;


	/**
	 * Constructor.
	 */
	public ALDOperatorChooserTree( Collection<String> standardOpNames, Collection<String> applicationOpNames) {

		if ( debug ) {
			System.out.println( "ALDOperatorChooserTree::ALDOperatorChooserTree");
		}
		
		// read configuration
		this.favoriteOperators = initFavoriteOperators();

		// initialize the operator tree
		this.standardTreeModel = createOpTreeModel( standardOpNames, "ALDOperators");
		this.applicationTreeModel = createOpTreeModel( applicationOpNames, "ALDOperators");

		String levelString = ALDEnvironmentConfig.getConfigValue("OPRUNNER","LEVEL");
		if (levelString == null || levelString.equalsIgnoreCase("application") ) {
				this.level = ALDAOperator.Level.APPLICATION;
				this.setModel( applicationTreeModel);
		} else {
				this.level = ALDAOperator.Level.STANDARD;
				this.setModel( standardTreeModel);
		}
		
		hideLeaves( new TreePath(this.getModel().getRoot()));
	}

	/** Return current mode of usage.
	 * 
	 * @return current mode
	 */
	public ALDAOperator.Level getLevel() {
		return level;
	}

	/**
	 * Set the level of the operators in this tree.
	 * 
	 * @param level new level
	 */
	public void setLevel( ALDAOperator.Level level) {
		if ( debug ) {
			System.out.println( "OperatorChooserTree::setLevel to " + level);
		}
		this.level = level;

		if ( level == ALDAOperator.Level.STANDARD) {
			this.setModel( standardTreeModel);
		} else {
			this.setModel( applicationTreeModel);
		}
		hideLeaves( new TreePath(this.getModel().getRoot()));
		
		this.updateUI();
	}
	
	@Override
	public ALDOperatorChooserTreeNode getLastSelectedPathComponent() {
		return  (ALDOperatorChooserTreeNode) super.getLastSelectedPathComponent();
	}

	private DefaultTreeModel createOpTreeModel( Collection<String> operators, String nameOfRootNode) {
		if (this.debug)
			System.out.println("Looking up all annotated @ALDAOperators for level " + level);

		// first we build the tree with our own data structure 
		Vector<String> allOpNames = new Vector<String>();
		AuxNode treeOfOpNames = new AuxNode();

		for (String operator : operators) {
			String className = operator;
			if (debug) {
				System.out.println("ALDOperatorChooserTree::createOpTreeModel adding "
								+ className);
			}

			allOpNames.add(className);

			// build a tree of opNames according to packages
			String[] parts = className.split("\\.");

			AuxNode currentNode = treeOfOpNames;
			// loop over the part but the last one
			int i;
			for (i = 0; i < parts.length - 1; i++) {
				String part = parts[i];
				if (currentNode.children.get(part) == null) {
					AuxNode newChild = new AuxNode(
							(currentNode.fullName.equals("") ? ""
									: currentNode.fullName + ".") + part);
					currentNode.children.put(part, newChild);
				}
				currentNode = currentNode.children.get(part);
			}
			currentNode.opNames.add(parts[i]);
		}

		ALDOperatorChooserTreeNode node = treeOfOpNames.createTree(nameOfRootNode);
		return new DefaultTreeModel(node);
	}

	/**
	 * Hide all leaves of this tree below the given path.
	 * The exception to this rule is that all favorite operators are unfolded 
	 * which also unfolds there siblings.
	 * 
	 * @param treePath Path below which leaves are to be hidden.
	 */
	private void hideLeaves( TreePath treePath) {
		ALDOperatorChooserTreeNode lastNode = (ALDOperatorChooserTreeNode) treePath.getLastPathComponent();
		TreeModel treeModel = this.getModel();
		if (!lastNode.isLeaf())
			this.collapsePath(treePath);
		if (this.favoriteOperators.contains(lastNode.getFullName()))
			this.makeVisible(treePath);

		if (treeModel.getChildCount(lastNode) >= 0) {
			for (int i = 0; i < treeModel.getChildCount(lastNode); i++) {
				TreeNode child = (TreeNode) treeModel.getChild(lastNode, i);
				TreePath extendedPath = treePath.pathByAddingChild(child);
				hideLeaves( extendedPath);
			}
		}
	}

	/**
	 * Initialize list of favorite operators to be unfolded at startup.
	 * <p>
	 * The list is initialized from environment configuration.
	 * 
	 * @return List of favorite operators.
	 */
	private Vector<String> initFavoriteOperators() {
		String favoriteOpsConfigFile = ALDEnvironmentConfig.getConfigValue(
				"OPRUNNER", "FAVORITEOPS");

		Vector<String> favoriteOps = new Vector<String>();
		if (this.debug) {
			System.out
					.println("ALDChooseOpNameFrame::initFavoriteOperators configfile = "
							+ favoriteOpsConfigFile);
		}

		if (favoriteOpsConfigFile != null) {
			try {
				BufferedReader reader = new BufferedReader(new FileReader(
						favoriteOpsConfigFile));
				String line;
				while ((line = reader.readLine()) != null) {
					favoriteOps.add(line);
				}

			} catch (Exception e) {
				// nothing to be done
			}
		}
		return favoriteOps;
	}

	// ========================================================================================

	/**
	 * Node to build up a tree of the package structure of <code>ALDOperator</code>s.
	 * Useful as we have a hash map of children.
	 */
	private class AuxNode {

		/**
		 * Children of this node.
		 */
		HashMap<String, AuxNode> children;

		/**
		 * Operator names associated with this node.
		 */
		LinkedList<String> opNames;

		/**
		 * Full name from the root operator to this node.
		 */
		private String fullName;

		/**
		 * Constructor.
		 * 
		 * @param fName
		 *            Full name of root operator.
		 */
		AuxNode(String fullName) {
			this.children = new HashMap<String, AuxNode>();
			this.opNames = new LinkedList<String>();
			this.fullName = fullName;
		}

		/**
		 * Default constructor.
		 */
		AuxNode() {
			this("");
		}

		/**
		 * Generates the ModelTree using the tree represented with <code>AuxNode</code>s.
		 * 
		 * @param name name of node.
		 * @return Created tree.
		 */
		ALDOperatorChooserTreeNode createTree(String name) {
			// create the root node
			ALDOperatorChooserTreeNode node = new ALDOperatorChooserTreeNode(this.fullName, name, false);

			java.util.Collections.sort(this.opNames);
			for (String oName : this.opNames) {
				ALDOperatorChooserTreeNode child = new ALDOperatorChooserTreeNode(node.getFullName() + "." + oName,
						oName, true);
				node.add(child);
			}

			LinkedList<String> keysSorted = new LinkedList<String>(
					this.children.keySet());
			java.util.Collections.sort(keysSorted);
			for (String str : keysSorted) {
				ALDOperatorChooserTreeNode child = this.children.get(str).createTree(str);
				node.add(child);
			}

			return node;
		}
	}
}
