/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 6091 $
 * $Date: 2012-10-11 14:31:03 +0200 (Do, 11. Okt 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.gui;

import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.util.HashMap;
import java.util.Set;

import javax.swing.*;

import de.unihalle.informatik.Alida.operator.*;
import de.unihalle.informatik.Alida.operator.events.ALDControlEvent;
import de.unihalle.informatik.Alida.operator.events.ALDControlEvent.ALDControlEventType;
import de.unihalle.informatik.Alida.operator.events.ALDConfigurationEvent;


import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.dataio.ALDDataIOManagerSwing.ProviderInteractionLevel;
import de.unihalle.informatik.Alida.dataio.provider.helpers.ALDOperatorParameterPanel;
import de.unihalle.informatik.Alida.dataio.provider.swing.events.ALDSwingValueChangeEvent;
import de.unihalle.informatik.Alida.dataio.provider.swing.events.ALDSwingValueChangeListener;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException.OperatorExceptionType;
import de.unihalle.informatik.Alida.gui.editor.ALDEditorFrame;

/**
 * Frame to configure and run an operator.
 * <p>
 * If the operator is properly configured, i.e. its input parameters are set, it
 * is executed using an <code>ALDExecuteOperatorFrame</code>. This control frame
 * is free to decide, what executing precisily means, usually it is to start the
 * operator via its <code>runOp()</code> method and display or store the results
 * appropriately.
 * <p>
 * Controllable operators trigger the display of additional buttons in GUI for
 * pausing, resuming and interrupting calculations.
 * 
 * @author Stefan Posch
 */

public class ALDOperatorControlFrame extends JFrame implements ActionListener,
		ItemListener, ALDSwingValueChangeListener {

	/**
	 * Local flag for debug output.
	 */
	@SuppressWarnings("unused")
	private boolean debug = false;

	/**
	 * Status of the associated thread to execute operator.
	 */
	protected static enum OpThreadStatus {
		/**
		 * Thread is not running.
		 */
		THREAD_STOPPED,
		/**
		 * Thread is running.
		 */
		THREAD_RUNNING,
		/**
		 * Thread is paused.
		 */
		THREAD_PAUSED
	}

	/**
	 * The operator associated with this frame.
	 */
	protected ALDOperator op = null;

	/**
	 * Current status of the operator execution thread.
	 */
	private OpThreadStatus threadStatus = OpThreadStatus.THREAD_STOPPED;

	/**
	 * Flag to indicate if selected operator is controllable.
	 */
	protected boolean isOpControllable = false;

	/**
	 * Frame to execute a configured operator.
	 * <p>
	 * Configured means that the input parameters have been set properly.
	 */
	protected ALDExecuteOperator executeOperator;

	/**
	 * The top level panel of this frame.
	 */
	protected JPanel inputPanel;

	/**
	 * Title string of window.
	 */
	protected String titleString = null;

	/**
	 * Mode for displaying parameters.
	 * <p>
	 * In standard mode only parameters annotated as standard are shown, in
	 * advanced mode all parameters annotated are displayed.
	 */
	protected Parameter.ExpertMode displayMode = Parameter.ExpertMode.STANDARD;

	/**
	 * Last directory visited, initially it's user's home.
	 */
	protected String lastDirectory = System.getProperty("user.home");

	/**
	 * Last selected file.
	 */
	protected File lastFile = new File("operatorParams.xml");

	/**
	 * Level of interaction providers are allowed to perform.
	 */
	protected ProviderInteractionLevel pLevel =
			ProviderInteractionLevel.ALL_ALLOWED;
	
	/*
	 * Buttons and other components for interactions with running operators.
	 */

	/**
	 * Run button.
	 */
	protected JButton runButton;
	/**
	 * Pause button, changes its label.
	 */
	protected JButton pauseButton;
	/**
	 * Stop button to stop snake optimization (after end of iteration).
	 */
	protected JButton stopButton;
	/**
	 * Button for stepping through optimization process.
	 */
	protected JButton stepButton;
	/**
	 * Box for activating step-through mode.
	 */
	protected JCheckBox stepThroughBox;
	/**
	 * Field for specifying number of iterations per step-through step.
	 */
	protected JTextField stepThroughStepSize;

	/**
	 * Label of status bar, changes dynamically.
	 */
	protected JLabel status;

	/**
	 * Panel where operator parameters are embedded.
	 */
	protected ALDOperatorParameterPanel operatorParameterPanel;

	/**
	 * Ok label to be used on button of Ok message boxes.
	 */
	private final Object[] okOption = { "OK" };

	/**
	 * Labels to be used on buttons of Yes/No message boxes.
	 */
	private final Object[] yesnoOption = { "YES", "NO" };

	/**
	 * Get the frame to execute a configured Operator
	 * 
	 * @return executeOperator
	 */
	public ALDExecuteOperator getALDExecuteOperator() {
		return this.executeOperator;
	}

	/**
	 * Get the operator.
	 * 
	 * @return operator
	 * @throws ALDOperatorException 
	 */
	public ALDOperator getOperator() throws ALDOperatorException {
		if (this.operatorParameterPanel == null || this.op == null)
			return null;
		this.operatorParameterPanel.setOperatorParameters();
		return this.op;
	}

	/**
	 * This method is called to instantiate an ALDOperator. May be overrridden
	 * may extending classes.
	 * 
	 * @param opName full name of operator to instatiate 
	 */
	protected ALDOperator instantiateOperator(String opName) {
		try {
			return (ALDOperator) (Class.forName(opName).newInstance());
		} catch (Exception e) {
			System.err.println("ALDOpRunner failed to instantiate the operator <"
							+ opName + ">");
			return null;
		}
	}

	// TODO comment
	private ALDEditorFrame editorFrame = null;
	private String nodeName = null;

	public void setEditorFrame(ALDEditorFrame frame, String nodeName) {
		this.editorFrame = frame;
		this.nodeName = nodeName;
		this.runButton.setText("Continue");
		this.runButton.setActionCommand("frame_close");
		// check if node is configured
		if (this.validateParameters(true, false)) {
			this.editorFrame.setToConfigured(this.nodeName, true);
		}
		else {
			this.editorFrame.setToConfigured(this.nodeName, false);
		}
	}

	private JRadioButtonMenuItem radioItemStandard;
	private JRadioButtonMenuItem radioItemAdvanced;

	public void changeMode(Parameter.ExpertMode handlingMode) {
		if (handlingMode == Parameter.ExpertMode.STANDARD)
			this.radioItemStandard.doClick();
		else
			this.radioItemAdvanced.doClick();
	}

	/** 
	 * Constructs a control frame for an operator object.
	 * @param _op Operator to be associated with this frame object.
	 * @throws ALDOperatorException
	 */
	public ALDOperatorControlFrame(ALDOperator _op) 
			throws ALDOperatorException {
		if (_op == null)
			throw new ALDOperatorException(OperatorExceptionType.INSTANTIATION_ERROR,
					"[ALDOperatorControlFrame(op)] no operator given, object is null!");
		this.op = _op;
		init();
	}
	
	/**
	 * Constructs a control frame for an operator by name.
	 * @param opName Name of operator to instantiate a control frame for.
	 * @throws ALDOperatorException
	 */
	public ALDOperatorControlFrame(String opName) throws ALDOperatorException {
		try {
			this.op = instantiateOperator(opName);
		} catch (Exception ex) {
//			ex.printStackTrace();
//			JOptionPane.showOptionDialog(
//							null,
//							"Problems instantiating chosen operator...\n"
//									+ "Please check class implementation or contact programmer.",
//							"Warning", JOptionPane.DEFAULT_OPTION,
//							JOptionPane.WARNING_MESSAGE, null, this.okOption,
//							this.okOption[0]);
			this.op = null;
			throw new ALDOperatorException(OperatorExceptionType.INSTANTIATION_ERROR,
					"[ALDOperatorControlFrame(name)] Problems instantiating operator " 
							+ "\"" + opName + "\"\n, something is wrong with the class!");
		}
		// if everything was ok, init the control frame
		init();
	}

	/** does the main work to instantiate the fram
	 * 
	 */
	private void init() {
		if (this.op == null) {
			JOptionPane.showOptionDialog(
							null,
							"Problems instantiating chosen operator...\n"
									+ "Please check class implementation or contact programmer.",
							"Warning", JOptionPane.DEFAULT_OPTION,
							JOptionPane.WARNING_MESSAGE, null, this.okOption,
							this.okOption[0]);
			return;
		}

		// figure out if selected operator is 'controllable'
		if (this.op instanceof ALDOperatorControllable) {
			this.isOpControllable = true;
		} else {
			this.isOpControllable = false;
		}

		if (this.titleString == null)
			this.titleString = "ALDOperatorControlFrame: " + this.op.getName();
		
		// init the operator executer
		this.executeOperator = new ALDExecuteOperator();

		// init the window
		this.setupWindow();

	}
	/**
	 * Specify the provider interaction level.
	 * @param pl		Level to be used in program execution.
	 */
	public void setProviderInteractionLevel(ProviderInteractionLevel pl) {
		this.pLevel = pl;
	}
	
	/**
	 * Construct the frame to configure an operator.
	 */
	protected void setupWindow() {

		// set up the main panel
		this.inputPanel = new JPanel();
		this.add(this.inputPanel);

		BoxLayout ylayout = new BoxLayout(this.inputPanel, BoxLayout.Y_AXIS);
		this.inputPanel.setLayout(ylayout);

		// add operator parameter configuration panel
		this.addParamConfigPanel();

		// add panel with control buttons
		this.addButtonPanel();

		// add status bar
		this.status = new JLabel("Status: Ready");
		JPanel statusBar = new JPanel();
		statusBar.setBorder(BorderFactory.createLoweredBevelBorder());
		statusBar.add(this.status);

		// add the panels to the main panel
		this.inputPanel.add(statusBar);

		// add a nice menubar
		JMenuBar mainWindowMenu = new JMenuBar();
		JMenu fileM = new JMenu("File");
		JMenuItem itemSave = new JMenuItem("Save Settings");
		itemSave.setActionCommand("fileM_save");
		itemSave.addActionListener(this);
		JMenuItem itemLoad = new JMenuItem("Load Settings");
		itemLoad.setActionCommand("fileM_load");
		itemLoad.addActionListener(this);
		JMenuItem itemQuit = new JMenuItem("Quit");
		itemQuit.setActionCommand("fileM_quit");
		itemQuit.addActionListener(this);
		fileM.add(itemSave);
		fileM.add(itemLoad);
		fileM.add(new JSeparator());
		fileM.add(itemQuit);

		JMenu actionsM = new JMenu("Actions");
		JMenuItem itemRun = new JMenuItem("Run");
		itemRun.setActionCommand("actionsM_run");
		itemRun.addActionListener(this);
		JMenuItem itemValidate = new JMenuItem("Validate");
		itemValidate.setActionCommand("actionsM_validate");
		itemValidate.addActionListener(this);
		actionsM.add(itemRun);
		actionsM.add(itemValidate);

		JMenu viewM = new JMenu("View");
		ButtonGroup operatorLevelGroup = new ButtonGroup();
		// default
		this.radioItemStandard = new JRadioButtonMenuItem("Standard");
		if (this.displayMode.equals(Parameter.ExpertMode.STANDARD))
			this.radioItemStandard.setSelected(true);
		this.radioItemStandard.setActionCommand("viewM_standard");
		this.radioItemStandard.addActionListener(this);
		this.radioItemAdvanced = new JRadioButtonMenuItem("Advanced");
		if (this.displayMode.equals(Parameter.ExpertMode.ADVANCED))
			this.radioItemAdvanced.setSelected(true);
		this.radioItemAdvanced.setActionCommand("viewM_advanced");
		this.radioItemAdvanced.addActionListener(this);
		operatorLevelGroup.add(this.radioItemStandard);
		operatorLevelGroup.add(this.radioItemAdvanced);
		viewM.add(this.radioItemStandard);
		viewM.add(this.radioItemAdvanced);

		// generate help menu
		JMenu helpM = this.generateHelpMenu();

		mainWindowMenu.add(fileM);
		mainWindowMenu.add(actionsM);
		mainWindowMenu.add(viewM);
		mainWindowMenu.add(Box.createHorizontalGlue());
		mainWindowMenu.add(helpM);

		// and go ..
		this.setTitle(this.titleString);

		// estimate size according to number of parameters in panel...
		int paramNum = this.op.getParameterNames().size();
		this.setSize(600, paramNum * 50 < 700 ? 700 : paramNum * 50);
		this.setJMenuBar(mainWindowMenu);
		this.pack();

		// check if operator is properly configured
		try {
			if (this.operatorParameterPanel.validateOperatorParameters(true) == null) {
				// operator is ready for execution, make button yellow
				this.runButton.setForeground(Color.green);
			} else {
				this.runButton.setForeground(Color.red);
			}
		} catch (ALDOperatorException e) {
			this.runButton.setForeground(Color.red);
		}
	}

	/**
	 * Sets up the help menu.
	 * 
	 * @return Generated help menu.
	 */
	protected JMenu generateHelpMenu() {
		JMenu helpM = new JMenu("Help");
		JMenuItem itemHelp = new JMenuItem("Online Help");
		itemHelp.addActionListener(OnlineHelpDisplayer.getHelpActionListener(
				itemHelp, this.op.getClass().getName()));
		JMenuItem itemAbout = new JMenuItem("About Alida");
		itemAbout.setActionCommand("helpM_about");
		itemAbout.addActionListener(this);
		helpM.add(itemHelp);
		helpM.add(itemAbout);
		return helpM;
	}

	/**
	 * Adds operator parameter configuration panel to input panel.
	 */
	protected void addParamConfigPanel() {
		this.operatorParameterPanel = new ALDOperatorParameterPanel(this.op,
				this.displayMode, true);
		this.operatorParameterPanel.addValueChangeEventListener(this);
		JScrollPane scrollPane = new JScrollPane(
				this.operatorParameterPanel.getJPanel());
		this.inputPanel.add(scrollPane);
	}

	/**
	 * Adds set of control buttons to the input panel.
	 */
	protected void addButtonPanel() {

		JPanel runPanel = new JPanel();
		runPanel.setLayout(new GridLayout(2, 1));
		// runPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));

		// controllable operators support step-wise execution...
		if (this.isOpControllable) {
			this.stepThroughBox = new JCheckBox("Step-wise execution");
			this.stepThroughBox.setActionCommand("stepFlagToggled");
			this.stepThroughBox.addItemListener(this);
			JLabel label = new JLabel("                 Step size = ");
			this.stepThroughStepSize = new JTextField("1", 5);
			this.stepThroughStepSize.setEnabled(false);
			this.stepButton = new JButton("Next Step");
			this.stepButton.setActionCommand("step");
			this.stepButton.addActionListener(this);
			this.stepButton.setBounds(50, 60, 80, 30);
			this.stepButton.setEnabled(false);

			JPanel stepPanel = new JPanel();
			stepPanel.setLayout(new BoxLayout(stepPanel, BoxLayout.LINE_AXIS));
			stepPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
			stepPanel.add(this.stepThroughBox);
			stepPanel.add(label);
			stepPanel.add(this.stepThroughStepSize);
			stepPanel.add(Box.createRigidArea(new Dimension(70, 0)));
			stepPanel.add(this.stepButton);
			runPanel.add(stepPanel);
		}

		// quit button
		JButton quitButton = new JButton("Quit");
		quitButton.setActionCommand("quit");
		quitButton.addActionListener(this);
		quitButton.setBounds(50, 60, 80, 30);

		// validate button
		// JButton validateButton = new JButton("Validate");
		// validateButton.setActionCommand( "validate");
		// validateButton.addActionListener( this);
		// validateButton.setBounds(50, 60, 80, 30);

		// run button
		this.runButton = new JButton("Run");
		this.runButton.setActionCommand("run");
		this.runButton.addActionListener(this);
		this.runButton.setBounds(50, 60, 80, 30);

		// validatedRun button
		// JButton vrunButton = new JButton("validatedRun");
		// vrunButton.setActionCommand( "validatedRun");
		// vrunButton.addActionListener( this);
		// vrunButton.setBounds(50, 60, 80, 30);

		// pause button for controllable plugins
		this.pauseButton = new JButton("Pause");
		this.pauseButton.setActionCommand("pause");
		this.pauseButton.addActionListener(this);
		this.pauseButton.setBounds(50, 60, 80, 30);
		this.pauseButton.setEnabled(false);

		// stop button for controllable plugins
		this.stopButton = new JButton("Stop");
		this.stopButton.setActionCommand("stop");
		this.stopButton.addActionListener(this);
		this.stopButton.setBounds(50, 60, 80, 30);
		this.stopButton.setEnabled(false);

		// now set up a panel to hold all buttons
		JPanel controlPanel = new JPanel();
		controlPanel
				.setLayout(new BoxLayout(controlPanel, BoxLayout.LINE_AXIS));
		controlPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		controlPanel.add(Box.createHorizontalGlue());
		// controlPanel.add(validateButton);
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		controlPanel.add(this.runButton);
		// if operator is not controllable, disable control buttons
		if (this.isOpControllable) {
			controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
			controlPanel.add(this.pauseButton);
			controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
			controlPanel.add(this.stopButton);
		}
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		controlPanel.add(quitButton);

		runPanel.add(controlPanel);
		this.inputPanel.add(runPanel);
	}

	/**
	 * Changes the status message to the given string.
	 * 
	 * @param msg
	 *            Message to be shown in statusbar.
	 */
	protected void setStatus(String msg) {
		this.status.setText("Status: " + msg);
	}

	/**
	 * Show the result frame of the operator.
	 */
	public void showResult() {
		this.executeOperator.getResultFrame().setVisible(true);
	}

	/**
	 * Run the operator.
	 * 
	 * @param wait
	 *            True if operator has to wait until finished.
	 * @return false if validation failed.
	 * @throws ALDOperatorException
	 */
	public boolean runOperator(boolean wait, boolean showResult) throws ALDOperatorException {
		HashMap<ALDParameterDescriptor, String> failedParams;
		failedParams = this.operatorParameterPanel
				.validateOperatorParameters(true);
		if (failedParams != null) {
			this.setVisible(true);
			return false;
		}
		this.runButton.setEnabled(false);
		if (this.isOpControllable) {
			this.pauseButton.setEnabled(true);
			this.stopButton.setEnabled(true);
		}
		this.threadStatus = OpThreadStatus.THREAD_RUNNING;
		this.setStatus("Running...");
		try {
			this.operatorParameterPanel.setOperatorParameters();
			if (this.op instanceof ALDOperatorControllable) {
				// tell the operator the requested step size via event...
				ALDConfigurationEvent confev = new ALDConfigurationEvent(this);
				confev.disableStepwiseExecution();
				// ... send it to operator
				((ALDOperatorControllable) this.op)
						.handleALDConfigurationEvent(confev);
			}
			this.executeOperator.init(this.op, this, showResult, this.pLevel);
			Thread t = new Thread(this.executeOperator);
			this.editorFrame.setToInExecution(this.nodeName, false);
			t.start();
			if (wait)
				t.join();
			return true;
		} catch (Exception ex) {
			System.out.println(ex.getMessage());
			ex.printStackTrace();
		}
		return false;
	}

	/**
	 * Validate the operator parameters by requesting values from the panels.
	 * 
	 * @return True, if validation is successful, otherwise false.
	 */
	public boolean validateParameters(boolean deepValidate, boolean allowWarnings) {
		try {
			HashMap<ALDParameterDescriptor, String> failedList = this.operatorParameterPanel
					.validateOperatorParameters(deepValidate);
			if (failedList != null && allowWarnings) {
				StringBuffer msg = new StringBuffer();
				Set<ALDParameterDescriptor> keys = failedList.keySet();
				for (ALDParameterDescriptor descr : keys) {
					msg.append("Reading parameter \"");
					msg.append(descr.getLabel() + " failed!\n");
					msg.append("--> " + failedList.get(descr) + "\n");
				}
				JOptionPane.showOptionDialog(null, msg,
						"Warning - Parameter Validation Failed!",
						JOptionPane.DEFAULT_OPTION,
						JOptionPane.WARNING_MESSAGE, null, this.okOption,
						this.okOption[0]);
				return false;
			}
			else if (failedList != null) {
				return false;
			}
		} catch (ALDOperatorException ex) {
			if (allowWarnings)
				JOptionPane.showOptionDialog(null,
						"Call of operator's validate() method failed! Reason:\n"
								+ ex.getCommentString(), "Warning",
						JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
						null, this.okOption, this.okOption[0]);
			return false;
		}
		return true;
	}

	@Override
	public void actionPerformed(ActionEvent e) {

		// local variables
		String command = e.getActionCommand();

		// run the configured operator
		if (command.equals("run") || command.equals("actionsM_run")) {
			if (!this.validateParameters(true,true))
				return;
			this.runButton.setEnabled(false);
			if (this.isOpControllable) {
				this.pauseButton.setEnabled(true);
				this.stopButton.setEnabled(true);
			}
			this.threadStatus = OpThreadStatus.THREAD_RUNNING;
			this.setStatus("Running...");
			try {
				this.operatorParameterPanel.setOperatorParameters();
				if (this.op instanceof ALDOperatorControllable) {
					// tell the operator the requested step size via event...
					ALDConfigurationEvent confev = new ALDConfigurationEvent(
							this);
					confev.disableStepwiseExecution();
					// ... send it to operator
					((ALDOperatorControllable) this.op)
							.handleALDConfigurationEvent(confev);
				}
				this.executeOperator.init(this.op, this, true, this.pLevel);
				Thread t = new Thread(this.executeOperator);
				t.start();
				// t.join();
				// this.showResult();
				// this.setVisible(false);
			} catch (Exception ex) {
				System.out.println(ex.getMessage());
				ex.printStackTrace();
			}
		}
		// validate
		else if (command.equals("actionsM_validate")) {
			this.validateParameters(true,true);
		}
		// TODO
		else if (command.equals("frame_close")) {
			try {
				if (this.operatorParameterPanel
						.validateOperatorParameters(false) == null) {
					this.setVisible(false);
				}
			} catch (ALDOperatorException e1) {
				this.runButton.setForeground(Color.red);
			}
		}
		// quit, i.e. close window
		else if (command.equals("quit") || command.equals("fileM_quit")) {
			if (this.threadStatus == OpThreadStatus.THREAD_RUNNING
					|| this.threadStatus == OpThreadStatus.THREAD_PAUSED) {
				int selection = JOptionPane
						.showOptionDialog(
								null,
								"The operator is running or paused!\n"
										+ "If you quit now, proper termination is impossible.\n"
										+ "Do you reallly want to quit?",
								"Warning", JOptionPane.DEFAULT_OPTION,
								JOptionPane.WARNING_MESSAGE, null,
								this.yesnoOption, this.yesnoOption[0]);
				switch (selection) {
				case JOptionPane.YES_OPTION:
					this.threadStatus = OpThreadStatus.THREAD_STOPPED;
					this.setVisible(false);
					break;
				case JOptionPane.NO_OPTION:
				case JOptionPane.CANCEL_OPTION:
					return;
				}
			} else {
				this.threadStatus = OpThreadStatus.THREAD_STOPPED;
				this.setVisible(false);
			}
		}
		// stop operator execution
		else if (command.equals("stop")) {
			((ALDOperatorControllable) this.op)
					.handleALDControlEvent(new ALDControlEvent(this,
							ALDControlEventType.STOP_EVENT));
			this.threadStatus = OpThreadStatus.THREAD_STOPPED;
			this.stopButton.setEnabled(false);
			this.pauseButton.setEnabled(false);
			this.setStatus("Stopped.");
		}
		// pause operator execution
		else if (command.equals("pause")) {
			((ALDOperatorControllable) this.op)
					.handleALDControlEvent(new ALDControlEvent(this,
							ALDControlEventType.PAUSE_EVENT));
			this.pauseButton.setActionCommand("continue");
			this.pauseButton.setText("Continue");
			this.stopButton.setEnabled(false);
			this.threadStatus = OpThreadStatus.THREAD_PAUSED;
			this.setStatus("Paused...");
		}
		// resume operator execution
		else if (command.equals("continue")) {
			((ALDOperatorControllable) this.op)
					.handleALDControlEvent(new ALDControlEvent(this,
							ALDControlEventType.RESUME_EVENT));
			this.pauseButton.setActionCommand("pause");
			this.pauseButton.setText("Pause");
			this.stopButton.setEnabled(true);
			this.threadStatus = OpThreadStatus.THREAD_RUNNING;
			this.setStatus("Running...");
		}
		// next step in step-wise execution
		else if (command.equals("step")) {
			if (this.threadStatus == OpThreadStatus.THREAD_RUNNING) {
				((ALDOperatorControllable) this.op)
						.handleALDControlEvent(new ALDControlEvent(this,
								ALDControlEventType.STEP_EVENT));
				this.stopButton.setEnabled(true);
			}
			// start the thread
			else {
				this.threadStatus = OpThreadStatus.THREAD_RUNNING;
				this.setStatus("Running...");
				try {
					this.operatorParameterPanel.setOperatorParameters();
					String stepSizeText = this.stepThroughStepSize.getText();
					int size = Integer.valueOf(stepSizeText).intValue();
					// tell the operator the requested step size via event...
					ALDConfigurationEvent confev = new ALDConfigurationEvent(
							this);
					confev.setStepsize(size);
					confev.enableStepwiseExecution();
					// ... send it to operator
					((ALDOperatorControllable) this.op)
							.handleALDConfigurationEvent(confev);
					this.executeOperator.init(this.op, this, true, this.pLevel);
					new Thread(this.executeOperator).start();
					// this.setVisible( false);
				} catch (Exception ex) {
					System.out.println(ex.getMessage());
					ex.printStackTrace();
				}
			}
		}
		// thread terminated (attention: event is not induced by a button,
		// but just from internally...)
		else if (command.equals("threadstop")) {
			this.setStatus("Ready.");
			this.runButton.setEnabled(true);
			this.pauseButton.setActionCommand("pause");
			this.pauseButton.setText("Pause");
			this.pauseButton.setEnabled(false);
			this.stopButton.setEnabled(false);
			this.threadStatus = OpThreadStatus.THREAD_STOPPED;

			if (this.editorFrame != null) {
				this.editorFrame.runNextNode();
//				this.editorFrame.runOpGraph(nodeName);
				this.editorFrame.setToInExecution(this.nodeName, true);
			}
		}

		// handle menu item commands

		else if (command.equals("fileM_save")) {
			// open file chooser
			JFileChooser getFileDialog = new JFileChooser();
			getFileDialog.setApproveButtonText("Save");
			getFileDialog.setCurrentDirectory(new File(this.lastDirectory));
			getFileDialog.setSelectedFile(this.lastFile);
			getFileDialog.setFileSelectionMode(JFileChooser.FILES_ONLY);
			int returnVal = getFileDialog.showOpenDialog(null);
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				try {
					HashMap<ALDOpParameterDescriptor, String> failedParams = this.operatorParameterPanel
							.setOperatorParameters();
					if (failedParams != null) {
						StringBuffer msg = new StringBuffer();
						Set<ALDOpParameterDescriptor> keys = failedParams
								.keySet();
						for (ALDOpParameterDescriptor descr : keys) {
							msg.append("Loading parameter " + descr.getLabel()
									+ "failed!\n");
							msg.append("--> " + failedParams.get(descr) + "\n");
						}
						JOptionPane.showOptionDialog(null,
								"Loading parameters failed, not all parameters could be loaded!\n"
										+ msg,
								"Warning - Problems loading parameters!",
								JOptionPane.DEFAULT_OPTION,
								JOptionPane.WARNING_MESSAGE, null,
								this.okOption, this.okOption[0]);
					}
				} catch (ALDOperatorException ex) {
					JOptionPane
							.showOptionDialog(
									null,
									"Loading parameters failed, \n operator could not be configured!",
									"Error - Loading parameters failed!",
									JOptionPane.DEFAULT_OPTION,
									JOptionPane.ERROR_MESSAGE, null,
									this.okOption, this.okOption[0]);
				}
				File file = getFileDialog.getSelectedFile();
				this.lastFile = file;
				this.op.serializeToXmlFile(file.getAbsolutePath());
				this.lastDirectory = file.getAbsolutePath();
			}
		} else if (command.equals("fileM_load")) {
			// open file chooser
			JFileChooser getFileDialog = new JFileChooser();
			getFileDialog.setCurrentDirectory(new File(this.lastDirectory));
			getFileDialog.setSelectedFile(this.lastFile);
			getFileDialog.setFileSelectionMode(JFileChooser.FILES_ONLY);
			int returnVal = getFileDialog.showOpenDialog(null);
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				File file = getFileDialog.getSelectedFile();
				this.lastDirectory = file.getAbsolutePath();
				this.lastFile = file;
				this.op = this.op
						.deserializeFromXmlFile(file.getAbsolutePath());
				// update the GUI with new parameter values
				// this.operatorParameterPanel.updateConfigPanel(this.op);
				// update the GUI with new parameter values
				try {
					HashMap<ALDOpParameterDescriptor, String> failedParams = this.operatorParameterPanel
							.updateConfigPanel(this.op);
					if (failedParams != null) {
						StringBuffer msg = new StringBuffer();
						Set<ALDOpParameterDescriptor> keys = failedParams
								.keySet();
						for (ALDOpParameterDescriptor descr : keys) {
							msg.append("Updating parameter " + descr.getLabel()
									+ "failed!\n");
							msg.append("--> " + failedParams.get(descr) + "\n");
						}
						JOptionPane.showOptionDialog(null,
								"Problems loading file, not all parameters could be updated!\n"
										+ msg, "Warning",
								JOptionPane.DEFAULT_OPTION,
								JOptionPane.WARNING_MESSAGE, null,
								this.okOption, this.okOption[0]);
					}
				} catch (ALDDataIOProviderException ex) {
					JOptionPane.showOptionDialog(null,
							"Problems loading file, operator does not match!\n"
									+ ex.getCommentString(), "Error",
							JOptionPane.DEFAULT_OPTION,
							JOptionPane.ERROR_MESSAGE, null, this.okOption,
							this.okOption[0]);
				}
			}
		} else if (command.equals("viewM_standard")) {
			this.operatorParameterPanel
					.changeViewMode(Parameter.ExpertMode.STANDARD);
			this.repaint();
		} else if (command.equals("viewM_advanced")) {
			this.operatorParameterPanel
					.changeViewMode(Parameter.ExpertMode.ADVANCED);
			this.repaint();
		} else if (command.equals("helpM_about")) {
			JOptionPane
					.showOptionDialog(
							null,
							"Alida / MiToBo Graphical Operator Runner - \n"
									+ "Chosen Operator: "
									+ this.op.getName()
									+ "\n"
									+ " @2012 Martin Luther University Halle-Wittenberg",
							"About Alida / MiToBo", JOptionPane.DEFAULT_OPTION,
							JOptionPane.PLAIN_MESSAGE, null, this.okOption,
							this.okOption[0]);
		}
	}

	// TODO
	/*public void configuration(String filename, boolean save)
			throws ALDOperatorException, ALDDataIOProviderException {
		if (save) {
			this.operatorParameterPanel.setOperatorParameters();
			this.op.serializeToXmlFile(filename);
		} else {
			this.op = this.op.deserializeFromXmlFile(filename);
			this.operatorParameterPanel.updateConfigPanel(this.op);
		}
	}
	*/
	

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
	 */
	@Override
	public void itemStateChanged(ItemEvent e) {

		String idString = null;

		// get the object affected by the event
		Object source = e.getItemSelectable();

		if (source instanceof JCheckBox) {
			JCheckBox box = (JCheckBox) source;
			idString = box.getActionCommand();
		}
		if (source instanceof JComboBox) {
			JComboBox box = (JComboBox) source;
			idString = box.getActionCommand();
		}

		// step through optimization
		if (idString == "stepFlagToggled") {
			if (this.stepThroughBox.isSelected()) {
				// step-wise execution
				this.stepThroughStepSize.setEnabled(true);
				this.stepButton.setEnabled(true);
				this.runButton.setEnabled(false);
				this.pauseButton.setEnabled(false);
			} else {
				// no step-wise execution
				this.stepThroughStepSize.setEnabled(false);
				this.stepButton.setEnabled(false);
				this.runButton.setEnabled(true);
				this.pauseButton.setEnabled(false);
			}
		}
		// update GUI
		this.repaint();
	}

	/**
	 * Replace the panel of given parameter to insert a link.
	 * 
	 * @param opLink
	 *            Name of parameter.
	 * @param value
	 *            Default value of the new panel.
	 */
	public void changePanel(String opLink, Object value) {
		this.operatorParameterPanel
				.changePanel(opLink, value, this.editorFrame);
	}
	
	public void setDescriptorLinked(String opLink, boolean b) {
		this.operatorParameterPanel.setDescriptorLinked(opLink, b);
	}

	@Override
	public void handleValueChangeEvent(ALDSwingValueChangeEvent event) {
		// check if operator is properly configured now
		try {
			HashMap<ALDParameterDescriptor, String> failedParams =
					this.operatorParameterPanel.validateOperatorParameters(false);
			if (failedParams == null) {
				// operator is ready for execution, make button green
				this.runButton.setForeground(Color.green);
				if (this.editorFrame != null)
					this.editorFrame.setToConfigured(this.nodeName, true);
			} 
			else {
				// operator is not ready for execution, make button red
				this.runButton.setForeground(Color.red);
				if (this.editorFrame != null)
					this.editorFrame.setToConfigured(this.nodeName, false);
			}
		} catch (ALDOperatorException e) {
			this.runButton.setForeground(Color.red);
			if (this.editorFrame != null)
				this.editorFrame.setToConfigured(this.nodeName, false);
		}
	}

	// TODO comments
	public void copyConfiguration(ALDOperatorControlFrame controlFrame)
			throws ALDDataIOException {
		this.operatorParameterPanel
				.copyConfiguration(controlFrame.operatorParameterPanel);
	}

}
