/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 6078 $
 * $Date: 2012-10-05 17:03:38 +0200 (Fr, 05. Okt 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.gui.editor;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Set;

import com.mxgraph.model.mxCell;
import com.mxgraph.view.mxGraph;




/**
 * Class to represent workflow graphs in Grappa.
 * <p>
 * This class basically extends the graph class {@link mxGraph} of the 
 * JGraphx library for routines to extract specific subsets of nodes.
 * 
 * @author moeller
 */

public class ALDWorkflowGraph extends mxGraph {
	
	/**
	 * Returns all nodes the graph currently contains on top level.
	 * @return	List of nodes.
	 */
	public LinkedList<mxCell> getAllNodes() {
		Object [] allCells = this.getChildVertices(this.getDefaultParent());
		LinkedList<mxCell> allNodes = new LinkedList<mxCell>();
		for (Object o : allCells) {
			if (((mxCell)o).isVertex())
				allNodes.add((mxCell)o);
		}
		return allNodes;
	}

	/**
	 * Returns all top level nodes of the graph in topological order. 
	 * @return	List of nodes, null if graph is cyclic and sorting is impossible.
	 */
	public LinkedList<mxCell> getAllNodesTopologicallyOrdered() {
		Object [] allCells = this.getChildVertices(this.getDefaultParent());
		LinkedList<mxCell> allNodes = new LinkedList<mxCell>();
		for (Object o : allCells) {
			if (((mxCell)o).isVertex())
				allNodes.add((mxCell)o);
		}
		return this.sortNodesTopologically(allNodes);
	}
	
	/**
	 * Returns all subsequent nodes of the given one in topological order.
	 * <p>
	 * Subsequent nodes are defined as nodes on the same graph level as the given 
	 * one, and linked to each other (only) via child links.
	 * @param node	Node to be considered.
	 * @return	List of nodes, null if subgraph is cyclic and sorting impossible.
	 */
	public LinkedList<mxCell>	getSubsequentNodesTopologicallyOrdered(
																																mxCell node) {
		return this.sortNodesTopologically(this.getSubsequentNodes(node));
	}

	/**
	 * Returns all subsequent nodes of the given one.
	 * <p>
	 * Subsequent nodes are defined as nodes on the same graph level as the given 
	 * one, and linked to each other (only) via child links.
	 * @param node	Node to be considered.
	 * @return	List of nodes.
	 */
	public LinkedList<mxCell>	getSubsequentNodes(mxCell node) {
		Object [] allCells = this.getChildVertices(this.getDefaultParent());
		LinkedList<mxCell> notChecked = new LinkedList<mxCell>();
		for (Object obj : allCells) {
			notChecked.add((mxCell)obj);
		}
		LinkedList<mxCell> successors = new LinkedList<mxCell>();
		// node itself belongs also to list
		successors.add(node);
		LinkedList<mxCell> nodesToCheck = new LinkedList<mxCell>();
		nodesToCheck.add(node);
		while (!nodesToCheck.isEmpty()) {
			mxCell nextToCheck = nodesToCheck.pollFirst();
			for (int i = 0; i < nextToCheck.getChildCount(); i++) {
				Object[] outEdges = this.getOutgoingEdges(nextToCheck.getChildAt(i));
				for (Object obj : outEdges) {
					mxCell edge = (mxCell)obj;
					mxCell succNode = (mxCell)edge.getTarget().getParent();
					if (!successors.contains(succNode))
						successors.add(succNode);
					if (notChecked.contains(succNode)) {
						nodesToCheck.add(succNode);
						notChecked.remove(succNode);
					}
				}
			}
		}
		return successors;
	}
	
	/**
	 * Returns subgraph of all subsequent nodes including their predessors
	 * in topological order.
	 * <p>
	 * Subsequent/preceding nodes are here defined as nodes on the same graph 
	 * level as the given one, and linked to each other (only) via child links.
	 * @param node	Node to be considered.
	 * @return List of nodes, null if subgraph is cyclic and sorting impossible.
	 */
	public LinkedList<mxCell>	getSubsequentSubgraphNodesTopologicallyOrdered(
																																mxCell node) {
		return this.sortNodesTopologically(this.getSubsequentSubgraphNodes(node));
	}

	/**
	 * Returns subgraph of all subsequent nodes including their predessors.
	 * <p>
	 * Subsequent/preceding nodes are here defined as nodes on the same graph 
	 * level as the given one, and linked to each other (only) via child links.
	 * @param node	Node to be considered.
	 * @return List of nodes.
	 */
	public LinkedList<mxCell>	getSubsequentSubgraphNodes(mxCell node) {
		LinkedList<mxCell> subseqNodes = this.getSubsequentNodes(node);		
		
		Object [] allCells = this.getChildVertices(this.getDefaultParent());
		LinkedList<mxCell> notChecked = new LinkedList<mxCell>();
		for (Object obj : allCells) {
			notChecked.add((mxCell)obj);
		}
		LinkedList<mxCell> successors = new LinkedList<mxCell>();
		LinkedList<mxCell> nodesToCheck = new LinkedList<mxCell>();
		nodesToCheck.addAll(subseqNodes);
		while (!nodesToCheck.isEmpty()) {
			mxCell nextToCheck = nodesToCheck.pollFirst();
			if (notChecked.contains(nextToCheck)) {
				successors.add(nextToCheck);
				notChecked.remove(nextToCheck);
			}
			// get preceding nodes
			LinkedList<mxCell> precNodes = this.getPrecedingNodes(nextToCheck);
			for (mxCell precNode : precNodes) {
				if (notChecked.contains(precNode)) {
					nodesToCheck.add(precNode);
				}
			}
		}
		return successors;
	}

	/**
	 * Returns set of preceding nodes in topological order.
	 * <p>
	 * Preceding nodes are defined as nodes on the same graph level as the given 
	 * one and being linked to that node (directly or indirectly) via 
	 * incoming edges (only) on the level of the children of the nodes.
	 * @param node	Node to be considered.
	 * @return	List of nodes, null if graph is cyclic and sorting impossible.
	 */
	public LinkedList<mxCell>	getPrecedingNodesTopologicallyOrdered(mxCell node) {
		return this.sortNodesTopologically(this.getPrecedingNodes(node));
	}

	public LinkedList<mxCell>	getPrecedingNodes(mxCell node) {
		Object [] allCells = this.getChildVertices(this.getDefaultParent());
		LinkedList<mxCell> notChecked = new LinkedList<mxCell>();
		for (Object obj : allCells) {
			notChecked.add((mxCell)obj);
		}
		LinkedList<mxCell> predessors = new LinkedList<mxCell>();

		// node itself belongs also to list
		predessors.add(node);
		
		LinkedList<mxCell> nodesToCheck = new LinkedList<mxCell>();
		nodesToCheck.add(node);
		while (!nodesToCheck.isEmpty()) {
			mxCell nextToCheck = nodesToCheck.pollFirst();
			for (int i = 0; i < nextToCheck.getChildCount(); i++) {
				Object[] inEdges = this.getIncomingEdges(nextToCheck.getChildAt(i));
				for (Object obj : inEdges) {
					mxCell edge = (mxCell)obj;
					mxCell precNode = (mxCell)edge.getSource().getParent();
					if (!predessors.contains(precNode))
						predessors.add(precNode);
					if (notChecked.contains(precNode)) {
						nodesToCheck.add(precNode);
						notChecked.remove(precNode);
					}
				}
			}
		}
		return predessors;
	}

	/**
	 * Sorts a list of graph nodes topologically in ascending order.
	 * @param inList		List with nodes to sort.
	 * @return	Sorted list in ascending order, null if graph is cyclic.
	 */
	private LinkedList<mxCell> sortNodesTopologically(
			Collection<mxCell> inList) {
		HashMap<mxCell, Integer> inEdgeCounter= new HashMap<mxCell, Integer>();
		for (mxCell cell : inList) {
			int count = 0;
			for (int i = 0; i < cell.getChildCount(); i++) {
				Object[] inEdges = this.getIncomingEdges(cell
						.getChildAt(i));
				for (Object obj : inEdges) {
					mxCell edge = (mxCell)obj;
					if (inList.contains(edge.getSource().getParent())) {
						count++;
					}
				}
			}
			inEdgeCounter.put(cell, new Integer(count));
		}
		
		// iterative removal of nodes from list according to input degree
		LinkedList<mxCell> outList = new LinkedList<mxCell>();
		boolean nodeFound = false;
		while (!inEdgeCounter.isEmpty()) {
			Set<mxCell> cells = inEdgeCounter.keySet();
			nodeFound = false;
			for (mxCell c : cells) {
				if (inEdgeCounter.get(c).intValue() == 0) {
					nodeFound = true;
					for (int i = 0; i < c.getChildCount(); i++) {
						Object[] outEdges = this.getOutgoingEdges(c.getChildAt(i));
						for (Object e : outEdges) {
							mxCell target = (mxCell) ((mxCell)e).getTarget();
							// Attention: 
							// skip edges linked to nodes not present in current subset
							if (inList.contains(target.getParent())) {
								inEdgeCounter.put((mxCell)target.getParent(), new Integer(
												inEdgeCounter.get(target.getParent()).intValue()-1));
							}
						}
					}
					outList.add(c);
					inEdgeCounter.remove(c);
					break;
				}
			}
			if (!nodeFound) {
				// graph is cyclic, no chance to sort nodes!
				return null;
			}
		}
		return outList;
	}

}
