/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5521 $
 * $Date: 2012-04-24 15:36:55 +0200 (Di, 24. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.core.datatypes;

import java.io.*;
import java.util.*;



import de.unihalle.informatik.Alida.exceptions.*;
import de.unihalle.informatik.Alida.operator.ALDData;
import de.unihalle.informatik.MiToBo.core.exceptions.MTBException;
import de.unihalle.informatik.MiToBo.core.operator.*;
import de.unihalle.informatik.MiToBo_xml.*;

/**
 * A set of 2D contours living in a common domain which is a rectangular subset
 * of R x R. Each contours is of type Contour2D
 * 
 * 
 * @author Stefan Posch
 */

public class MTBContour2DSet extends ALDData implements Cloneable {
		// subset of R x R
		/**
		 * Minimal x coordinate of the domain of this contour set.
		 */
		private double xMin;

		/**
		 * Minimal y coordinate of the domain of this contour set.
		 */
		private double yMin;

		/**
		 * Maximal x coordinate of the domain of this contour set.
		 */
		private double xMax;

		/**
		 * Maximal y coordinate of the domain of this contour set.
		 */
		private double yMax;

		/**
		 * The set of contours represented as a Vector.
		 */
		Vector<MTBContour2D> contourSet;

		/**
		 * Standard constructor.
		 */
		public MTBContour2DSet() {
				this.xMin = 0.0;
				this.yMin = 0.0;
				this.xMax = 0.0;
				this.yMax = 0.0;
				contourSet = new Vector<MTBContour2D>();
				setProperty("xMin", 0.0);
				setProperty("yMin", 0.0);
				setProperty("xMax", 0.0);
				setProperty("yMax", 0.0);
		}

		/**
		 * Construct an empty set of contours with given extent of domain.
		 * 
		 * @param xMin
		 *          minimal x-coordinate of the domain
		 * @param yMin
		 *          minimal y-coordinate of the domain
		 * @param xMax
		 *          maximal x-coordinate of the domain
		 * @param yMax
		 *          maximal y-coordinate of the domain
		 */
		public MTBContour2DSet(double xMin, double yMin, double xMax, double yMax) {
				this.xMin = xMin;
				this.yMin = yMin;
				this.xMax = xMax;
				this.yMax = yMax;
				contourSet = new Vector<MTBContour2D>();
				setProperty("xMin", xMin);
				setProperty("yMin", yMin);
				setProperty("xMax", xMax);
				setProperty("yMax", yMax);
		}

		@Override
		public MTBContour2DSet clone() {
			MTBContour2DSet newSet = 
					new MTBContour2DSet(this.xMin, this.yMin, this.xMax, this.yMax);
			newSet.contourSet = new Vector<MTBContour2D>();
			for (MTBContour2D cont: this.contourSet) {
				newSet.contourSet.add(cont.clone());
			}
			newSet.setProperty("xMin", new Double(this.xMin));
			newSet.setProperty("yMin", new Double(this.yMin));
			newSet.setProperty("xMax", new Double(this.xMax));
			newSet.setProperty("yMax", new Double(this.yMax));
			return newSet;
		}
		
		/**
		 * Get the number of contours of this contour set
		 * 
		 * @return Number of contours.
		 */
		public int size() {
				return contourSet.size();
		}

		/**
		 * Get a contour by index.
		 * 
		 * @return The i-th contour.
		 */
		public MTBContour2D elementAt(int i) {
				if (i < 0 || i >= size())
						throw new ArrayIndexOutOfBoundsException();
				return contourSet.elementAt(i);
		}

		/**
		 * Append a contour (at the end) to the set of contours.
		 * 
		 * @param contour
		 *          contour to add
		 */
		public boolean add(MTBContour2D contour) {
				return contourSet.add(contour);
		}

		/**
		 * Set the i-th contour from the set.
		 * 
		 * @param contour
		 *          new contour element
		 * @param i
		 *          index of the new contour
		 */
		public void setElementAt(MTBContour2D contour, int i) {
				contourSet.setElementAt(contour, i);
		}

		/**
		 * Delete the i-th contour from the set
		 * 
		 * @param i
		 *          index of contour to remove
		 */
		public void removeElementAt(int i) {
				contourSet.removeElementAt(i);
		}

		/**
		 * Read a contour set from an xml file <code>filename</code> and set the MTB
		 * polygon set accordingly. The processing history is read also if available.
		 * <p>
		 * WARNING: currently assume filename WITHOUT extension.
		 * 
		 * @param filename
		 *          Filename to read from, WITHOUT extension (for the moment).
		 */
		public void read(String filename) throws ClassNotFoundException, MTBException {

				MTBXMLContour2DSetDocument xmlContourSetDocument;
				try {
						File file = new File(filename + ".xml");
						setLocation(file.getCanonicalPath());

						// instantiate XML document by file parsing
						xmlContourSetDocument = MTBXMLContour2DSetDocument.Factory.parse(file);
						MTBXMLContour2DSetType xmlContourSet = xmlContourSetDocument
						    .getMTBXMLContour2DSet();

						// get bounding box
						this.xMin = xmlContourSet.getXMin();
						this.yMin = xmlContourSet.getYMin();
						this.xMax = xmlContourSet.getXMax();
						this.yMax = xmlContourSet.getYMax();

						// resize contour set and copy contours from xml
						this.contourSet = new Vector<MTBContour2D>();
						this.contourSet.setSize(xmlContourSet.sizeOfContoursArray());
						for (int idx = 0; idx < xmlContourSet.sizeOfContoursArray(); idx++) {
								MTBXMLContour2DType xmlContour = xmlContourSet.getContoursArray(idx);
								this.contourSet.setElementAt(getContour2DFromXml(xmlContour), idx);
						}
				} catch (Exception e) {
						e.printStackTrace();
						// TODO: throw Mitobo exception instead ??
						throw new ClassNotFoundException(
						    "MTBXMLContour2DSet::read failed to read " + filename + ".xml");
				}
				// MTBPortHashAccess.readHistory(this, filename);
				MTBOperator.readHistory(this, filename);
		}

		/**
		 * Write this contour set into XML file <code>filename.xml</code>
		 * Additionally, its processing history is written.
		 * <p>
		 * WARNING: currently assume filename WITHOUT extension.
		 * 
		 * @param filename
		 *          Filename to write to, WITHOUT extension (for the moment)
		 */
		public void write(String filename) throws ALDException {
				try {
						BufferedWriter file = new BufferedWriter(
						    new FileWriter(filename + ".xml"));

						// generate XML documents
						MTBXMLContour2DSetDocument xmlContourSetDocument = MTBXMLContour2DSetDocument.Factory
						    .newInstance();
						MTBXMLContour2DSetType xmlContourSet = xmlContourSetDocument
						    .addNewMTBXMLContour2DSet();

						// transfer the contours to XML
						MTBXMLContour2DType[] cList = new MTBXMLContour2DType[this.contourSet
						    .size()];
						for (int p = 0; p < this.contourSet.size(); p++) {
								MTBContour2D contour = this.contourSet.elementAt(p);
								cList[p] = this.getContour2DAsXml(contour, null);
						}
						xmlContourSet.setContoursArray(cList);
						xmlContourSet.setXMin(xMin);
						xmlContourSet.setYMin(yMin);
						xmlContourSet.setXMax(xMax);
						xmlContourSet.setYMax(yMax);

						// write the xml file
						file.write(xmlContourSetDocument.toString());
						file.close();
				} catch (Exception e) {
						System.err.println("Exception" + e);
				}
				// write processing history
				// MTBPortHashAccess.writeHistory(this, filename);
				MTBOperator.writeHistory(this, filename);
		}

		/**
		 * Copy the information of <code>contour</code> into the corresponding xml
		 * element <code>xmlContour</code>. If <code>xmlContour</code> is null, a new
		 * obejct is created, otherwise the passed object filled.
		 */
		public MTBXMLContour2DType getContour2DAsXml(MTBContour2D contour,
		    MTBXMLContour2DType xmlC) {

				MTBXMLContour2DType xmlContour;

				// instantiate contour
				if (xmlC == null)
						xmlContour = MTBXMLContour2DType.Factory.newInstance();
				else
						xmlContour = xmlC;

				// list of contour points
				MTBXMLPointVectorType xmlPlist = MTBXMLPointVectorType.Factory
				    .newInstance();
				Vector<java.awt.geom.Point2D.Double> points = contour.getPoints();
				for (int i = 0; i < points.size(); i++) {
						MTBXMLPoint2DDoubleType xmlPoint = xmlPlist.addNewPoint();
						xmlPoint.setX((int) points.elementAt(i).getX());
						xmlPoint.setY((int) points.elementAt(i).getY());
				}
				xmlContour.setPoints(xmlPlist);

				// inner contours
				Vector<MTBContour2D> innerContours = contour.getAllInner();
				MTBXMLContour2DType[] inList = new MTBXMLContour2DType[innerContours.size()];
				for (int i = 0; i < innerContours.size(); ++i) {
						inList[i] = this.getContour2DAsXml(innerContours.get(i), null);
				}
				xmlContour.setInnerArray(inList);
				return xmlContour;
		}

		/**
		 * Get a new <code>MTBContour2D</code> from the information of the
		 * <code>xmlContour</code>.
		 * 
		 * @param xmlContour
		 *          xml contour object
		 * 
		 * @return MTBContour2D object.
		 */
		public MTBContour2D getContour2DFromXml(MTBXMLContour2DType xmlContour) {

				MTBContour2D contour = new MTBContour2D();

				// get contour points
				MTBXMLPointVectorType plist = xmlContour.getPoints();
				for (int i = 0; i < plist.sizeOfPointArray(); i++) {
						MTBXMLPoint2DDoubleType p = plist.getPointArray(i);
						contour.addPixel((int) p.getX(), (int) p.getY());
				}
				// inner contours
				MTBXMLContour2DType[] inConts = xmlContour.getInnerArray();
				if (inConts != null && inConts.length > 0) {
						for (int i = 0; i < inConts.length; ++i) {
								MTBContour2D inCont = getContour2DFromXml(inConts[i]);
								contour.addInner(inCont);
						}
				}
				return contour;
		}
}
