/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 6030 $
 * $Date: 2012-09-24 14:27:37 +0200 (Mo, 24. Sep 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.morphology;

import ij.process.BinaryProcessor;
import ij.process.ByteProcessor;

import java.awt.geom.*;
import java.util.*;

import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException.OperatorExceptionType;


import de.unihalle.informatik.Alida.annotations.ALDAOperator;
import de.unihalle.informatik.Alida.annotations.ALDAOperator.ExecutionMode;
import de.unihalle.informatik.Alida.annotations.ALDAOperator.Level;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.annotations.Parameter.Direction;
import de.unihalle.informatik.MiToBo.core.datatypes.*;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImageByte;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.core.operator.*;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.ContourOnLabeledComponents;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.ContourOnLabeledComponents.ContourType;
import de.unihalle.informatik.MiToBo.segmentation.regions.labeling.*;

/**
 * Methods for post-processing components/pixels in binary images.
 * <p>
 * Thresholding algorithms yield binary images with distinct connected
 * components of foreground pixels, i.e. regions. The routines of this class
 * provide functionality for post-processing such binary (unlabeled) component
 * images. Exemplary post-processing steps might include linking/merging
 * adjacent components or changing a component's morphological shape.
 * <p>
 * Note that the different processing modes may require different parameter 
 * settings. These settings have to be done explicitly by calling the 
 * corresponding setter-functions the class provides. If the parameters are not
 * set by the user default values are used.
 * 
 * @author moeller
 */
@ALDAOperator(genericExecutionMode=ExecutionMode.ALL,level=Level.STANDARD)

public class ComponentPostprocess extends MTBOperator {

	/**
	 * Processing mode identifiers.
	 * 
	 * @author moeller
	 */
	public static enum ProcessMode {
		/**
		 * Delete components below a size threshold from the image.
		 * <p>
		 * Set threshold with method @link{setMinimalComponentSize}.
		 */
		ERASE_SMALL_COMPS,
		/**
		 * Link adjacent components with a line if distance below threshold.
		 * <p>
		 * Set distance with method @link{setMaximalComponentDistance}.
		 */
		LINK_ADJ_COMPS,
		/**
		 * Link adjacent pixels if distance below threshold.
		 * <p>
		 * Set distance with method @link{setMaximalComponentDistance}.
		 */
		LINK_ADJ_PIXELS,
		/**
		 * Perform Voronoi expansion for pixels close to components.
		 * <p>
		 * Set dilation mask size with 
		 * method @link{setMaximalVoronoiExpansionDistance}.
		 */
		VORONOI_EXPAND,
		/**
		 * Delete components with a rounded shape instead of an sustained shape.
		 * The roundness threshold is given by the user and is in range [0,1], 
		 * 1 means a sustained region.
		 * <p>
		 * Set threshold with method @link{setRoundnessThreshold}.
		 */
		ERASE_ROUND_COMPS
	}

	/**
	 * Binary input image.
	 */
	@Parameter( label= "Binary Input Image", required = true, dataIOOrder = -10,
			direction = Direction.IN,	description = "Binary input image")
	private transient MTBImage inImg = null;

	/**
	 * Binary result image.
	 */
	@Parameter( label= "Result Image", required = true, 
			direction = Direction.OUT, description = "Result image")
	private transient MTBImageByte resultImg = null;

	/**
	 * Minimal size of components in mode ERASE_SMALL_COMPS.
	 */
	@Parameter( label= "Min. Component Size", required = false, 
			direction = Direction.IN,	dataIOOrder = -8,
			description = "Minimal component size, smaller ones are deleted.")
	private int minCompSize = 0;

	/**
	 * Maximal distance of adjacent components.
	 */
	@Parameter( label= "Max. Distance of Componens", required = false, 
			direction = Direction.IN, dataIOOrder = -6,
			description = "Maximum allowed distance of components to be linked.")
	private int maxCompDist = 0;

	/**
	 * Voronoi expansion distance.
	 */
	@Parameter( label= "Max. Voronoi Expansion Distance", required = false, 
			direction = Direction.IN, dataIOOrder = -4,
			description = "Distance to which components are expanded in " +
					"Voronoi expansion mode.")
	private int maxVoronoiExpDist = 0;

	/**
	 * Flag for using 8-neighborhood.
	 */
	@Parameter( label= "Use Diagonal Neighbors?   ", required = false, 
			direction = Direction.IN, dataIOOrder = -10,
			description = "Flag for considering 8-neighborhood instead of 4-NB")
	private boolean diagonalNeighbors = false;

	/**
	 * Roundness threshold for components.
	 */
	@Parameter( label= "Roundness Threshold", required = false, 
			direction = Direction.IN,	dataIOOrder = -2,
			description = "Threshold for removing round components, " + 
				" larger values enforce greater eccentricity of surviving components.")
	private double roundnessThreshold = 0.0;

	/**
	 * Processing mode.
	 */
	@Parameter( label= "Process Mode", required = true, direction = Direction.IN, 
			description = "Process Modus", dataIOOrder = -8)
	private ProcessMode processMode = null; 

	/**
	 * Default constructor.
	 * 
	 * @throws ALDOperatorException
	 */
	public ComponentPostprocess() throws ALDOperatorException {
		// nothing to do here
	}

	/**
	 * Default constructor. 
	 * <p>
	 * Note that parameters for the chosen processing mode have to be
	 * set explicitly. Parameters for other modes are ignored.
	 *  
	 * @param img
	 *          Image to work on.
	 * @param pm
	 *          Process mode of operator.
	 *          
	 * @throws ALDProcessingDAGException
	 */
	public ComponentPostprocess(MTBImage img, ProcessMode pm)
	throws ALDOperatorException {
		this.inImg = img;
		this.processMode = pm;
	}

	/**
	 * Default constructor. A new empty meta data object is initialized.
	 * 
	 * @param img
	 *          Image to work on.
	 * @param pm
	 *          Process mode of operator.
	 * @param _minCompSize
	 *          Minimal region size for eliminating small comps.
	 * @param _maxCompDist
	 *          Maximal region distance in linking.
	 * @param maxVoroExpandDist
	 *          Max. distance of pixel considered in Voronoi expansion.
	 * 
	 *          Note: Depending on the process mode each time only one of the
	 *          parameters is actually in use.
	 * @throws ALDProcessingDAGException
	 */
	@Deprecated
	public ComponentPostprocess(MTBImage img, ProcessMode pm,
			int _minCompSize, int _maxCompDist, int maxVoroExpandDist)
	throws ALDOperatorException {
		this.inImg = img;
		this.processMode = pm;
		this.minCompSize = _minCompSize;
		this.maxCompDist = _maxCompDist;
		this.maxVoronoiExpDist = maxVoroExpandDist;
	}

	/**
	 * Default constructor. A new empty meta data object is initialized.
	 * 
	 * @param inImg
	 *          Image to work on.
	 * @param pm
	 *          Process mode of operator.
	 * @param minCompSize
	 *          Minimal region size for eliminating small comps.
	 * @param maxCompDist
	 *          Maximal region distance in linking.
	 * @param maxVoronoiExpandDist
	 *          Max. distance of pixel considered in Voronoi expansion.
	 * @param roundnessThreshold
	 *          Threshold for round objects in range [0,1]. A sustained region has
	 *          eccentricity 1, a round region has eccentricity 0.
	 * 
	 *          Note: Depending on the process mode each time only one of the
	 *          parameters is actually in use.
	 * @throws ALDProcessingDAGException
	 */
	@Deprecated
	public ComponentPostprocess(MTBImage img, ProcessMode pm,
			int _minCompSize, int _maxCompDist, int maxVoroExpandDist,
			double _roundnessThreshold) throws ALDOperatorException {
		this.inImg = img;
		this.processMode = pm;
		this.minCompSize = _minCompSize;
		this.maxCompDist = _maxCompDist;
		this.maxVoronoiExpDist = maxVoroExpandDist;
		this.roundnessThreshold = _roundnessThreshold;
	}

	/**
	 * Set minimal component size for erasing small components.
	 * <p>
	 * Only used in mode 'ERASE_SMALL_COMPS'.
	 */
	public void setMinimalComponentSize(int _minCompSize) { 
		this.minCompSize = _minCompSize;
	}

	/**
	 * Set maximal component/pixel distance for linking.
	 * <p>
	 * Only used in modes 'LINK_ADJ_COMPS' and 'LINK_ADJ_PIXELS'.
	 */
	public void setMaximalComponentDistance(int _maxCompDist) { 
		this.maxCompDist = _maxCompDist;
	}

	/**
	 * Set radius of dilation mask in Voronoi expansion.
	 * <p>
	 * Only used in mode 'VORONOI_EXPAND'.
	 */
	public void setMaximalVoronoiExpansionDistance(int _maxVoroExpandDist) {
		this.maxVoronoiExpDist = _maxVoroExpandDist;
	}

	/**
	 * Set roundness threshold for eliminating round components.
	 * <p>
	 * Only used in mode 'ERASE_ROUND_COMPS'.
	 */
	public void setRoundnessThreshold(double _roundnessThreshold) {
		this.roundnessThreshold = _roundnessThreshold;
	}

	/**
	 * If true 8-neighborhood, i.e. diagonal neighbors, will be used in 
	 * mode 'VORONOI_EXPAND' for Voronoi dilation.
	 */
	public void setDiagonalNeighbors(boolean value) {
		this.diagonalNeighbors = value;
	}

	/**
	 * Get reference to the current input image.
	 * 
	 * @return Input image to work on.
	 */
	public MTBImage getInputImage() {
		return this.inImg;
	}

	/**
	 * Get current process mode.
	 * 
	 * @return Process mode of the component process.
	 */
	public ProcessMode getProcessMode() {
		return this.processMode;
	}

	/**
	 * Returns the minimal valid component size for component removal.
	 */
	public int getMinimalComponentSize() {
		return this.minCompSize;
	}

	/**
	 * Returns the maximal component distance for linking.
	 */
	public int getMaximalComponentDist() {
		return this.maxCompDist;
	}

	/**
	 * Returns the maximal distance of pixels considered in Voronoi expansion.
	 */
	public int getMaxVoronoiExpandDist() {
		return this.maxVoronoiExpDist;
	}

	/**
	 * Get value of Parameter argument DiagonalNeighbors.
	 * 
	 * @return value of DiagonalNeighbors
	 */
	public boolean getDiagonalNeighbors() {
		return this.diagonalNeighbors;
	}

	/**
	 * Returns the currently active roundness threshold.
	 * 
	 * @throws ALDOperatorException
	 */
	public double getRoundnessThreshold() {
		return this.roundnessThreshold;
	}

	/**
	 * Get the result image after applying operator. Attention, reference might be
	 * null.
	 * 
	 * @return Reference to result image.
	 */
	public MTBImageByte getResultImage() {
		return this.resultImg;
	}

	// input image width and height
	private int width;
	private int height;

	// temporary variables needed during internal calculations
	private Vector<Vector<Point2D.Double>> endPointHash;

	@Override
	protected void operate() throws ALDOperatorException,
	ALDProcessingDAGException {

		MTBImageByte byteImg;
		if (this.inImg.getType().equals(MTBImage.MTBImageType.MTB_BYTE))
			byteImg = (MTBImageByte)this.inImg;
		else
			byteImg = (MTBImageByte)this.inImg.convertType(
																			MTBImage.MTBImageType.MTB_BYTE, true);

		MTBImageByte result = null;
		switch (this.getProcessMode()) {
		case ERASE_SMALL_COMPS:
			result= this.EraseSmallComponents(byteImg, 
																				this.getMinimalComponentSize());
			break;
		case LINK_ADJ_COMPS:
			result= this.LinkAdjacentComponents(byteImg);
			break;
		case LINK_ADJ_PIXELS:
			result= this.LinkAdjacentPixels(byteImg, this.getMaximalComponentDist());
			break;
		case VORONOI_EXPAND:
			result= this.VoronoiExpandComponents(byteImg, 
																						this.getMaxVoronoiExpandDist());
			break;
		case ERASE_ROUND_COMPS:
			result= this.EraseRoundComponents(byteImg, this.getRoundnessThreshold());
			break;
		}
		this.resultImg = result;

		if (result == null)
			throw new ALDOperatorException(OperatorExceptionType.OPERATE_FAILED,
					"ComponentPostprocess.operate() failed: " + "Unknown process mode "
					+ this.getProcessMode());
	}

	/**
	 * Function to remove small components from a binary image.
	 * <p>
	 * Small components are components that have a
	 * size below the given threshold. Suitable for noise reduction in binary
	 * images.
	 * 
	 * @param binIP
	 *          binary input image
	 * @param minComponentSize
	 *          minimum size of valid components
	 * @return image with too small components erased
	 * @throws ALDOperatorException
	 */
	protected MTBImageByte EraseSmallComponents(MTBImageByte binIP,
			int minComponentSize) throws ALDOperatorException,
			ALDProcessingDAGException {

		this.width = binIP.getSizeX();
		this.height = binIP.getSizeY();

		// initialize result image
		MTBImageByte resultImage = (MTBImageByte) MTBImage.createMTBImage(
				this.width, this.height, 1, 1, 1, MTBImageType.MTB_BYTE);
		resultImage.fillWhite();

		// copy input image data
		resultImage.setImagePart(binIP, 0, 0, 0, 0, 0);

		// label components in binary image
		LabelComponentsSequential regionLabler = new LabelComponentsSequential(
				binIP, getDiagonalNeighbors());
		regionLabler.runOp(false);
		MTBRegion2DSet binRegions = regionLabler.getResultingRegions();

		for (int i = 0; i < binRegions.size(); ++i) {
			MTBRegion2D r = binRegions.elementAt(i);
			if (r.getArea() < minComponentSize) {
				// iterate over all region points...
				Vector<Point2D.Double> points = r.getPoints();
				for (int j = 0; j < points.size(); ++j) {

					int x = (int) points.elementAt(j).x;
					int y = (int) points.elementAt(j).y;

					resultImage.putValueInt(x, y, 0);
				}
			}
		}
		return resultImage;
	}

	/**
	 * Remove components with small eccentricity.
	 * 
	 * @param binImage		Input image.
	 * @param threshold		Eccentricity threshold.
	 * @return	Result image.
	 * @throws ALDOperatorException
	 * @throws ALDProcessingDAGException
	 */
	protected MTBImageByte EraseRoundComponents(MTBImageByte binImage,
			double threshold) throws ALDOperatorException, ALDProcessingDAGException {
		this.width = binImage.getSizeX();
		this.height = binImage.getSizeY();
		// initialize result image
		MTBImageByte resultImage = (MTBImageByte) MTBImage.createMTBImage(
				this.width, this.height, 1, 1, 1, MTBImageType.MTB_BYTE);
		resultImage.fillBlack();
		// label components in binary image
		LabelComponentsSequential regionLabler = new LabelComponentsSequential(
				binImage, getDiagonalNeighbors());
		regionLabler.runOp(false);
		MTBRegion2DSet binRegions = regionLabler.getResultingRegions();
		for (int i = 0; i < binRegions.size(); ++i) {
			MTBRegion2D r = binRegions.elementAt(i);
			if (r.getEccentricity() >= threshold) {
				// iterate over all region points...
				Vector<Point2D.Double> points = r.getPoints();
				for (int j = 0; j < points.size(); ++j) {
					int x = (int) points.elementAt(j).x;
					int y = (int) points.elementAt(j).y;
					resultImage.putValueInt(x, y, 255);
				}
			}
		}
		return resultImage;
	}

	/**
	 * Function for linking adjacent components/regions in a binary image.
	 * <p>
	 * This function processes a binary image with foreground objects being white.
	 * Based on an initial component labeling step and subsequent skeletonization
	 * of resulting components, adjacent components are linked by a line of width
	 * one pixel in the result image.
	 * <p>
	 * The criterion for two components to be adjacent is defined based on the
	 * distance of pairs of skeleton endpoints of both components. If there exists
	 * at least one pair of endpoints between two components with a distance below
	 * the given threshold, the components are linked. If there is more than one
	 * pair of such points, the pair with the smallest distance is selected to be
	 * linked to each other.
	 * 
	 * @param binImage
	 *          input binary image with regions
	 * @return binary image where adjacent components are linked by a line of
	 *         width 1 pixel
	 * @throws ALDOperatorException
	 */
	protected MTBImageByte LinkAdjacentComponents(MTBImageByte binIP) 
		throws ALDOperatorException, ALDProcessingDAGException {

		this.width = binIP.getSizeX();
		this.height = binIP.getSizeY();

		// initialize result image
		MTBImageByte resultImage = (MTBImageByte) MTBImage.createMTBImage(
				this.width, this.height, 1, 1, 1, MTBImageType.MTB_BYTE);
		resultImage.fillWhite();

		// copy input image data
		resultImage.setImagePart(binIP, 0, 0, 0, 0, 0);

		// label components in binary image
		LabelComponentsSequential regionLabler = new LabelComponentsSequential(
				binIP, getDiagonalNeighbors());
		regionLabler.runOp(false);
		MTBRegion2DSet binRegions = regionLabler.getResultingRegions();

		// calculate skeletons of initial binary regions,
		// skeletonize function requires inverse input image (object = black)
		ByteProcessor bP = new ByteProcessor(this.width, this.height);
		for (int y = 0; y < this.height; ++y) {
			for (int x = 0; x < this.width; ++x) {
				if (binIP.getValueInt(x, y) > 0)
					bP.putPixel(x, y, 0);
				else
					bP.putPixel(x, y, 255);
			}
		}
		BinaryProcessor bbP = new BinaryProcessor(bP);
		bbP.skeletonize();

		// array where endpoints are marked
		int endPointArray[][] = new int[this.width][this.height];

		// pseudo-hashtable with all endpoints of binary regions:
		// remember the endpoints of all regions, indexed by region index
		this.endPointHash = new Vector<Vector<Point2D.Double>>();
		for (int i = 0; i < binRegions.size(); ++i) {
			this.endPointHash.add(new Vector<Point2D.Double>());
		}

		// find all endpoints of all binary regions
		for (int i = 0; i < binRegions.size(); ++i) {
			MTBRegion2D r = binRegions.elementAt(i);
			Vector<Point2D.Double> points = r.getPoints();

			// iterate over all region points...
			for (int j = 0; j < points.size(); ++j) {

				int x = (int) points.elementAt(j).x;
				int y = (int) points.elementAt(j).y;

				// found a point of the skeleton?
				if (bbP.getPixel(x, y) == 0) {

					// if so, count neighboring skeleton points
					int nCount = 0;
					for (int dx = -1; dx <= 1; dx++) {
						for (int dy = -1; dy <= 1; ++dy) {
							if (dx == 0 && dy == 0)
								continue;
							if (bbP.getPixel(x + dx, y + dy) == 0)
								nCount++;
						}
					}
					// if there is only one neighbor in 8-NB, (x,y) is endpoint
					if (nCount == 1) {
						endPointArray[y][x] = 1;
						// binEndPoints.add(new Point2D.Double(x,y));

						// put endpoint in hash table
						this.endPointHash.elementAt(i).add(new Point2D.Double(x, y));
					}
				}
			}
		} // end of for-loop: all binary regions

		// iterate over all regions and link them to neighbor regions
		for (int i = 0; i < binRegions.size() - 1; ++i) {

			// get endpoints of this region
			Vector<Point2D.Double> regionEndPoints = this.endPointHash.elementAt(i);

			// find close endpoints in neighboring regions
			for (int j = i + 1; j < binRegions.size(); ++j) {

				// get endpoints of neighbor region
				Vector<Point2D.Double> neighborEndPoints = this.endPointHash
				.elementAt(j);

				double regionDist = this.maxCompDist;
				int closestPointID_region = -1;
				int closestPointID_neighbor = -1;

				for (int m = 0; m < regionEndPoints.size(); ++m) {
					double rx = regionEndPoints.elementAt(m).x;
					double ry = regionEndPoints.elementAt(m).y;

					for (int n = 0; n < neighborEndPoints.size(); ++n) {
						double nx = neighborEndPoints.elementAt(n).x;
						double ny = neighborEndPoints.elementAt(n).y;

						double dist = (rx - nx) * (rx - nx) + (ry - ny) * (ry - ny);

						// accept endpoints with distance below threshold
						if (dist < regionDist) {
							closestPointID_region = m;
							closestPointID_neighbor = n;
							regionDist = dist;
						}
					}
				}
				// check if regions are close to each other
				if (regionDist < this.maxCompDist) {
					// if so, draw line between the two endpoints
					Point2D.Double regPoint = regionEndPoints
					.elementAt(closestPointID_region);
					Point2D.Double neighbPoint = neighborEndPoints
					.elementAt(closestPointID_neighbor);
					resultImage.drawLine2D((int) regPoint.x, (int) regPoint.y,
							(int) neighbPoint.x, (int) neighbPoint.y, 255);
				}
			} // end of for-loop: iterate over all pairs of region endpoints
		} // end of for-loop: iterate over all regions

		return resultImage;

	} // end of function: LinkRegions(...)

	/**
	 * Function for linking adjacent pixels in a binary image.
	 * <p>
	 * This function processes a binary image containing only isolated pixels in
	 * the foreground. The foreground color is white. Based on the given maximal
	 * distance adjacent pixels are linked together to components. The result is
	 * comparable to a single-linkage clustering. As result an image is returned
	 * where adjacent pixels are linked to each other by a line of width one.
	 * <p>
	 * Note that the behaviour of the function is undefined if there are
	 * foreground objects present in the image other than single pixels!
	 * 
	 * @param binImage
	 *          input binary image with regions
	 * @return binary image where adjacent pixels are linked by a line of width 1
	 *         pixel
	 * @throws ALDOperatorException
	 * 
	 * @todo Rückgabe ändern, es sollten Regionen oder so zurückkommen, damit man
	 *       auf Punkte pro Region schliessen kann...
	 */
	protected MTBImageByte LinkAdjacentPixels(MTBImageByte binIP, int maxDist)
	throws ALDOperatorException, ALDProcessingDAGException {

		this.width = binIP.getSizeX();
		this.height = binIP.getSizeY();

		// initialize result image
		MTBImageByte resultImage = (MTBImageByte) MTBImage.createMTBImage(
				this.width, this.height, 1, 1, 1, MTBImageType.MTB_BYTE);
		resultImage.fillWhite();

		// copy input image data
		resultImage.setImagePart(binIP, 0, 0, 0, 0, 0);

		// label components in binary image
		LabelComponentsSequential regionLabler = new LabelComponentsSequential(
				binIP, getDiagonalNeighbors());
		regionLabler.runOp(false);
		MTBRegion2DSet binRegions = regionLabler.getResultingRegions();

		// iterate over all regions and link them to neighbor regions
		for (int i = 0; i < binRegions.size() - 1; ++i) {

			int reg_ax = (int) binRegions.get(i).getCenterOfMass_X();
			int reg_ay = (int) binRegions.get(i).getCenterOfMass_Y();

			// find neighboring regions
			for (int j = i + 1; j < binRegions.size(); ++j) {

				int reg_bx = (int) binRegions.get(j).getCenterOfMass_X();
				int reg_by = (int) binRegions.get(j).getCenterOfMass_Y();

				double dist = (reg_ax - reg_bx) * (reg_ax - reg_bx) + (reg_ay - reg_by)
				* (reg_ay - reg_by);

				// accept endpoints with distance below threshold
				if (dist < maxDist * maxDist)
					resultImage.drawLine2D(reg_ax, reg_ay, reg_bx, reg_by, 255);
			}
		}
		return resultImage;
	} // end of function: LinkRegions(...)

	/**
	 * Dilate components, but avoid merges.
	 * 
	 * @param binImg		Input image.
	 * @param maxDist		Size of dilation mask.
	 * @return	Result image.
	 * @throws ALDOperatorException
	 * @throws ALDProcessingDAGException
	 */
	protected MTBImageByte VoronoiExpandComponents(MTBImageByte binImg,
			int maxDist) throws ALDOperatorException, ALDProcessingDAGException {

		// ImageProcessor binIP = binImg.getImagePlus().getProcessor();

		ImgDilate dilOp = new ImgDilate(binImg, maxDist * 2);
		dilOp.runOp(false);
		MTBImage dilateBinImg = dilOp.getResultImage();

		// pixels in dilated area
		MTBImage relPixels = MTBImage.createMTBImage(binImg.getSizeX(), binImg
				.getSizeY(), 1, 1, 1, MTBImageType.MTB_BYTE);
		for (int y = 0; y < binImg.getSizeY(); ++y) {
			for (int x = 0; x < binImg.getSizeX(); ++x) {
				if (dilateBinImg.getValueInt(x, y) > 0 && binImg.getValueInt(x, y) == 0)
					relPixels.putValueInt(x, y, 255);
			}
		}

		// component labeling
		LabelComponentsSequential regionLabler = new LabelComponentsSequential(
				binImg, getDiagonalNeighbors());
		regionLabler.runOp(false);
		MTBRegion2DSet comps = regionLabler.getResultingRegions();

		/**
		 * to do: Change ImagePlus and ImageProcessors to MTBImage objects. new
		 * version: changed by danny, using new ContourOnLabeledComponents class
		 */
		MTBImageByte binImgByte = (MTBImageByte) binImg.convertType(
				MTBImageType.MTB_BYTE, false);
		ContourOnLabeledComponents clc = new ContourOnLabeledComponents(binImgByte,
				comps, ContourType.OUTER_CONTOUR, 1);
		clc.runOp(false);
		MTBContour2DSet conts = clc.getResultContours();

		/**
		 * old version: changed by danny
		 */
		// Vector<Contour2D> conts=
		// Contour_on_Labeled_Components.get_ContoursOnRegions(binIP,comps);

		// check each pixel in the image
		for (int y = 0; y < binImg.getSizeY(); ++y) {
			for (int x = 0; x < binImg.getSizeX(); ++x) {

				// pixel is not new, i.e. was there before -> ignore it
				if (relPixels.getValueInt(x, y) == 0)
					continue;

				double minDist_l = 100000000;
				int minID = -1;

				// alle Konturen
				minDist_l = 100000000;
				for (int i = 0; i < conts.size(); ++i) {
					MTBContour2D c = conts.elementAt(i);
					Vector<Point2D.Double> pc = c.getPoints();

					// alle Punkte der Kontur
					double c_dist = 100000000;
					for (int k = 0; k < pc.size(); ++k) {
						Point2D.Double pk = pc.elementAt(k);
						double d = Math.sqrt((x - pk.getX()) * (x - pk.getX())
								+ (y - pk.getY()) * (y - pk.getY()));
						if (d < c_dist) {
							c_dist = d;
						}
					}
					if (c_dist < minDist_l) {
						minDist_l = c_dist;
						minID = i;
					}
				}
				if (minDist_l < maxDist && minID != -1)
					comps.elementAt(minID).addPixel(x, y);
			}
		}

		MTBImageByte resultImage = (MTBImageByte) MTBImage.createMTBImage(binImg
				.getSizeX(), binImg.getSizeY(), 1, 1, 1, MTBImageType.MTB_BYTE);
		for (int i = 0; i < comps.size(); ++i) {
			MTBRegion2D r = comps.elementAt(i);
			Vector<Point2D.Double> points = r.getPoints();
			for (int j = 0; j < points.size(); ++j) {
				Point2D.Double p = points.elementAt(j);
				resultImage.putValueInt((int) Math.round(p.getX()), (int) Math.round(p
						.getY()), i + 1);
			}
		}
		return resultImage;
	}
}
