/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5795 $
 * $Date: 2012-07-13 17:21:14 +0200 (Fr, 13. Jul 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.snakes.energies;

import java.awt.geom.Point2D;
import java.util.Vector;

import Jama.Matrix;



import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;
import de.unihalle.informatik.Alida.annotations.ALDParametrizedClass;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.energies.MTBActiveContourEnergy_CVRegionFit;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.exceptions.*;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.exceptions.MTBActiveContourException.ExceptionType;
import de.unihalle.informatik.MiToBo.segmentation.snakes.datatypes.MTBSnake;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingle;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingleGreedy;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingleVarCalc;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingle.EnergyNormalizationMode;

/**
 * Snake energy based on Chan-Vese region fitting.
 * <p>
 * This energy models the inner and outer regions of a snake by a Gaussian 
 * model, i.e. each region is specified by a mean intensity value and 
 * deviations from this value are penalized.
 * 
 * @author moeller
 */
@ALDDerivedClass
@ALDParametrizedClass

public class MTBSnakeEnergyCD_CVRegionFit 
	extends MTBActiveContourEnergy_CVRegionFit 
		implements MTBSnakeEnergyDerivable, MTBSnakeEnergyComputable {

	/**
	 * Interval boundaries of adapted energy interval.
	 * <p>
	 * energyRange[0] -> left boundary
	 * energyRange[1] -> right boundary
	 * energyRange[2] -> interval width
	 */
	private double [] energyRange = {0.0, 0.0, 0.0};
	
	/**
	 * Width of working image.
	 */
	private transient int iWidth;
	
	/**
	 * Height of working image. 
	 */
	private transient int iHeight;
	
	/**
	 * Mode of normalization.
	 */
	protected EnergyNormalizationMode normMode = 
		EnergyNormalizationMode.NORM_BALANCED_DERIVATIVES;
	
	/**
	 * Scaling factor to rescale image coordinates in range [0,1] or several
	 * ranges to the original coordinates range like [1000, 1000] in a image of
	 * size 1000 x 1000. Default scaling factor is 1.
	 */
	protected transient double scaleFactor = 1.0;

	/**
	 * Current snake the energy is based on.
	 */
	private transient MTBSnake currentSnk;
	
	/**
	 * Global exclude mask for image to be segmented.
	 */
	private transient boolean [][] excludeMask;
	
	/**
	 * Default constructor.
	 */
	public MTBSnakeEnergyCD_CVRegionFit() {
		// nothing to do here
	}
	
  /**
   * Default constructor.
   * 
   * @param im	Image to work on.
   * @param l_in	Weighting factor for inner region fit.
   * @param l_out	Weighting factor for outer region fit.
   * @throws ALDOperatorException 
   * @throws ALDProcessingDAGException 
   */
  public MTBSnakeEnergyCD_CVRegionFit(MTBImage im, double lin, double lout) {
  	super(im,lin,lout);
  }
  
	@Override
  public boolean initEnergy(SnakeOptimizerSingleVarCalc opt) 
		throws MTBSnakeException {
		if (this.inImg == null) {
			// get normalized (!) image from snake optimizer
			this.inImg = opt.getInputImage();
		}
		else {
			// make sure that the input image intensities are normalized
			this.inImg = opt.normalizeInputImage(this.inImg);
		}
		try {
	    super.init();
    } catch (MTBActiveContourException e) {
    	throw new MTBSnakeException(ExceptionType.INITIALIZATION_ERROR,
    			"MTBSnakeEnergyCD_CVRegionFit - " +
    			"init failed! Reason: " + e.getCommentString());
    }
  	this.iWidth = this.inImg.getSizeX();
  	this.iHeight = this.inImg.getSizeY();
  	this.normMode = opt.getNormalizationMode();
		opt.requireCounterClockwiseSnakePointOrder();
		opt.requireOverlapMask();
		// get energy minimum and maximum;
		// minimum of Chan-Vese energy is always smaller than zero and 
		// maximum always larger
		double eMin = this.getEnergyDerivativeMinVal();
		double eMax = this.getEnergyDerivativeMaxVal();
		if (Math.abs(eMin) > Math.abs(eMax)) {
			this.energyRange[0] = - Math.abs(eMin);
			this.energyRange[1] =   Math.abs(eMin);
		}
		else {
			this.energyRange[0] = - Math.abs(eMax);
			this.energyRange[1] =   Math.abs(eMax);
		}
		this.energyRange[2] = this.energyRange[1] - this.energyRange[0];
//		// make sure that all parameters are initialized correctly
//		this.updateStatus(opt);
	  return true;
  }

	@Override
  public boolean initEnergy(SnakeOptimizerSingleGreedy opt) {
//		if (this.inImg == null)
//			this.inImg = opt.getInputImage();
//		try {
//	    super.init();
//    } catch (ALDOperatorException e) {
//	    // TODO Auto-generated catch block
//	    e.printStackTrace();
//    }
//  	this.iWidth = this.inImg.getSizeX();
//  	this.iHeight = this.inImg.getSizeY();
//		opt.requireCounterClockwiseSnakePointOrder();
//		opt.requireOverlapMask();
//		// get energy minimum and maximum;
//		// minimum of Chan-Vese energy is always smaller than zero and 
//		// maximum always larger
//		double eMin = this.getEnergyMinVal().doubleValue();
//		double eMax = this.getEnergyMaxVal().doubleValue();
//		if (Math.abs(eMin) > Math.abs(eMax)) {
//			this.energyRange[0] = - Math.abs(eMin);
//			this.energyRange[1] =   Math.abs(eMin);
//		}
//		else {
//			this.energyRange[0] = - Math.abs(eMax);
//			this.energyRange[1] =   Math.abs(eMax);
//		}
//		this.energyRange[2] = this.energyRange[1] - this.energyRange[0];
//	  return true;
		return false;
  }

	@Override
  public void updateStatus(SnakeOptimizerSingleVarCalc opt) 
  	throws MTBSnakeException {
		this.currentSnk = opt.getCurrentSnake();
		this.excludeMask = opt.getExcludeMask();
		// set excluded pixels invalid - if there are some...
    boolean [][] vMask = new boolean[this.iHeight][this.iWidth];
    if (this.excludeMask != null) {
    	for (int y=0;y<this.iHeight;++y) {
    		for (int x=0;x<this.iWidth;++x) {
    			vMask[y][x] = !this.excludeMask[y][x];
    		}
    	}
    }
    else {
    	// all pixels are valid, non is excluded
    	for (int y=0;y<this.iHeight;++y) {
    		for (int x=0;x<this.iWidth;++x) {
    			vMask[y][x] = true;
    		}
    	}    	
    }
		// fetch current overlap mask
    if (opt.getSnakeNumber() > 1) {
    	int[][] overlapMask = opt.getOverlapMask();
    	int[][] snakeMask;
    	snakeMask = this.currentSnk.getBinaryMask(opt.getInputImage().getSizeX(),
    			opt.getInputImage().getSizeY());	    
    	for (int y=0;y<this.iHeight;++y) {
    		for (int x=0;x<this.iWidth;++x) {
    			if (snakeMask[y][x] == 0 && overlapMask[y][x] != 0)
    				vMask[y][x] = false;
    		}
    	}
    }
		this.currentSnk.setValidmask(vMask);
		// do the actual update
		try {
	    this.updateParameters(this.currentSnk);
    } catch (MTBActiveContourException e) {
    	throw new MTBSnakeException(ExceptionType.INITIALIZATION_ERROR,
    			"MTBSnakeEnergyCD_CVRegionFit - " +
    			"update failed! Reason: " + e.getCommentString());
    }
	}
	
	@Override
  public void updateStatus(SnakeOptimizerSingleGreedy opt) {
		// nothing to do here
	}
	
	/**
   * Updates the fitting term based on new region average values.
   */
	@Override
  public Matrix getDerivative_MatrixPart(SnakeOptimizerSingleVarCalc opt) {
		
		int snakePointNum = this.currentSnk.getPointNum();
		Vector<Point2D.Double> polyPoints = this.currentSnk.getPoints();
		
		Matrix A = new Matrix(snakePointNum * 2, snakePointNum * 2);
		for (int counter = 0; counter < snakePointNum - 1; ++counter) {
			Point2D.Double p = polyPoints.get(counter);
			double eDerive = this.getDerivative(this.currentSnk,
					                                (int)(p.x*this.scaleFactor), 
										 	 			 					    (int)(p.y*this.scaleFactor),0);
			double tau_i = 0.0;
			switch(this.normMode)
			{
			case NORM_BALANCED_DERIVATIVES:
				// calculate normalized final derivative value
				tau_i = (eDerive - this.energyRange[0])/this.energyRange[2]                                              
						* MTBSnakeEnergyDerivable.targetEnergyRange[2] 
								+ MTBSnakeEnergyDerivable.targetEnergyRange[0];
				break;
			case NORM_NONE:
			default:
				tau_i = eDerive;
				break;
			}
//			double oldVal= A.get(counter, snakePointNum + counter);
			A.set(counter, snakePointNum + counter, - tau_i);
//			oldVal= A.get(counter, snakePointNum + counter + 1);
			A.set(counter, snakePointNum + counter + 1, tau_i);
//			oldVal= A.get(snakePointNum + counter, counter);
			A.set(snakePointNum + counter, counter, tau_i);
//			oldVal= A.get(snakePointNum + counter, counter + 1);
			A.set(snakePointNum + counter, counter + 1, - tau_i);
		}
		// add last point
		Point2D.Double p = polyPoints.get(snakePointNum - 1);
		double eDerive = this.getDerivative(this.currentSnk,
				                                (int)(p.x*this.scaleFactor), 
				                                (int)(p.y*this.scaleFactor), 0);
		double tau_i = 0.0;
		switch(this.normMode)
		{
		case NORM_BALANCED_DERIVATIVES:
			// calculate normalized final derivative value
			tau_i = (eDerive - this.energyRange[0])/this.energyRange[2]                                              
					* MTBSnakeEnergyDerivable.targetEnergyRange[2] 
							+ MTBSnakeEnergyDerivable.targetEnergyRange[0];
			break;
		case NORM_NONE:
		default:
			tau_i = eDerive;
			break;
		}
		// fill matrix - upper right block
//		double oldVal = A.get(snakePointNum - 1, snakePointNum);
		A.set(snakePointNum - 1, snakePointNum, tau_i);
//		oldVal = A.get(snakePointNum - 1, snakePointNum * 2 - 1);
		A.set(snakePointNum - 1, snakePointNum * 2 - 1, - tau_i);
		// fill matrix - lower left block
//		oldVal = A.get(2 * snakePointNum - 1, 0);
		A.set(2 * snakePointNum - 1, 0, - tau_i);
//		oldVal = A.get(2 * snakePointNum - 1, snakePointNum - 1);
		A.set(2 * snakePointNum - 1, snakePointNum - 1, tau_i);
		return A;
	}

	@Override
  public Matrix getDerivative_VectorPart(SnakeOptimizerSingleVarCalc opt) {
	  return null;
  }

  @Override
  public double calcEnergy(SnakeOptimizerSingle opt) {
  	// get current snake
  	MTBSnake snake= opt.getCurrentSnake();
  	// calculate energy
  	return 1.0/this.energyRange[1]*super.calcEnergy(snake);
  }

	/**
	 * Set the scaling factor.
	 * 
	 * @param s
	 *          new scaling factor.
	 */
	@Override
  public void setScaleFactor(double s) {
		this.scaleFactor = s;
	}

	/**
	 * Get scaling factor.
	 * 
	 * @return Scaling factor.
	 */
	@Override
  public double getScaleFactor() {
		return this.scaleFactor;
	}

	@Override
  public String toString() {
		return "MTBSnakeEnergyCD_CVRegionFit: Chan-Vese Region Fitting "  
				+ "(lambda_in = " + this.lambda_in + 
					" , lamdba_out = " + this.lambda_out + ")";
  }
}
