/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2011
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5795 $
 * $Date: 2012-07-13 17:21:14 +0200 (Fr, 13. Jul 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.snakes.energies;

import java.awt.geom.Point2D;
import java.util.Vector;

import Jama.Matrix;

import de.unihalle.informatik.Alida.annotations.ALDClassParameter;
import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;
import de.unihalle.informatik.Alida.annotations.ALDParametrizedClass;
import de.unihalle.informatik.Alida.annotations.Parameter.ExpertMode;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.energies.MTBActiveContourEnergy_MeanSep;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.exceptions.*;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.exceptions.MTBActiveContourException.ExceptionType;
import de.unihalle.informatik.MiToBo.segmentation.snakes.datatypes.MTBSnake;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingle;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingle.EnergyNormalizationMode;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingleGreedy;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingleVarCalc;

/**
 * Snake energy based on region mean separation.
 * <p>
 * This energy models tries to best separate the average intensity
 * values of the inner and outer regions of a snake. For details refer to:
 * 
 * Yezzi, Tsai, Willsky, "A Fully Global Approach to Image Segmentation via 
 * Coupled Curve Evolution Equations", JVCIR, 13, 195-216 (2002)  
 * <p>
 * Note that there are different normalization modes implemented in this 
 * class. In mode 'linear' energy values are linearly mapped to the range 
 * [-1,1] with the maximal/minimal energy values as boundaries. Note that this 
 * often results in quite small energy values. In mode 'linear_cropped ' energy 
 * values are also mapped linearly, however, an upper/lower bound can be 
 * specified above/below the values are cropped.
 * 
 * @author moeller
 */


@ALDDerivedClass
@ALDParametrizedClass
public class MTBSnakeEnergyCD_MeanSep 
	extends MTBActiveContourEnergy_MeanSep
		implements MTBSnakeEnergyDerivable, MTBSnakeEnergyComputable {

	/**
	 * Modes of energy value transformation.
	 */
	public static enum EnergyMapping {
		/**
		 * Energy values are mapped linearly into the target interval.
		 */
		LINEAR,
		/**
		 * Energy values are mapped linearly into the target interval, but in 
		 * addition boundaries for valid energy values can be specified which 
		 * are mapped to -1 and 1. Values larger or smaller than the given energy 
		 * boundaries are cropped. This is to restrict energy values to a 
		 * reasonable interval and avoid crucial numerical behaviour. 
		 */
//		LINEAR_CROPPED
	}

	/**
	 * Energy value normalization to be used.
	 */
	@ALDClassParameter(label = "Mapping Mode", 
			mode = ExpertMode.ADVANCED, dataIOOrder = -10)
	protected EnergyMapping mapMode = EnergyMapping.LINEAR;
	
//	@ALDClassParameter(label="Maximum Energy Value")
	protected double maxEnergyBound;
	
	/**
	 * Mode of normalization.
	 */
	protected EnergyNormalizationMode normMode = 
		EnergyNormalizationMode.NORM_BALANCED_DERIVATIVES;
	
	/**
	 * Interval boundaries of adapted energy interval.
	 * <p>
	 * energyRange[0] -> left boundary
	 * energyRange[1] -> right boundary
	 * energyRange[2] -> interval width
	 */
	private double [] energyRange = {0.0, 0.0, 0.0};

	/**
	 * Scaling factor to rescale image coordinates in range [0,1] or several
	 * ranges to the original coordinates range like [1000, 1000] in a image of
	 * size 1000 x 1000. Default scaling factor is 1.
	 */
	protected transient double scaleFactor = 1.0;

	/**
	 * Current snake the energy is based on.
	 */
	private transient MTBSnake currentSnk;
	
	/**
	 * Global exclude mask for image to be segmented.
	 */
	private boolean [][] excludeMask;
	
	/**
	 * Default constructor.
	 */
	public MTBSnakeEnergyCD_MeanSep() {
		super();
	}
	
  /**
   * Default constructor.
   * @param im	Image to work on.
   * @throws ALDProcessingDAGException 
   */
  public MTBSnakeEnergyCD_MeanSep(MTBImage im) {
  	this.inImg = im;
  	this.mapMode = EnergyMapping.LINEAR;
//		this.enerMinVal = this.getEnergyMinVal().doubleValue();
  }

  /**
   * Constructor with energy value mapping mode.
   * @param im	Image to work on.
   * @param m		Mode for mapping energy values to interval [-1,1].
   */
  public MTBSnakeEnergyCD_MeanSep(MTBImage im, EnergyMapping m) {
  	this.inImg = im;
  	this.mapMode = m;
//  	this.enerMinVal = this.getEnergyMinVal().doubleValue();
  }
	
  /**
   * Constructor with energy value mapping mode and maximal energy bound.
   * @param im	Image to work on.
   * @param m		Mode for mapping energy values to interval [-1,1].
   * @param maxEnergBound	Maximal valid value for energies, must be >= 0!
   */
  public MTBSnakeEnergyCD_MeanSep(MTBImage im, EnergyMapping m, 
  		double _maxEnergBound) {
  	this.inImg = im;
  	this.mapMode = m;
  	this.maxEnergyBound = _maxEnergBound;
//  	if (maxEnergBound < 0)
//  		this.enerMinVal = this.getEnergyMinVal().doubleValue();
//  	else
//  		this.enerMinVal = -maxEnergBound;
  }

  /**
	 * Set the scaling factor.
	 * @param s	New scaling factor.
	 */
	@Override
  public void setScaleFactor(double s) {
		this.scaleFactor = s;
	}

	/**
	 * Get scaling factor.
	 * @return Scaling factor.
	 */
	@Override
  public double getScaleFactor() {
		return this.scaleFactor;
	}

	@Override
  public boolean initEnergy(SnakeOptimizerSingleVarCalc opt) 
  	throws MTBSnakeException {
		if (this.inImg == null) {
			// get normalized (!) image from snake optimizer
			this.inImg = opt.getInputImage();
		}
		else {
			// make sure that the input image intensities are normalized
			this.inImg = opt.normalizeInputImage(this.inImg);
		}
		try {
	    super.init();
    } catch (MTBActiveContourException e) {
    	throw new MTBSnakeException(ExceptionType.INITIALIZATION_ERROR,
    			"MTBSnakeEnergyCD_MeanSep - " +
    			"init failed! Reason: " + e.getCommentString());
    }
		this.normMode = opt.getNormalizationMode();
		opt.requireCounterClockwiseSnakePointOrder();
		// get energy minimum and maximum of energy values
		double eMin = this.getEnergyDerivativeMinVal();
		double eMax = this.getEnergyDerivativeMaxVal();
		if (Math.abs(eMin) > Math.abs(eMax)) {
			this.energyRange[0] = - Math.abs(eMin);
			this.energyRange[1] =   Math.abs(eMin);
		}
		else {
			this.energyRange[0] = - Math.abs(eMax);
			this.energyRange[1] =   Math.abs(eMax);
		}
		this.energyRange[2] = this.energyRange[1] - this.energyRange[0];

//		if (this.maxEnergyBound < 0)
//			this.enerMinVal = this.getEnergyDerivativeMinVal();
//		else
//			this.enerMinVal = -this.maxEnergyBound;
		return true;
  }
	
	@Override
  public boolean initEnergy(SnakeOptimizerSingleGreedy opt) {
//		this.enerMinVal = this.getEnergyMinVal().doubleValue();
//		if (this.maxEnergyBound < 0)
//			this.enerMinVal = this.getEnergyMinVal().doubleValue();
//		else
//			this.enerMinVal = -this.maxEnergyBound;
//		opt.requireCounterClockwiseSnakePointOrder();
//		this.enerRange = -2 * this.enerMinVal;
//		return true;
		return false;
  }

	@Override
  public void updateStatus(SnakeOptimizerSingleVarCalc opt) 
  	throws MTBSnakeException {
		this.currentSnk = opt.getCurrentSnake();
		this.excludeMask = opt.getExcludeMask();
		// set excluded pixels invalid
    boolean [][] vMask = new boolean[this.iHeight][this.iWidth];
    for (int y=0;y<this.iHeight;++y) {
    	for (int x=0;x<this.iWidth;++x) {
    		vMask[y][x] = !this.excludeMask[y][x];
    	}
    }
		this.currentSnk.setValidmask(vMask);
		// do the actual update
		try {
	    this.updateParameters(this.currentSnk);
    } catch (MTBActiveContourException e) {
    	throw new MTBSnakeException(ExceptionType.INITIALIZATION_ERROR,
    			"MTBSnakeEnergyCD_MeanSep - " +
    			"update failed! Reason: " + e.getCommentString());
    }
  }

	@Override
  public void updateStatus(SnakeOptimizerSingleGreedy opt) {
		// nothing to do here
	}
	
	@Override
  public Matrix getDerivative_MatrixPart(SnakeOptimizerSingleVarCalc opt) {

		// init result matrix
		int snakePointNum = this.currentSnk.getPointNum();
		Matrix A = new Matrix(snakePointNum * 2, snakePointNum * 2);

		// iterate over all points and fill the matrix
		Vector<Point2D.Double> polyPoints = this.currentSnk.getPoints();
		for (int counter = 0; counter < snakePointNum - 1; ++counter) {
			Point2D.Double p = polyPoints.get(counter);
			double eDerive = this.getDerivative(this.currentSnk,
          (int)(p.x*this.scaleFactor), (int)(p.y*this.scaleFactor),0);
			// calculate normalized final derivative value
			double xi_i = this.normalizeEnergyValue(eDerive);
			
			// update matrix: top right, first entry
//			double oldVal= A.get(counter, snakePointNum + counter);
			A.set(counter, snakePointNum + counter, - xi_i);
			// update matrix: top right, second entry
//			oldVal= A.get(counter, snakePointNum + counter + 1);
			A.set(counter, snakePointNum + counter + 1, xi_i);
			// update matrix: top right
//			oldVal= A.get(snakePointNum + counter, counter);
			A.set(snakePointNum + counter, counter, xi_i);
			// update matrix: top right
//			oldVal= A.get(snakePointNum + counter, counter + 1);
			A.set(snakePointNum + counter, counter + 1, - xi_i);
		}
		// add last point
		Point2D.Double p = polyPoints.get(snakePointNum - 1);
		double eDerive = this.getDerivative(this.currentSnk,
        (int)(p.x*this.scaleFactor), (int)(p.y*this.scaleFactor),0);
		// calculate normalized final derivative value
		double xi_i = this.normalizeEnergyValue(eDerive);

		// fill matrix - upper right block
//		double oldVal = A.get(snakePointNum - 1, snakePointNum);
		A.set(snakePointNum - 1, snakePointNum, - xi_i);
//		oldVal = A.get(snakePointNum - 1, snakePointNum * 2 - 1);
		A.set(snakePointNum - 1, snakePointNum * 2 - 1, xi_i);
		// fill matrix - lower left block
//		oldVal = A.get(2 * snakePointNum - 1, 0);
		A.set(2 * snakePointNum - 1, 0, xi_i);
//		oldVal = A.get(2 * snakePointNum - 1, snakePointNum - 1);
		A.set(2 * snakePointNum - 1, snakePointNum - 1, - xi_i);
		return A;
	}

	/**
	 * Map energy value into target interval.
	 * 
	 * @param v	Input energy value.
	 * @return	Mapped value lying in target interval.
	 */
	private double normalizeEnergyValue(double v) {
		double normedV = v;
		switch(this.normMode)
		{
		case NORM_BALANCED_DERIVATIVES:
			switch(this.mapMode)
			{
			case LINEAR:
				normedV = ( v - this.energyRange[0])/this.energyRange[2] 
						* MTBSnakeEnergyDerivable.targetEnergyRange[2] 
								+ MTBSnakeEnergyDerivable.targetEnergyRange[0];
				break;
//			case LINEAR_CROPPED:
//				if (Math.abs(v) > Math.abs(this.enerMinVal)) {
//					if (v < 0)
//						normV= -1.0;
//					else
//						normV= 1.0;
//				}
//				else {
//					normV= (v-this.enerMinVal)/this.enerRange * this.enerTargetRange 
//																											+ this.enerTargetInvalMin;
//				}
//				break;
			}
			break;
		case NORM_NONE:
		default:
			break;
		}			
		return normedV;
	}
	
	@Override
  public Matrix getDerivative_VectorPart(SnakeOptimizerSingleVarCalc opt) {
	  return null;
  }

  @Override
  public double calcEnergy(SnakeOptimizerSingle opt) {
  	
  	// TODO only partially implemented!
//  	
//  	MTBSnake snake= (MTBSnake)opt.getCurrentSnakes().elementAt(0);
//  	
//    // result values
//    double lengthTerm = 0;
//    double curvatureTerm = 0;
//    double regionHomoTerm = 0;
//
//    // iterate over all points of the snake
////    Vector<Point2D.Double> spoints = snake.getPoints();
////    int pointNum = snake.getPointNum();
//
////    double x_p, y_p, x_i, y_i, x_n, y_n;
////    for (int i = 1; i < pointNum - 1; ++i) {
////
////      x_p = spoints.get(i - 1).x;
////      y_p = spoints.get(i - 1).y;
////      x_i = spoints.get(i).x;
////      y_i = spoints.get(i).y;
////      x_n = spoints.get(i + 1).x;
////      y_n = spoints.get(i + 1).y;
//
////      // length term
////      lengthTerm += alpha
////          * ((x_n - x_i) * (x_n - x_i) + (y_n - y_i) * (y_n - y_i));
////
////      // curvature term
////      curvatureTerm += beta
////          * ((x_n - 2 * x_i + x_p) * (x_n - 2 * x_i + x_p) + (y_n - 2 * y_i + y_p)
////              * (y_n - 2 * y_i + y_p));
////
////    }
//
//    // region homogenity term
//    double sumInside = 0, sumOutside = 0;
//    for (int y = 0; y < this.iHeight; ++y) {
//      for (int x = 0; x < this.iWidth; ++x) {
//        // check if image point lies inside or outside
//        int value = this.inImg.getValueInt(x, y);
//        try {
//          if (!snake.jni_containsPoint(x / this.scaleFactor, y
//              / this.scaleFactor)) {
////            sumInside += (value - this.c_in) * (value - this.c_in)
////                / (255 * 255);
//          } else {
////            sumOutside += (value - this.c_out) * (value - this.c_out)
////                / (255 * 255);
//          }
//        } catch (MTBPolygon2DException e) {
//          // TODO Auto-generated catch block
//          e.printStackTrace();
//        }
//      }
//    }
////    regionHomoTerm = this.lambda_in * sumInside + this.lambda_out * sumOutside;
//    return 0.5 * (lengthTerm + curvatureTerm) + regionHomoTerm;
  	return 0.0;
  }

	@Override
  public String toString() {
	  return "MTBSnakeEnergyCD_MeanSep: Mean Separation ";
  }
}
