/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5679 $
 * $Date: 2012-06-22 11:41:56 +0200 (Fr, 22. Jun 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.snakes.energies;

import java.awt.geom.Point2D;
import java.util.Vector;

import Jama.Matrix;



import de.unihalle.informatik.Alida.annotations.ALDClassParameter;
import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;
import de.unihalle.informatik.Alida.annotations.ALDParametrizedClass;
import de.unihalle.informatik.MiToBo.core.datatypes.defines.MTBConstants;
import de.unihalle.informatik.MiToBo.segmentation.basics.MTBSegmentationInterface;
import de.unihalle.informatik.MiToBo.segmentation.snakes.datatypes.MTBSnake;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingle;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingleGreedy;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingleVarCalc;

/**
 * Snake energy to avoid overlap of multiple snakes.
 * 
 * @author moeller
 */
@ALDDerivedClass
@ALDParametrizedClass

public class MTBSnakeEnergyCD_OverlapPenalty 
	implements MTBSnakeEnergyDerivable, MTBSnakeEnergyComputable {

	/**
	 * Weighting factor and overlap penalty, respectively.
	 */
	@ALDClassParameter(label="Weighting Parameter")
	private double rho = 1.0;
	
	/**
	 * Number of snakes to be optimized.
	 */
	protected transient int snakeNum = 0;
	
	/**
	 * Maxmimum value of energy, dependent on number of snakes and rho.
	 */
	protected transient double maxEnergyVal = 0; 
	
	/**
	 * Scaling factor to rescale image coordinates in range [0,1] or several
	 * ranges to the original coordinates range like [1000, 1000] in a image of
	 * size 1000 x 1000. Default scaling factor is 1.
	 */
	protected double scaleFactor = 1.0;

	/**
	 * Set the scaling factor.
	 * 
	 * @param s
	 *          new scaling factor.
	 */
	@Override
  public void setScaleFactor(double s) {
		this.scaleFactor = s;
	}

	/**
	 * Get scaling factor.
	 * 
	 * @return Scaling factor.
	 */
	@Override
  public double getScaleFactor() {
		return this.scaleFactor;
	}

	/**
	 * Default constructor.
	 */
	public MTBSnakeEnergyCD_OverlapPenalty() {
		// nothing to do here
	}
	
  /**
   * Default constructor.
   * 
   * @param _rho	Overlap penalty coefficient.
   */
  public MTBSnakeEnergyCD_OverlapPenalty(double _rho, int snakes) {
  	this.rho= _rho;
  	this.snakeNum = snakes;
  }
  
  @Override
  public boolean initEnergy(SnakeOptimizerSingleVarCalc opt) {
  	this.snakeNum = opt.getSnakeNumber();
  	this.maxEnergyVal = 
  		this.rho > MTBConstants.epsilon ? this.rho*this.snakeNum : 1.0;
  	// indicate that this energy requires the global overlap mask
		opt.requireOverlapMask();
		opt.requireCounterClockwiseSnakePointOrder();
		return true;
	}

  @Override
  public boolean initEnergy(SnakeOptimizerSingleGreedy opt) {
  	this.snakeNum = opt.getSnakeNumber();
  	this.maxEnergyVal = 
  		this.rho > MTBConstants.epsilon ? this.rho*this.snakeNum : 1.0;
  	// indicate that this energy requires the global overlap mask
		opt.requireOverlapMask();
		opt.requireCounterClockwiseSnakePointOrder();
		return true;
	}

  @Override
  public void updateStatus(SnakeOptimizerSingleVarCalc o) {
	  // nothing to do here...  
  }

	@Override
  public void updateStatus(SnakeOptimizerSingleGreedy o) {
	  // nothing to do here...  
  }

	/**
   * Updates the region coupling energy term in matrix A.
   */
	@Override
  public Matrix getDerivative_MatrixPart(SnakeOptimizerSingleVarCalc opt) {

		// get snake data
		MTBSnake snake = (MTBSnake)opt.getCurrentSnakes().elementAt(0);
    Vector<Point2D.Double> polyPoints = snake.getPoints();
    int snakePointNum = polyPoints.size();
		Matrix A = new Matrix(snakePointNum * 2, snakePointNum * 2);

		// fetch current overlap mask
	  int[][] overlapMask = opt.getOverlapMask();
	  if (overlapMask == null) {
	  	return A;
	  }
	  	
	  // get mask of this snake
	  int[][] snakeMask;
    snakeMask = snake.getBinaryMask(opt.getInputImage().getSizeX(),
    		                            opt.getInputImage().getSizeY());	    
    // update matrix A
    for (int counter = 0; counter < snakePointNum - 1; ++counter) {
    	Point2D.Double p = polyPoints.get(counter);
    	// check in how many other snakes the point actually lies
    	double overlapCount = 
    		overlapMask[(int)(p.y*this.scaleFactor)][(int)(p.x*this.scaleFactor)]
    	- snakeMask[(int)(p.y*this.scaleFactor)][(int)(p.x*this.scaleFactor)];
//	    	double oldEntryA = -A.get(counter, snakePointNum + counter);
    	double newEntryA = (this.rho * overlapCount)/this.maxEnergyVal;
    	A.set(counter, snakePointNum + counter, -newEntryA);
    	A.set(counter, snakePointNum + counter + 1, newEntryA);
    	A.set(snakePointNum + counter, counter, newEntryA);
    	A.set(snakePointNum + counter, counter + 1, -newEntryA);
    }
    // add last point
    Point2D.Double p = polyPoints.get(snakePointNum - 1);
    // check in how many other snakes the point actually lies
    double overlapCount = 
     	  overlapMask[(int)(p.y*this.scaleFactor)][(int)(p.x*this.scaleFactor)] 
      - snakeMask[(int)(p.y*this.scaleFactor)][(int)(p.x*this.scaleFactor)];
//	    double oldEntryA = A.get(snakePointNum - 1, snakePointNum);
  	double newEntryA = (this.rho * overlapCount)/this.maxEnergyVal;
    A.set(snakePointNum - 1, snakePointNum, newEntryA);
    A.set(snakePointNum - 1, snakePointNum * 2 - 1, -newEntryA);
    A.set(2 * snakePointNum - 1, 0, -newEntryA);
    A.set(2 * snakePointNum - 1, snakePointNum - 1, newEntryA);
    return A;
	}

	@Override
  public Matrix getDerivative_VectorPart(SnakeOptimizerSingleVarCalc opt) {
	  return null;
  }

  @Override
  public double calcEnergy(SnakeOptimizerSingle opt) {
	  // TODO Auto-generated method stub
	  return 0;
  }

	@Override
  public String toString() {
		return new String("MTBSnkEner: Snake overlap penalty (rho= "+this.rho+")");
  }
}
