/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5679 $
 * $Date: 2012-06-22 11:41:56 +0200 (Fr, 22. Jun 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.snakes.energies;

import java.awt.geom.Point2D;
import java.util.Vector;

import Jama.Matrix;

import de.unihalle.informatik.MiToBo.core.datatypes.MTBPolygon2DException;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImageDouble;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.Alida.annotations.ALDClassParameter;
import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;
import de.unihalle.informatik.Alida.annotations.ALDParametrizedClass;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;
import de.unihalle.informatik.MiToBo.segmentation.basics.MTBSegmentationInterface;
import de.unihalle.informatik.MiToBo.segmentation.snakes.datatypes.MTBSnake;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingle;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingleGreedy;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingleVarCalc;
import de.unihalle.informatik.MiToBo.tools.image.ImageValueTools;

/**
 * Snake energy based on region variance separation.
 * <p>
 * This energy models tries to best separate the intensity variance
 * of the inner and outer regions of a snake. For details refer to:
 * 
 * Yezzi, Tsai, Willsky, "A Fully Global Approach to Image Segmentation via 
 * Coupled Curve Evolution Equations", JVCIR, 13, 195-216 (2002) 
 * <p>
 * Important! Note that this energy is quite instable, so always use 
 * together with curvature regularization!
 * 
 * @author moeller
 */


@ALDDerivedClass
@ALDParametrizedClass
public class MTBSnakeEnergyCD_VarSep 
	implements MTBSnakeEnergyDerivable, MTBSnakeEnergyComputable {

  /**
	 * Original image.
	 */
	@ALDClassParameter(label="Input Image")
	private MTBImage inImgOrig;

	/**
	 * Scaling factor to rescale image coordinates in range [0,1] or several
	 * ranges to the original coordinates range like [1000, 1000] in a image of
	 * size 1000 x 1000. Default scaling factor is 1.
	 */
	protected double scaleFactor = 1.0;

	/**
	 * Set the scaling factor.
	 * 
	 * @param s
	 *          new scaling factor.
	 */
	@Override
  public void setScaleFactor(double s) {
		this.scaleFactor = s;
	}

	/**
	 * Get scaling factor.
	 * 
	 * @return Scaling factor.
	 */
	@Override
  public double getScaleFactor() {
		return this.scaleFactor;
	}

	/**
	 * Configuration modes of this energy.
	 */
	public static enum mode {
		/**
		 * Original operating mode.
		 * <p>
		 * Runs exactly the original approach as published by Yezzi et al. 
		 */
		ORIGINAL,
		/**
		 * Local variance averaging.
		 * <p>
		 * Slightyly modified version where the local differences between region
		 * means and local intensity are calculated over a local neighborhood.
		 * This improves performance if the are large intensity variances in the 
		 * image to be segmented. 
		 */
		LOCAL_AVG
	}

	private boolean debug = true;
	
//	private MTBSnakeE_VarSep.mode energyMode = mode.ORIGINAL;
	private MTBSnakeEnergyCD_VarSep.mode energyMode = mode.LOCAL_AVG;

  /**
   * Mask size for averaging in local average mode. 
   */
  private int averageMaskSize = 13;

  /**
	 * Original image.
	 */
//	private MTBImage inImgOrig;

	/**
	 * Normalized image to work on.
	 */
//	private MTBImageDouble inImg;
	
	/**
	 * Width of working image.
	 */
	private int iWidth;
	
	/**
	 * Height of working image. 
	 */
	private int iHeight;
	
  /**
   * Binary mask of the snake.
   */
  private int[][] snakeMask = null;
  
  /**
   * Binary snake mask from previous iteration.
   */
  private int[][] snakeMaskOld = null;

  /**
	 * Mask to indicate overlap regions between snakes.
	 */
  protected int[][] overlapMask = null;

	/**
	 * Mask to exclude image pixels from fitting term calculation.
	 */
  protected int[][] excludeMask = null;

  /*
   * Some private helpers during calculations.
   */
  
  /**
   * Interior region area.
   */
  private int a_in = -1;
  
  /**
   * Exterior region area.
   */
  private int a_out = -1;
  /**
   * Inner average intensity.
   */
  private double c_in = -1;
  /**
   * Outer average intensity.
   */
  private double c_out = -1;
  /**
   * Inner intensity variance.
   */
  private double var_in = -1;
  /**
   * Outer intensity variance.
   */
  private double var_out = -1;

  /**
   * Memory variable for last inner average value.
   */
  private double final_c_in = -1;

  /**
   * Memory variable for last outer average value.
   */
  private double final_c_out = -1;
  
  /**
   * Memory variable for last inner variance value.
   */
  private double final_c_in_var = -1;
  
  /**
   * Memory variable for last outer variance value.
   */
  private double final_c_out_var = -1;

  /**
   * Default constructor.
   * @throws ALDOperatorException 
   */
  public MTBSnakeEnergyCD_VarSep() throws ALDOperatorException {
  	// nothing to do here
  }
  
  /**
   * Default constructor.
   * 
   * @param im	Image to work on.
   * @throws ALDOperatorException 
   */
  public MTBSnakeEnergyCD_VarSep(MTBImage im) throws ALDOperatorException {
  	if (im != null) {
  		this.inImgOrig= im;
//  		this.inImgOrig= im;
//  		this.iHeight= im.getSizeY();
//  		this.iWidth= im.getSizeX();
  	}
  }
  
	@Override
  public boolean initEnergy(SnakeOptimizerSingleVarCalc opt) {
		if (this.inImgOrig != null) {
			this.iHeight= this.inImgOrig.getSizeY();
			this.iWidth= this.inImgOrig.getSizeX();
		}
//		try {
			MTBImageDouble newInImg = 
				(MTBImageDouble)this.inImgOrig.convertType(null, 
																						MTBImageType.MTB_DOUBLE, true);
//	    ImageValueTools.normImageFromTo(newInImg, this.inImgOrig.getTypeMin(), 
//	    														this.inImgOrig.getTypeMax(),0.0,1.0, null);
//	    ImageValueTools.normImageTo(newInImg, 0.0,1.0, null);
//	    this.inImg = newInImg;
//    } catch (ALDOperatorException e) {
//	    // TODO Auto-generated catch block
//	    e.printStackTrace();
//    } catch (ALDProcessingDAGException e) {
//	    // TODO Auto-generated catch block
//	    e.printStackTrace();
//    }
		opt.requireCounterClockwiseSnakePointOrder();
		opt.requireOverlapMask();
	  return true;
  }

	@Override
  public boolean initEnergy(SnakeOptimizerSingleGreedy opt) {
		if (this.inImgOrig != null) {
			this.iHeight= this.inImgOrig.getSizeY();
			this.iWidth= this.inImgOrig.getSizeX();
		}
		try {
			MTBImageDouble newInImg = 
				(MTBImageDouble)this.inImgOrig.convertType(null, 
																						MTBImageType.MTB_DOUBLE, true);
	    ImageValueTools.normImageFromTo(newInImg, this.inImgOrig.getTypeMin(), 
	    														this.inImgOrig.getTypeMax(),0.0,1.0, null);
//	    ImageValueTools.normImageTo(newInImg, 0.0,1.0, null);
//	    this.inImg = newInImg;
    } catch (ALDOperatorException e) {
	    // TODO Auto-generated catch block
	    e.printStackTrace();
    } catch (ALDProcessingDAGException e) {
	    // TODO Auto-generated catch block
	    e.printStackTrace();
    }
		opt.requireCounterClockwiseSnakePointOrder();
		opt.requireOverlapMask();
	  return true;
  }

	@Override
  public void updateStatus(SnakeOptimizerSingleGreedy o) {
	  // nothing to do here...  
  }

	/**
   * Updates the fitting term based on new region average/variance values.
   */
	@Override
  public Matrix getDerivative_MatrixPart(SnakeOptimizerSingleVarCalc opt) {

		MTBSnake snake = (MTBSnake)opt.getCurrentSnakes().elementAt(0);
		int snakePointNum = snake.getPointNum();
		Matrix A = new Matrix(snakePointNum * 2, snakePointNum * 2);
	
    if (this.snakeMask != null) 
    	this.snakeMaskOld =	this.snakeMask.clone();	
    this.snakeMask = snake.getBinaryMask(this.iWidth, this.iHeight);
    
    this.calcRegionStats(true);

    if (this.debug) {
    	System.out.println("Snakes - VarSepEnergy:");
    	System.out.println("AREA_IN= "+ this.a_in);
    	System.out.println("AREA_OUT= "+ this.a_out);
    	System.out.println("MEAN_IN= "+ this.c_in);
    	System.out.println("MEAN_OUT= "+ this.c_out);
    	System.out.println("VAR_IN= "+ this.var_in);
    	System.out.println("VAR_OUT= "+ this.var_out);
    }
    
    double minEnergy= Double.MAX_VALUE;
    double maxEnergy= 0.0;
    
    double [] energies = new double[snakePointNum];
    
		double inRegTerm = 0;
		double outRegTerm = 0;

		Vector<Point2D.Double> polyPoints = snake.getPoints();
		for (int counter = 0; counter < snakePointNum - 1; ++counter) {
			Point2D.Double p = polyPoints.get(counter);

			switch(this.energyMode) 
			{
			case ORIGINAL:
				// get point intensity
//				double pointInt = 
//					this.inImg.getValueDouble((int)(p.x * this.scaleFactor),
//																		(int) (p.y * this.scaleFactor));
//				inRegTerm = 
//					((pointInt-this.c_in)*(pointInt-this.c_in)-this.var_in)/this.a_in;
//				outRegTerm = 
//					((pointInt-this.c_out)*(pointInt-this.c_out)-this.var_out)/this.a_out;
				break;
			case LOCAL_AVG:
				int x = (int)(p.x * this.scaleFactor);
				int y = (int)(p.y * this.scaleFactor);
				inRegTerm =  (this.getAvgVarIn(x, y)-this.var_in)/this.a_in;
				outRegTerm = (this.getAvgVarOut(x, y)-this.var_out)/this.a_out;
				break;
			}
			
			// calculate final energy term
			double nu_i = - (this.var_in - this.var_out) * (inRegTerm + outRegTerm);

			// remember min and max values 
			if (nu_i < minEnergy)
				minEnergy = nu_i;
			if (nu_i > maxEnergy)
				maxEnergy = nu_i;

			// remember energy itself
			energies[counter] = nu_i;
		}		
		// last point
		Point2D.Double p = polyPoints.get(snakePointNum - 1);
		switch(this.energyMode) 
		{
		case ORIGINAL:
//			double pointInt = 
//				this.inImg.getValueDouble((int) (p.x * this.scaleFactor),
//						(int) (p.y * this.scaleFactor));
//			inRegTerm = 
//				((pointInt-this.c_in)*(pointInt-this.c_in)-this.var_in) / this.a_in;
//			outRegTerm = 
//				((pointInt-this.c_out)*(pointInt-this.c_out)-this.var_out)/this.a_out;
//			break;
		case LOCAL_AVG:
			int x = (int)(p.x * this.scaleFactor);
			int y = (int)(p.y * this.scaleFactor);
			inRegTerm = (this.getAvgVarIn(x, y)-this.var_in)/this.a_in;
			outRegTerm = (this.getAvgVarOut(x, y)-this.var_out)/this.a_out;
			break;
		}
		
		double nu_i = - (this.var_in-this.var_out) * (inRegTerm + outRegTerm);
		
		// remember min and max values 
		if (nu_i < minEnergy)
			minEnergy = nu_i;
		if (nu_i > maxEnergy)
			maxEnergy = nu_i;

		// remember energy itself
		energies[snakePointNum-1] = nu_i;

		// normalize energies to [-1,1]
		double absMax = ( Math.abs(minEnergy) > Math.abs(maxEnergy) ?
				Math.abs(minEnergy) : Math.abs(maxEnergy) );
		for (int i=0; i<energies.length; ++i) {
			energies[i] = energies[i] / absMax;
		}
		
		// update matrix
		for (int counter = 0; counter < snakePointNum - 1; ++counter) {

			nu_i = energies[counter];
			
//			double oldVal= A.get(counter, snakePointNum + counter);
			A.set(counter, snakePointNum + counter, - nu_i);
//			oldVal= A.get(counter, snakePointNum + counter + 1);
			A.set(counter, snakePointNum + counter + 1, nu_i);
//			oldVal= A.get(snakePointNum + counter, counter);
			A.set(snakePointNum + counter, counter, nu_i);
//			oldVal= A.get(snakePointNum + counter, counter + 1);
			A.set(snakePointNum + counter, counter + 1, - nu_i);
		}		
		// update matrix for last point
		nu_i = energies[snakePointNum-1];
		
		// fill matrix - upper right block
//		double oldVal = A.get(snakePointNum - 1, snakePointNum);
		A.set(snakePointNum - 1, snakePointNum,- nu_i);
//		oldVal = A.get(snakePointNum - 1, snakePointNum * 2 - 1);
		A.set(snakePointNum - 1, snakePointNum * 2 - 1, nu_i);
		// fill matrix - lower left block
//		oldVal = A.get(2 * snakePointNum - 1, 0);
		A.set(2 * snakePointNum - 1, 0, nu_i);
//		oldVal = A.get(2 * snakePointNum - 1, snakePointNum - 1);
		A.set(2 * snakePointNum - 1, snakePointNum - 1, - nu_i);

		return A;
	}

	@Override
  public Matrix getDerivative_VectorPart(SnakeOptimizerSingleVarCalc opt) {
	  return null;
  }

  @Override
  public double calcEnergy(SnakeOptimizerSingle opt) {
  	
  	// TODO only partially implemented!
  	
  	MTBSnake snake= (MTBSnake)opt.getCurrentSnakes().elementAt(0);
  	
//    // calculate region means, if not done before
//    if (this.c_in < 0 || this.c_out < 0)
  	this.calcRegionStats(false);

    // result values
    double lengthTerm = 0;
    double curvatureTerm = 0;
    double regionHomoTerm = 0;

    // iterate over all points of the snake
//    Vector<Point2D.Double> spoints = snake.getPoints();
//    int pointNum = snake.getPointNum();

//    double x_p, y_p, x_i, y_i, x_n, y_n;
//    for (int i = 1; i < pointNum - 1; ++i) {
//
//      x_p = spoints.get(i - 1).x;
//      y_p = spoints.get(i - 1).y;
//      x_i = spoints.get(i).x;
//      y_i = spoints.get(i).y;
//      x_n = spoints.get(i + 1).x;
//      y_n = spoints.get(i + 1).y;

//      // length term
//      lengthTerm += alpha
//          * ((x_n - x_i) * (x_n - x_i) + (y_n - y_i) * (y_n - y_i));
//
//      // curvature term
//      curvatureTerm += beta
//          * ((x_n - 2 * x_i + x_p) * (x_n - 2 * x_i + x_p) + (y_n - 2 * y_i + y_p)
//              * (y_n - 2 * y_i + y_p));
//
//    }

    // region homogenity term
    double sumInside = 0, sumOutside = 0;
    for (int y = 0; y < this.iHeight; ++y) {
      for (int x = 0; x < this.iWidth; ++x) {
        // check if image point lies inside or outside
      	double value = 0.0;
//        double value = this.inImg.getValueDouble(x, y);
        try {
          if (!snake.jni_containsPoint(x / this.scaleFactor, y
              / this.scaleFactor)) {
            sumInside += (value - this.c_in) * (value - this.c_in)
                / (255 * 255);
          } else {
            sumOutside += (value - this.c_out) * (value - this.c_out)
                / (255 * 255);
          }
        } catch (MTBPolygon2DException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }
//    regionHomoTerm = this.lambda_in * sumInside + this.lambda_out * sumOutside;
    return 0.5 * (lengthTerm + curvatureTerm) + regionHomoTerm;
  }

  /**
   * Calculates new average values, variances and region areas 
   * for snake interior and background.
   * 
   * @param incremental
   *          If true, calculation is incremental.
   * 
   *          TODO implement incremental calculation
   */
  private void calcRegionStats(boolean incremental) {

    int countInside = 0;
    int countOutside = 0;
    double sumInside = 0;
    double sumOutside = 0;

    int [][] changeMask = new int[this.iHeight][this.iWidth];

    // in first iteration mean values need to be calculated with
    // dummy change mask
    if (this.snakeMaskOld == null || !incremental) {
      for (int y = 0; y < this.iHeight; ++y)
        for (int x = 0; x < this.iWidth; ++x)
          if (this.snakeMask[y][x] == 0)
            // pixel lies outside
            changeMask[y][x] = 0;
          else {
            // pixel lies inside
            changeMask[y][x] = 1;
          }
    }
    // calculate change mask
    else {
      for (int y = 0; y < this.iHeight; ++y) {
        for (int x = 0; x < this.iWidth; ++x) {
          // pixel is both times outside
          if (this.snakeMask[y][x] == 0 && this.snakeMaskOld[y][x] == 0)
            changeMask[y][x] = 0;
          // pixel is now inside
          else if (this.snakeMask[y][x] == 1 && this.snakeMaskOld[y][x] == 0)
            changeMask[y][x] = 1;
          // pixel is now outside
          else if (this.snakeMask[y][x] == 0 && this.snakeMaskOld[y][x] == 1)
            changeMask[y][x] = 2;
          // both times outside
          else
            changeMask[y][x] = 3;
        }
      }
    }
    // calculate average intensities
    for (int y = 0; y < this.iHeight; ++y) {
      for (int x = 0; x < this.iWidth; ++x) {
        // exclude masked pixels from mean value calculations
        if (this.excludeMask != null && this.excludeMask[y][x] > 0)
          continue;
        switch (changeMask[y][x]) {
          case 0:
          case 2:
//            if (this.overlapMask != null) {
//              if (this.overlapMask[y][x] == 0) {
//                countOutside++;
//                sumOutside += this.inImg.getValueDouble(x, y);
//              }
//            } else {
//              countOutside++;
//              sumOutside += this.inImg.getValueDouble(x, y);
//            }
//            break;
//          case 1:
//          case 3:
//            countInside++;
//            sumInside += this.inImg.getValueDouble(x, y);
            break;
        }
      }
    }
    // remember areas
    this.a_in = countInside;
    this.a_out = countOutside;
    
    // compute final intensity parameters
    this.c_in = sumInside / countInside;
    this.c_out = sumOutside / countOutside;

    // store for access from outside of the class
    this.final_c_in = this.c_in;
    this.final_c_out = this.c_out;

    // calculate variance
    sumInside = 0;
    sumOutside = 0;
    countInside = 0;
    countOutside = 0;
    for (int y = 0; y < this.iHeight; ++y) {
      for (int x = 0; x < this.iWidth; ++x) {
        // exclude masked pixels from mean value calculations
        if (this.excludeMask != null && this.excludeMask[y][x] > 0)
          continue;
        switch (changeMask[y][x]) {
          case 0:
          case 2:
//            if (this.overlapMask != null) {
//              if (this.overlapMask[y][x] == 0) {
//                countOutside++;
//                sumOutside += (this.inImg.getValueDouble(x, y) - this.c_out)
//                    * (this.inImg.getValueDouble(x, y) - this.c_out);
//              }
//            } else {
//              countOutside++;
//              sumOutside += (this.inImg.getValueDouble(x, y) - this.c_out)
//                  * (this.inImg.getValueDouble(x, y) - this.c_out);
//            }
//            break;
          case 1:
          case 3:
//            countInside++;
//            sumInside += (this.inImg.getValueDouble(x, y) - this.c_in)
//                * (this.inImg.getValueDouble(x, y) - this.c_in);
            break;
        }
      }
    }
    // compute final intensity parameters
    this.var_in = sumInside / countInside;
    this.var_out = sumOutside / countOutside;
    this.final_c_in_var = sumInside / countInside;
    this.final_c_out_var = sumOutside / countOutside;
  }

  /**
   * Returns the recently computed inner average intensity value.
   */
  public double getCin() {
    return this.final_c_in;
  }

  /**
   * Returns the recently computed outer average intensity value.
   */
  public double getCout() {
    return this.final_c_out;
  }

  /**
   * Returns the recently computed inner intensity variance.
   */
  public double getCinVariance() {
    return this.final_c_in_var;
  }

  /**
   * Returns the recently computed outer intensity variance.
   */
  public double getCoutVariance() {
    return this.final_c_out_var;
  }

	@Override
  public String toString() {
	  return new String("MTBSnkEner: Variance Separation");
  }
	
	protected double getAvgVarIn(int xx, int yy) {
		int mask_size_half = this.averageMaskSize/2;
		
//		double avgVar= 0;
//		int count = 0;
//		for (int dy = -mask_size_half; dy <= mask_size_half; ++dy) {
//			for (int dx = -mask_size_half; dx <= mask_size_half; ++dx) {
//				int x = xx+dx;
//				int y = yy+dy;
//				if (x<0||y<0||x>=this.inImg.getSizeX()||y>=this.inImg.getSizeY())
//					continue;
//				avgVar = avgVar + (this.inImg.getValueDouble(x, y) - this.c_in) *
//								 					(this.inImg.getValueDouble(x, y) - this.c_in);
//				count++;
//			}			
//		}
//		return avgVar/count;
		return 0;
	}
	protected double getAvgVarOut(int xx, int yy) {
		int mask_size_half = this.averageMaskSize/2;
		
//		double avgVar= 0;
//		int count = 0;
//		for (int dy = -mask_size_half; dy <= mask_size_half; ++dy) {
//			for (int dx = -mask_size_half; dx <= mask_size_half; ++dx) {
//				int x = xx+dx;
//				int y = yy+dy;
//				if (x<0||y<0||x>=this.inImg.getSizeX()||y>=this.inImg.getSizeY())
//					continue;
//				avgVar = avgVar + (this.inImg.getValueDouble(x, y) - this.c_out) *
//								          (this.inImg.getValueDouble(x, y) - this.c_out);
//				count++;
//			}			
//		}
//		return avgVar/count;
		return 0.0;
	}

	@Override
  public void updateStatus(SnakeOptimizerSingleVarCalc o) {
	  // TODO Auto-generated method stub
	  
  }
}
