/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5795 $
 * $Date: 2012-07-13 17:21:14 +0200 (Fr, 13. Jul 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.snakes.optimize;

import java.awt.geom.Point2D;
import java.io.IOException;
import java.util.Vector;

import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;


import de.unihalle.informatik.Alida.annotations.ALDAOperator;
import de.unihalle.informatik.Alida.annotations.ALDAOperator.ExecutionMode;
import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.MiToBo.core.datatypes.MTBPolygon2DException;
import de.unihalle.informatik.MiToBo.core.datatypes.MTBPolygon2DSet;
import de.unihalle.informatik.MiToBo.core.datatypes.images.*;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.gui.MTBTableModel;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.exceptions.MTBSnakeException;
import de.unihalle.informatik.MiToBo.segmentation.snakes.datatypes.MTBSnake;
import de.unihalle.informatik.MiToBo.segmentation.snakes.energies.MTBSet_SnakeEnergyDerivable;
import de.unihalle.informatik.MiToBo.segmentation.snakes.energies.MTBSnakeEnergyDerivable;

/**
 * <pre>
 * 
 * Image contour segmentation using parametric snakes.
 * 
 * This class provides methods to segment contours in an image based on
 * parametric active contour models, i.e. snakes. The snake optimization
 * is done based on a greedy algorithm. Its main intention is for comparison
 * and demontration purposes. You should always keep in mind that greedy
 * snake algorithms are slow - particularly since for each point movement
 * excessive local energy calculations have to be done. 
 * 
 * </pre>
 * 
 * 
 * @author Birgit Möller
 */
@ALDAOperator(genericExecutionMode=ExecutionMode.NONE)
@ALDDerivedClass

public class SnakeOptimizerSingleGreedy extends SnakeOptimizerSingle {

	@Parameter(label="List of Energies", type=Parameter.Type.PARAMETER,
			direction=Parameter.Direction.IN, required=true,
			description = "List of snake energies.")
	protected MTBSet_SnakeEnergyDerivable energies= null;



  /************************************************************************/
  /*** Internal helpers for optimization procedure. ***/
  /************************************************************************/

//  /**
//   * Iteration counter.
//   */
//  protected int itCounter = 0;
//
//  protected double energy = 0;
//
//  protected double oldEnergy = 0;
//
//  protected double minEnergy = Double.MAX_VALUE;
//  protected double minEnergyOld = Double.MAX_VALUE;
//
//  protected double c_in = -1;
//  protected double c_out = -1;
//
//  protected double final_c_in = -1;
//  protected double final_c_out = -1;
//  protected double final_c_in_var = -1;
//  protected double final_c_out_var = -1;
//
//  private StringBuffer input = new StringBuffer();
//
//  private StringBuffer inputExt = new StringBuffer();
//
//  private int[][] snakeMask = null;
//  private int[][] snakeMaskOld = null;
//
//  private int[][] overlapMask = null;
//
//  private int[][] excludeMask = null;

	/**
	 * Current energy of the snake.
	 */
	protected double energy = -1.0;
	
  /***********************************************************************/
  /*** Data Access. ***/
  /***********************************************************************/

  /***********************************************************************/
  /*** Optimizer initialization. ***/
  /***********************************************************************/

	public SnakeOptimizerSingleGreedy() throws ALDOperatorException{
			// nothing to do here
	}
	
  /**
   * Initializes the optimizer.
   * <p>
   * Here internal member variables are initialized according to the given
   * parameters, and memory for intermediate results and debug data is
   * allocated.
   * 
   * @param p
   *          Parameter object (should be of type
   *          {@link de.unihalle.informatik.MiToBo.segmentation.snakes.SnakeOptimizerSingleParams}
   *          ).
   * @throws IllegalArgumentException 
   * @throws ALDOperatorException 
   */
  @SuppressWarnings("unused")
  @Override
  protected void initOptimizer() 
  		throws MTBSnakeException, IllegalArgumentException {

    // reset iterations counter
    this.itCounter = 0;

    this.printParams();

//    MTBSnakeExternalEnergy d = this.getExternalEnergy();
    this.inImg = this.getInputImage();
    MTBSnake initS = (MTBSnake)this.getInitialSnakes().elementAt(0);
// System.out.println("init= " + this.snake.getPoints().size());
//    if (snakeParamsSingle.normalizeSnake()) {
//      System.out.println("Doing normalization...");
//      this.scaleFactor = (this.inImg.getSizeX() > this.inImg.getSizeY()) ? this.inImg
//          .getSizeX()
//          : this.inImg.getSizeY();
//      this.setSnake(new Snake(initS.getSnakePoints(), initS.isClosed(),
//          this.scaleFactor, false));
//      if (snakeParamsSingle.useExternalEnergies)
//        snakeParamsSingle.getExternalEnergy().setScaleFactor(this.scaleFactor);
//    } else
    this.snake = initS;

    // check if it is sorted correctly
    try {
      if (this.snake.getPointNum() > 5 
          && !this.snake.jni_isCounterClockwiseOriented())
        this.snake.reversePolypoints();
    } catch (MTBPolygon2DException e) {
      System.out.println(">>> " + e.getError());
    }

    if (this.snake.getPointNum() > 5 && this.doResampling)
      this.resampleSnake();

    if (this.inImg != null) {
      this.iWidth = this.inImg.getSizeX();
      this.iHeight = this.inImg.getSizeY();
    }
    if (this.showIntermediateResults) {
      // result image
      this.dispImg = (MTBImageRGB) MTBImage.createMTBImage(this.iWidth,
          this.iHeight, 1, 1, 1, MTBImageType.MTB_RGB);
      this.dispImg.show();
    }

    // calculate energy, if required
//    this.energy = this.calcSnakeEnergy();

    // display current snake energy
    System.out.println("Initial snake energy: " + this.energy);
    /*
     * re-sample before starting b/c of no re-sampling in run for including all
     * points in printing the energies (added by danny)
     */
    // this.snake.resample(5);
    // this.snake.resample(5);
  }

  /***********************************************************************/
  /*** Optimization. ***/
  /***********************************************************************/

  /**
   * Performs one single optimization step, i.e. iterates once over all points.
   * 
   * @return Status of the optimization process.
   * @throws ALDOperatorException 
   */
  @Override
  protected Snake_status doIteration() {

    // check if you have enough snake points, if not, stop
    if (this.snake.getPointNum() <= 5)
      return Snake_status.SNAKE_FAIL;

    // save current snake data for later reference
//    Snake oldSnake = this.snake.clone();
//    if (this.snakeMask != null)
//      this.snakeMaskOld = this.snakeMask.clone();
//    this.snakeMask = this.snake.getBinaryMask(this.iWidth, this.iHeight);

    // get parameters
//    Stop_criterion sc = this.getStopCrit();
//    int maxIterations = this.getIterations();

    // increase iteration counter
    this.itCounter++;

    // move each point to local minimum in neighborhood
    int pointsMoved= 0;
    for (Point2D.Double p: this.snake.getPoints()) {
    	Point2D.Double saveP= (Point2D.Double)p.clone();
	double startE= this.calcSnakeEnergy();
	double Emin= startE;
    	int minDx= 0; int minDy= 0;
		for (int dy=-1;dy<=1;++dy) {
			for (int dx=-1;dx<=1;++dx) {
				int ptx= (int)saveP.x + dx;
				int pty= (int)saveP.y + dy;
				p.setLocation(ptx, pty);
				double E= this.calcSnakeEnergy();
				if (E < Emin) {
					Emin= E;
					minDx= dx;
					minDy= dy;
				}
			}
		}
		if (Emin < startE && (minDx != 0 || minDy != 0)) {
		    p.setLocation((int)saveP.x + minDx, (int)saveP.y + minDy);
			pointsMoved++;
		}
		else
		    p.setLocation(saveP);		    
    }

//    // resample the snake
//    if (snakeParamsSingle.resampleSnake() && (this.itCounter % 2 == 0)) {
//      this.resampleSnake();
//    }
//    if (this.snake.getPointNum() <= 5)
//      return SnakeOptimizerSingle.Snake_status.SNAKE_FAIL;

    // loop elimination, every 5 iterations
    // TODO check number of iterations
//     if (this.itCounter % 3 == 0)

//       if (!this.snake.jni_isSimple()) {
//         System.out.println("Snake is not simple: simplifying it...");
//         this.snake.jni_makePolySimple();
//       }
    // not used any more, deleted by danny
    // if ((snakeParamsSingle.useRegionCriterion ||
    // snakeParamsSingle.coupleSnakes)
    // && !this.snake.jni_isCounterClockwiseOriented()) {
    // System.out.println("After simplification: snake is in wrong order, "
    // + "reverting it...");
    // this.snake.reversePolypoints();
    // }

    // calculate energy
    // this.energy = this.calcSnakeEnergy();

    // check if there are snake points outside of the image
    // => project them back to image border
    // ToDo @Danny: test for ID's, should be the same as before !!!
//    newpoints = new Vector<SnakePoint2D>();
//    for (int i = 0; i < this.snake.getPointNum(); ++i) {
//      SnakePoint2D po = this.snake.getSnakePoints().elementAt(i);
//      if (po.x < 0)
//        po.x = 0;
//      if (po.x * this.scaleFactor > this.iWidth - 1)
//        po.x = (this.iWidth - 1) / this.scaleFactor;
//      if (po.y < 0)
//        po.y = 0;
//      if (po.y * this.scaleFactor > this.iHeight - 1)
//        po.y = (this.iHeight - 1) / this.scaleFactor;
//      newpoints.addElement(po);

      // System.out.println(po.getOldId() +
      // this.snake.getPoints().get(i).getOldId);

//    }
//    newSnake = new Snake(newpoints, true, this.scaleFactor, true);
//    this.setSnake(newSnake);

    // this.oldEnergy= this.energy;
    // boolean minEnergyChanged= false;
    // if (this.energy < this.minEnergy) {
    // this.minEnergyOld= this.minEnergy;
    // this.minEnergy= this.energy;
    // minEnergyChanged= true;
    // }

    // if (minEnergyChanged && this.minEnergyOld > this.minEnergy) {
    // double energyImprove= (this.minEnergyOld -
    // this.minEnergy)/this.minEnergyOld;
    // this.gammaAdaptive= this.gammaAdaptive*(1+2*energyImprove);
    // }

//     System.out.println("Iteration: " + this.itCounter + " , Energy= \t"
//         + this.energy + " , Pointnum= " + this.snake.getPointNum()
//         + " , gamma= " + this.gammaAdaptive[0][0]);
//     System.out.println("\t\t Energy-details: ");
//     if (snakeParamsSingle.useInternalEnergies) {
//       System.out.print("length= " + this.calcEnergyInternal_lengthTerm()
//           + " , ");
//       System.out.print("curv= " + this.calcEnergyInternal_curvTerm() + "\n");
//     }
//     if (snakeParamsSingle.useExternalEnergies)
//       System.out.println("ext= " + this.calcEnergyExternal());

    // check if energy minimum has been reached
//    switch (sc) {
//      case SNAKE_STOP_MAXITERATIONS:
//        if (this.itCounter >= maxIterations)
//          return SnakeOptimizerSingle.Snake_status.SNAKE_DONE;
//        return SnakeOptimizerSingle.Snake_status.SNAKE_SUCCESS;
//      case SNAKE_STOP_ENERGYDIFF:
//        if (Math.abs((this.energy - this.oldEnergy) / this.oldEnergy) < snakeParamsSingle
//            .getEnergyImprove())
//          return SnakeOptimizerSingle.Snake_status.SNAKE_DONE;
//        return SnakeOptimizerSingle.Snake_status.SNAKE_SUCCESS;
//      case SNAKE_STOP_AREADIFF:
//        int[][] maskNew = this.snake.getBinaryMask(this.iWidth, this.iHeight);
//        int[][] maskOld = oldSnake.getBinaryMask(this.iWidth, this.iHeight);
//        int snakeSize_new = 0;
//        int snakeSize_old = 0;
//        for (int y = 0; y < this.iHeight; ++y) {
//          for (int x = 0; x < this.iWidth; ++x) {
//            if (maskOld[y][x] > 0)
//              snakeSize_old++;
//            if (maskNew[y][x] > 0)
//              snakeSize_new++;
//          }
//        }
//        double areaChange = (double) snakeSize_new / (double) snakeSize_old;
//        if (Math.abs(1.0 - areaChange) < snakeParamsSingle.getAreaFraction()
//            || this.itCounter > snakeParamsSingle.getIterations())
//          return SnakeOptimizerSingle.Snake_status.SNAKE_DONE;
//        return SnakeOptimizerSingle.Snake_status.SNAKE_SUCCESS;
//      case SNAKE_STOP_IT_AND_NOMOVE:
//        // check for point correspondences
//        // set gamma to 0, so that point is not moving anymore
//        
//        
//        
//        
//        
//        
//    }
    
    try {
	    System.in.read();
    } catch (IOException e) {
	    // TODO Auto-generated catch block
	    e.printStackTrace();
    }
    
    if (pointsMoved < 1) {
	System.out.println("Fertig!!!");
    	return SnakeOptimizerSingle.Snake_status.SNAKE_DONE;
    }
    return SnakeOptimizerSingle.Snake_status.SNAKE_SUCCESS;
  }

	@Override
  public SnakeOptimizerSingle clone() {
	  // TODO Auto-generated method stub
	  return null;
  }

	@Override
  protected void showSnake() {
	  // TODO Auto-generated method stub
	  
  }

	@Override
  protected void saveSnake() {
	  // TODO Auto-generated method stub
	  
  }

	/**
	 * Print important parameters to standard output stream.
	 * 
	 * @throws ALDOperatorException
	 */
	@Override
  public void printParams() {
		
	}

//  /**
//   * Method to update gamma in an online fashion
//   * 
//   * @param oldS
//   *          Old snake before optimization step.
//   * @param newS
//   *          New snake after optimization step.
//   * @param g
//   *          Update strategy for gamma.
//   */
//  private void updateGamma(Snake oldS, Snake newS,
//      SnakeOptimizerSingleParams.Gamma_update_strategy g) {
//    // convert parameter pointer
//    SnakeOptimizerSingleParams snakeParamsSingle = (SnakeOptimizerSingleParams) this.snakeParams;
//
//    SnakeOptimizerSingleParams.Gamma_update_strategy gas = g;
//    if (!snakeParamsSingle.getAdaptGamma())
//      gas = SnakeOptimizerSingleParams.Gamma_update_strategy.GAMMA_UPDATE_NONE;
//    switch (gas) {
//      case GAMMA_UPDATE_NONE: // no update at all, just resizing
//        System.out.println("No gamma update!");
//        this.gammaAdaptive = new double[this.snake.getPointNum() * 2][1];
//        for (int i = 0; i < this.snake.getPointNum() * 2; ++i)
//          this.gammaAdaptive[i][0] = snakeParamsSingle.getGamma();
//        break;
//      case GAMMA_UPDATE_FIXED_DECREASE: // fixed decrement
//        System.out.println("Decreasing gamma!");
//        double[][] newGamma = new double[this.snake.getPointNum() * 2][1];
//        double oldVal = this.gammaAdaptive[0][0];
//        double newVal = oldVal - oldVal * 0.01;
//        for (int i = 0; i < this.snake.getPointNum() * 2; ++i)
//          newGamma[i][0] = newVal;
//        this.gammaAdaptive = newGamma;
//        break;
//      case GAMMA_UPDATE_POINTWISE_EXT_ENERGY: // analysis of local external
//        // energy
//        System.out.println("Pointwise gamma update!");
//        double[][] gammaUpdate = new double[newS.getPointNum() * 2][1];
//        /*
//         * Check every snake point for gamma update.
//         */
//        for (int i = 0; i < newS.getPointNum(); ++i) {
//          // get current snake point
//          SnakePoint2D curPoint = newS.getSnakePoints().elementAt(i);
//          /*
//           * Do gamma update only for points which are not added in this
//           * iteration step.
//           */
//          double energyCur = 0.0;
//          double energyOld = 0.0;
//          double diff = 0.0;
//          if (curPoint.getOldId() >= 0) {
//            // current external energy from snake point at current position in
//            // the snake
//            energyCur = snakeParamsSingle.getExternalEnergy().getValue_norm(
//                curPoint.x, curPoint.y);
//            // get old values from current snake point from its old position in
//            // the snake
//            SnakePoint2D oldPoint = oldS.getSnakePoints().elementAt(
//                curPoint.getOldId());
//            // get old external energy on current snake point
//            energyOld = snakeParamsSingle.getExternalEnergy().getValue_norm(
//                oldPoint.x, oldPoint.y);
//            // difference between old and new external energy in percent of the
//            // old energy value
//            diff = ((energyCur - energyOld) / (energyOld));
//            /*
//             * Update the gamma value if external energy is decreased more than
//             * 5 percent.
//             */
//            if ((int) (diff * 100.0) > 5) {
//              // get old gamma value on current snake point
//              double gammaOld = this.gammaAdaptive[curPoint.getOldId()][0];
//              // update gamma for current snake point about 10 percent
//              double gammaNew = gammaOld - ((diff / energyOld) * gammaOld);
//              // set new gamma for current snake point
//              gammaUpdate[i][0] = gammaNew;
//              gammaUpdate[i + newS.getPointNum()][0] = gammaNew;
//            } else { // no update, keep gamma
//              /*
//               * Case 1: Energy is not changing anymore, so set the point as
//               * fix. Problem is, that energy is not changing anymore, but the
//               * local minimum is not reached. How to handle this ?
//               */
//              // ToDo check this...
//              gammaUpdate[i][0] = 0.0;
//              gammaUpdate[i + newS.getPointNum()][0] = 0.0;
//            }
//            /*
//             * Case 2: New energy value is smaller than the new energy
//             * value...how to handle gamma ?
//             */
//            // ToDo check this...
//            // if((int) (diff * 100.0) < 0.0) {
//            // 
//            // }
//          } else { // gamma for a new added point is set to initial gamma value
//            gammaUpdate[i][0] = snakeParamsSingle.getGamma();
//            gammaUpdate[i + newS.getPointNum()][0] = snakeParamsSingle
//                .getGamma();
//          }
//          // System.out.println(energyOld + ", " + energyCur + " | " + diff
//          // + " | " + gammaUpdate[i][0] + ", "
//          // + gammaUpdate[i + newS.getPointNum()][0]);
//        }
//        // update gamma values for current snake
//        this.gammaAdaptive = gammaUpdate;
//
//        break;
//    }
//  }

//  /**
//   * Calculates new average values and variances for snake interior and
//   * background.
//   * 
//   * @param incremental
//   *          If true, calculation is incremental.
//   * 
//   *          TODO implement incremental calculation
//   */
//  private void calcRegionMeans(boolean incremental) {
//
//    int countInside = 0;
//    int countOutside = 0;
//    double sumInside = 0;
//    double sumOutside = 0;
//
//    int[][] changeMask = new int[this.iHeight][this.iWidth];
//
//    // in first iteration mean values need to be calculated with
//    // dummy change mask
//    if (this.snakeMaskOld == null || !incremental) {
//      for (int y = 0; y < this.iHeight; ++y)
//        for (int x = 0; x < this.iWidth; ++x)
//          if (this.snakeMask[y][x] == 0)
//            // pixel lies outside
//            changeMask[y][x] = 0;
//          else
//            // pixel lies inside
//            changeMask[y][x] = 1;
//    }
//    // calculate change mask
//    else {
//      for (int y = 0; y < this.iHeight; ++y) {
//        for (int x = 0; x < this.iWidth; ++x) {
//          // pixel is both times outside
//          if (this.snakeMask[y][x] == 0 && this.snakeMaskOld[y][x] == 0)
//            changeMask[y][x] = 0;
//          // pixel is now inside
//          else if (this.snakeMask[y][x] == 1 && this.snakeMaskOld[y][x] == 0)
//            changeMask[y][x] = 1;
//          // pixel is now outside
//          else if (this.snakeMask[y][x] == 0 && this.snakeMaskOld[y][x] == 1)
//            changeMask[y][x] = 2;
//          // both times outside
//          else
//            changeMask[y][x] = 3;
//        }
//      }
//    }
//    // calculate average intensities
//    for (int y = 0; y < this.iHeight; ++y) {
//      for (int x = 0; x < this.iWidth; ++x) {
//        // exclude masked pixels from mean value calculations
//        if (this.excludeMask != null && this.excludeMask[y][x] > 0)
//          continue;
//        switch (changeMask[y][x]) {
//          case 0:
//          case 2:
//            if (this.overlapMask != null) {
//              if (this.overlapMask[y][x] == 0) {
//                countOutside++;
//                sumOutside += this.inImg.getValueInt(x, y);
//              }
//            } else {
//              countOutside++;
//              sumOutside += this.inImg.getValueInt(x, y);
//            }
//            break;
//          case 1:
//          case 3:
//            countInside++;
//            sumInside += this.inImg.getValueInt(x, y);
//            break;
//        }
//      }
//    }
//    // compute final intensity parameters
//    this.c_in = sumInside / countInside;
//    this.c_out = sumOutside / countOutside;
//
//    // store for access from outside of the class
//    this.final_c_in = this.c_in;
//    this.final_c_out = this.c_out;
//
//    // calculate variance
//    sumInside = 0;
//    sumOutside = 0;
//    countInside = 0;
//    countOutside = 0;
//    for (int y = 0; y < this.iHeight; ++y) {
//      for (int x = 0; x < this.iWidth; ++x) {
//        // exclude masked pixels from mean value calculations
//        if (this.excludeMask != null && this.excludeMask[y][x] > 0)
//          continue;
//        switch (changeMask[y][x]) {
//          case 0:
//          case 2:
//            if (this.overlapMask != null) {
//              if (this.overlapMask[y][x] == 0) {
//                countOutside++;
//                sumOutside += (this.inImg.getValueInt(x, y) - this.c_out)
//                    * (this.inImg.getValueInt(x, y) - this.c_out);
//              }
//            } else {
//              countOutside++;
//              sumOutside += (this.inImg.getValueInt(x, y) - this.c_out)
//                  * (this.inImg.getValueInt(x, y) - this.c_out);
//            }
//            break;
//          case 1:
//          case 3:
//            countInside++;
//            sumInside += (this.inImg.getValueInt(x, y) - this.c_in)
//                * (this.inImg.getValueInt(x, y) - this.c_in);
//            break;
//        }
//      }
//    }
//    // compute final intensity parameters
//    this.final_c_in_var = sumInside / countInside;
//    this.final_c_out_var = sumOutside / countOutside;
//  }

  /***********************************************************************/
  /*** Other utilities. ***/
  /***********************************************************************/
	/**
	 * Calculates total snake energy.
	 * 
	 * @return Value of local energy, -1 if error during calculation.
	 */
	protected double calcSnakeEnergy() {
		double sEnergy = 0.0;
//
//		// process all internal energies
//		if (this.energies == null)
//			return -1.0;
//		Vector<MTBSnakeEnergyPDE> eVec = this.energies.getEnergyList(); 
//		for (MTBSnakeEnergyPDE ener : eVec) {
//			sEnergy += ener.calcEnergy(this);
//		}
		return sEnergy;
	}

//  /**
//   * Specify new overlap mask.
//   */
//  protected void setOverlapMask(int[][] m) {
//    this.overlapMask = m;
//  }

//  /**
//   * Specify new exclude mask.
//   */
//  protected void setExcludeMask(int[][] m) {
//    this.excludeMask = m;
//  }

//  /**
//   * Set background average externally.
//   */
//  protected void setBackgroundAverage(double avg) {
//    this.c_out = avg;
//  }

}
