/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 6030 $
 * $Date: 2012-09-24 14:27:37 +0200 (Mo, 24. Sep 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.thresholds;

import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException.OperatorExceptionType;


import de.unihalle.informatik.Alida.annotations.ALDAOperator;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.annotations.Parameter.Direction;
import de.unihalle.informatik.MiToBo.core.datatypes.MTBImageHistogram;
import de.unihalle.informatik.MiToBo.core.datatypes.images.*;
import de.unihalle.informatik.MiToBo.core.datatypes.wrapper.MTBDoubleData;
import de.unihalle.informatik.MiToBo.core.operator.*;

/**
 * This class implements routines to calculate Otsu's
 * threshold from given grayscale images. 
 * <p>
 * Basically, a threshold is extracted that separates the image
 * histogram in two classes, more or less equal in size
 * and well-separated in terms of their average intensities.
 * <p>
 * Otsu thresholding is based on the following goal function,
 * 
 * {@latex.ilb %preamble{\\usepackage{amssymb}}
 * 		  \\begin{eqnarray*}
 *  		J(T) &=& \\pi_{fg} \\cdot \\pi_{bg} \\cdot ( \\mu_{fg} - \\mu_{bg} )^2, \\\\
 *        \\end{eqnarray*}}
 * 
 * which is to be maximized with regard to threshold T.
 * 
 * @author moeller
 */
@ALDAOperator(genericExecutionMode=ALDAOperator.ExecutionMode.ALL)

public class CalcGlobalThreshOtsu extends MTBOperator {
	
	/**
	 * Input image.
	 */
	@Parameter( label= "Input Image", required = false,
			supplemental = false, dataIOOrder = -2,
			direction = Direction.IN, description = "Input image")
	private transient MTBImage inImg = null;

	/**
	 * Input image.
	 */
	@Parameter( label= "Input Histogram", required = false,
			supplemental = false, dataIOOrder = -1,
			direction = Direction.IN, description = "Input histogram")
	private MTBImageHistogram inHisto = null;

	/**
	 * Result threshold.
	 */
	@Parameter( label= "Threshold", supplemental = false,
			direction = Direction.OUT, description = "Result threshold")
	private MTBDoubleData threshold = null;

	/**
	 * Default constructor.
	 * @throws ALDOperatorException
	 */
	public CalcGlobalThreshOtsu() throws ALDOperatorException {
		// nothing to do here
	}

	/**
	 * Default constructor with input image as argument.
	 * 
	 * @param img	Input image.
	 * @throws ALDOperatorException
	 */
	public CalcGlobalThreshOtsu(MTBImage img) throws ALDOperatorException {
		this.inImg = img;
	}

	/**
	 * Default constructor with input histogram as argument.
	 * 
	 * @param histo Input histogram.
	 * @throws ALDOperatorException
	 */
	public CalcGlobalThreshOtsu(MTBImageHistogram histo) 
			throws ALDOperatorException {
		this.inHisto = histo;
	}

	/* (non-Javadoc)
	 * @see de.unihalle.informatik.Alida.operator.ALDOperator#validateCustom()
	 */
	@Override
  public void validateCustom() throws ALDOperatorException {
		if (this.inImg == null && this.inHisto == null)
			throw new ALDOperatorException(OperatorExceptionType.OPERATE_FAILED, 
				"[CalcGlobalThreshOtsu] " +
						"No input data, specify either an image or a histogram!");
	}
	
	/**
	 * Get the resulting threshold.
	 * 
	 * @return	Otsu threshold.
	 */
	public MTBDoubleData getOtsuThreshold() {
		return this.threshold;
	}
		
	@Override
	protected void operate() {
		if (this.verbose.booleanValue())
			System.out.println("[CalcGlobalThreshOtsu] Processing image...");
		if (this.inImg != null)
			this.threshold= this.calcThreshold(this.inImg);
		else
			this.threshold= this.calcThreshold(this.inHisto);
	}

	/**
	 * Calculate Otsu threshold on given image.
	 * 
	 * @param	image	MiToBo image.
	 * @return	Calculated Otsu threshold.
 	 */
	private MTBDoubleData calcThreshold(MTBImage image) {

		// calculate histogram from given image, 
		// but with not less than 128 bins neither more than 1024 bins
		double minBinValue = image.getMinMaxDouble()[0];
		double maxBinValue = image.getMinMaxDouble()[1];
		
		if (this.verbose.booleanValue()) {
			System.out.println("[CalcGlobalThreshOtsu] Min. histo entry = " + 
																																	minBinValue);
			System.out.println("[CalcGlobalThreshOtsu] Max. histo entry = " + 
																																	maxBinValue);
		}
		
		int binCount = (int)(maxBinValue-minBinValue+1);
		if (binCount < 128)
			binCount = 128;
		if (binCount > 1024) 
			binCount = 1024;

		if (this.verbose.booleanValue()) {
			System.out.println("[CalcGlobalThreshOtsu] Using "+binCount+" bins.");
		}
		MTBImageHistogram histo = 
			new MTBImageHistogram(image,binCount,minBinValue,maxBinValue);
		return this.calcThreshold(histo);
	}

	/**
	 * Calculate Otsu threshold on given histogram.
	 * 
	 * @param	histo 	Histogram.
	 * @return	Calculated Otsu threshold.
 	 */
	private MTBDoubleData calcThreshold(MTBImageHistogram histo) {

		int minBinID = 0;
		int maxBinID = histo.getSize()-1;
		int binCount = histo.getSize();
		
		// normalize histogram
		histo.normalize();
		
		// calculate mean
		double mean= histo.getMean();
		if (this.verbose.booleanValue()) {
			System.out.println("[CalcGlobalThreshOtsu] Mean value = " + mean);
		}
		
		// search threshold
		double j_max= -1;
		
		int thresh= 0;
		for (int t=0; t<binCount; ++t) {
			
			double omega_1= 0;
			for (int i=minBinID; i<=t; ++i)
				omega_1= omega_1 + histo.getBinValue(i);
			double omega_2= 0;
			for (int i=t+1; i<=maxBinID; ++i)
				omega_2= omega_2 + histo.getBinValue(i);
				
			double mu_1= 0;
			for (int i=minBinID; i<=t; ++i)
				mu_1= mu_1 + histo.mapIndexToValue(i)*histo.getBinValue(i)/omega_1;
			double mu_2= 0;
			for (int i=t+1; i<=maxBinID; ++i)
				mu_2= mu_2 + histo.mapIndexToValue(i)*histo.getBinValue(i)/omega_2;
			
			if (this.verbose.booleanValue()) {
				System.out.print("[CalcGlobalThreshOtsu] Testing " +  
																							histo.mapIndexToValue(t));
				System.out.print("... J =  " + 
					(omega_1*(mu_1-mean)*(mu_1-mean) + omega_2*(mu_2-mean)*(mu_2-mean)) + "\n");
			}
			if (  omega_1*(mu_1-mean)*(mu_1-mean) + omega_2*(mu_2-mean)*(mu_2-mean) 
					> j_max) {
				j_max = 
					omega_1*(mu_1-mean)*(mu_1-mean) + omega_2*(mu_2-mean)*(mu_2-mean);
				thresh= t;
			}
		}
		if (this.verbose.booleanValue()) {
			System.out.println("[CalcGlobalThreshOtsu] Otsu threshold = " + 
																								histo.mapIndexToValue(thresh));
		}
		return new MTBDoubleData(new Double(histo.mapIndexToValue(thresh)));
	}
}
