/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 4636 $
 * $Date: 2011-12-01 10:57:20 +0100 (Do, 01. Dez 2011) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.tools;

import de.unihalle.informatik.MiToBo.core.datatypes.defines.MTBConstants;

/**
 * <pre>
 * Class to convert values from HSI color space into other color spaces like RGB
 * or so.
 * 
 * H: color value.
 * S: saturation of color.
 * I: color intensity.
 * 
 * Ranges of HSI:
 *   H in [0,360) (representing degrees),
 *   S and I in [0, 1] (representing [0, 100 %]).
 * </pre>
 * 
 * 
 * @author misiak
 */



public class HSIConverter {
  /**
   * Hue of HIS space.
   */
  private double h;
  /**
   * Saturation of HIS space.
   */
  private double s;
  /**
   * Intensity of HIS space.
   */
  private double i;

  /**
   * Standard constructor.
   */
  public HSIConverter() {
    this.h = 0.0;
    this.s = 0.0;
    this.i = 0.0;
  }

  /**
   * Constructor to create a HSIconverter object
   * 
   * @param h
   *          hue of HIS space
   * @param s
   *          saturation of HIS space
   * @param i
   *          intensity of HIS space
   */
  public HSIConverter(double h, double s, double i) {
    this.h = h;
    this.s = s;
    this.i = i;
  }

  /**
   * Method to convert a single color from HSI color space into RGB color space.
   * 
   * @return Integer array with values for r,g,b.
   */
  public int[] toRGB() {
    int[] rgbColor = new int[3];
    // make sure arguments stay in range [0�,360�) for H and in [0,1] for S,I
    h = Math.max(0, Math.min(360.0, h));
    s = Math.max(0, Math.min(1.0, s));
    i = Math.max(0, Math.min(1.0, i));
    // test if color is monochrome (gray)
    if (s < MTBConstants.epsilon) {
      rgbColor[0] = (int) Math.round(i * 255);
      rgbColor[1] = (int) Math.round(i * 255);
      rgbColor[2] = (int) Math.round(i * 255);
      return rgbColor;
    } else {
      int Hi = (int) Math.floor(h / 60.0);
      double f = (h / 60) - Hi;
      double p = i * (1 - s);
      double q = i * (1 - s * f);
      double t = i * (1 - s * (1 - f));
      // calculate r,g,b values for RGB color space
      switch (Hi) {
        case 0:
          rgbColor[0] = (int) Math.round(i * 255);
          rgbColor[1] = (int) Math.round(t * 255);
          rgbColor[2] = (int) Math.round(p * 255);
          break;
        case 1:
          rgbColor[0] = (int) Math.round(q * 255);
          rgbColor[1] = (int) Math.round(i * 255);
          rgbColor[2] = (int) Math.round(p * 255);
          break;
        case 2:
          rgbColor[0] = (int) Math.round(p * 255);
          rgbColor[1] = (int) Math.round(i * 255);
          rgbColor[2] = (int) Math.round(t * 255);
          break;
        case 3:
          rgbColor[0] = (int) Math.round(p * 255);
          rgbColor[1] = (int) Math.round(q * 255);
          rgbColor[2] = (int) Math.round(i * 255);
          break;
        case 4:
          rgbColor[0] = (int) Math.round(t * 255);
          rgbColor[1] = (int) Math.round(p * 255);
          rgbColor[2] = (int) Math.round(i * 255);
          break;
        case 5:
          rgbColor[0] = (int) Math.round(i * 255);
          rgbColor[1] = (int) Math.round(p * 255);
          rgbColor[2] = (int) Math.round(q * 255);
          break;
        case 6:
          rgbColor[0] = (int) Math.round(i * 255);
          rgbColor[1] = (int) Math.round(t * 255);
          rgbColor[2] = (int) Math.round(p * 255);
          break;
      }
      return rgbColor;
    }
  }
}
